/* eslint-disable no-async-promise-executor */
require('./type/thirdweb.type')
const EventEmitter = require('events')
const { promisify } = require('util')
const url = require('url')
const rpc = require('json-rpc2')
const hazama = require('hazama-core')
const bitcoin = require('bitcoinjs-lib')
const fs = require('fs')
const messages = require('./constant/messages')
const {
  ThirdwebSDK,
  getNativeTokenByChainId,
  NATIVE_TOKEN_ADDRESS,
  getRpcUrl,
  getContractAddressByChainId,
  EditionDropInitializer,
  NFTDropInitializer,
  TokenInitializer,
  NFTCollectionInitializer,
  MultiwrapInitializer,
} = require('@thirdweb-dev/sdk')
const TokenType = require('./constant/token-type')
const {
  getRpcUrlsByNetWork,
  getAlchemyApiKey,
  getAlchemyNetwork,
  WaitingAlchemyPeriod,
  SUPPORTED_CHAIN_NAME,
  HazamaHost: HOST,
  getChainId,
} = require('./constant/network-info')
const { ERC20MetadataAbi } = require('./constant/abi')
const ethers = require('ethers')
const {
  isArray,
  isArrayFull,
  sleep,
  rejectNull,
  objectToLowerCase,
  generatePath,
  jointAsm,
  iterate,
  isStringFull,
  isObjectFull,
  getContractInfoFromPath,
  getContractName,
  encodeQueryData,
  isNil,
} = require('./utils/common')
const ContractType = require('./constant/contract-type')
const {
  StandardContractType,
  HazBaseContractType,
} = require('./constant/standard-contract-type')
const {
  CustomContract,
  getParentContractType,
} = require('./constant/custom-contract')
const TokenStandard = require('./constant/token-standard')
const DEFAULT_QUERY_ALL_COUNT = 100
const { Alchemy } = require('alchemy-sdk')
const { parse } = require('qs')
const Role = require('./constant/role')
const moment = require('moment')
const { formatUnits } = require('ethers/lib/utils')
const { v4: uuidv4 } = require('uuid')
const { HttpService } = require('./services/http-service')
const RequestTransactionStatus = require('./constant/request-transaction-status')
const ApiRequestStatus = require('./constant/api-request-status')
const twFactoryAbi = require('@thirdweb-dev/contracts-js/dist/abis/TWFactory.json')
const { ThirdwebStorage } = require('@thirdweb-dev/storage')
const VoteType = require('./constant/vote-type')

/**
 * Library for Hazama base platform
 * @namespace
 * @name hazama-lib
 */
const lib = new EventEmitter()
lib.on('error', (err) => {
  const message = err?.message || (isStringFull(err) ? err : '')

  if (message.includes('transaction underpriced')) {
    throw new Error(messages.TRANSACTIONUNDERPRICED)
  } else if (message.includes('exceeded-max-quantity')) {
    throw new Error(messages.EXCEEDEDMAXQUANTITY)
  } else if (message.includes('word mismatch')) {
    throw new Error(messages.WORDMISMATCH)
  } else if (message.toLowerCase().includes('user rejected transaction')) {
    throw new Error(messages.USERDENIEDTRANSACTION)
  } else if (
    message.includes('User denied transaction signature') ||
    message.includes('User denied message signature') ||
    message.toLowerCase().includes('caller is not owner nor approved')
  ) {
    throw new Error(messages.USERDENIEDTRANSACTION)
  } else if (message.includes('execution-only-allowed-by-owner')) {
    throw new Error(messages.EXECUTIONONLYALLOWEDBYOWNER)
  } else if (
    message.includes('insufficient funds for gas * price + value') ||
    message
      .toLowerCase()
      .includes('insufficient funds for intrinsic transaction cost')
  ) {
    throw new Error(messages.INSUFFICIENTFUNDSFORGAS)
  } else if (message.includes('is not a valid address')) {
    throw new Error(messages.WLINVALIDADDRESS)
  } else if (err?.reason) {
    throw new Error(err?.reason)
  } else {
    throw new Error(messages.UNKNOWN_ERROR)
  }
})

let currentClient
let currentConnection
let generalServerInfo
let hazNetworkSetupManually = false
let currentClientKey
let gatewayUrls
let thirdwebClientId
let thirdwebClientIdFromAPI
let thirdwebSecretKey

function isAbiCompatible(abi, entrypoints) {
  return abi.reduce((p, x) => p && x in entrypoints, true)
}

lib.connect = (endpoint, cb) => {
  try {
    // eslint-disable-next-line n/no-deprecated-api
    const endpointData = url.parse(endpoint)

    let user
    let pass

    if (endpointData.auth) {
      ;[user, pass] = endpointData.auth.split(':')
    }

    let connectHost
    if (endpointData.hostname) {
      connectHost = endpointData.hostname
    } else {
      connectHost = endpointData.pathname
    }

    let connectPort
    if (endpointData.port) {
      connectPort = parseInt(endpointData.port)
    } else {
      if (
        endpointData.protocol === 'wss:' ||
        endpointData.protocol === 'https:'
      ) {
        connectPort = 443
      } else {
        connectPort = 80
      }
    }
    const client =
      typeof window !== 'undefined'
        ? { host: connectHost, connectWebsocket: () => {}, expose: () => {} }
        : rpc.Client.$create(connectPort, connectHost, user, pass)
    if (
      endpointData.protocol === 'wss:' ||
      endpointData.protocol === 'https:'
    ) {
      console.log('wss://' + client.host + ':443' + endpointData.path)
    }
    client.host = endpointData.href

    const connector = async (err, result) => {
      if (err) {
        throw err
      } else if (!result) {
        throw new Error("Couldn't connect to endpoint")
      } else {
        currentConnection = result
        currentConnection.callAsync = promisify(currentConnection.call)
        currentClient = client

        const info = await currentConnection.callAsync('core.info', [])
        if (info.network) {
          generalServerInfo = info
          hazama.networks.mainnet = info.network
          if (hazNetworkSetupManually) {
            setNetworkDeprecationNotice()
          }
        }

        if (cb) cb()

        lib.watch('\\/_\\/block\\..*', (k) => {
          lib.emit('new block', k)
        })

        lib.watch('\\/_\\/tx\\..*', (k) => {
          lib.emit('new tx', k)
        })

        currentClient.expose('event', ([key]) => {
          lib.emit('_raw_event', key)
        })

        currentClient.expose('verifyevent', ([msg]) => {
          lib.emit('_verify_event', msg)
        })

        currentConnection.on('close', () => {
          currentConnection = null
          lib.emit('close')
        })

        lib.emit('connected', currentConnection)
      }
    }

    if (typeof window !== 'undefined') {
      let lastId = 0
      let sendMsg
      const waitCbs = {}

      const connectWebsocket = (connector) => {
        const socket = new WebSocket(client.host)
        socket.addEventListener('open', () => {
          console.log('open', client.host)
          sendMsg = (method, params, cb) => {
            const ob = {
              jsonrpc: '2.0',
              id: lastId,
              method,
              params,
            }
            waitCbs[lastId] = { cb, ts: Date.now(), method, ob }
            socket.send(JSON.stringify(ob))
            lastId++
          }

          socket.addEventListener('message', (event) => {
            let ob
            try {
              ob = JSON.parse(event.data)
            } catch (e) {
              ob = null
            }

            if (ob) {
              if ('id' in ob && ob.id in waitCbs) {
                if (!ob.error) {
                  waitCbs[ob.id].cb(null, ob.result)
                } else {
                  waitCbs[ob.id].cb(ob.error)
                }

                delete waitCbs[ob.id]
              } else if ('method' in ob && ob.method === 'event') {
                lib.emit('_raw_event', ob.params[0])
              }
            }
          })

          socket.addEventListener('close', () => {
            console.log('closed', client.host)
            lib.emit('close')
          })

          socket.on = (e, cb) => {
            socket.addEventListener(e, cb)
          }

          socket.end = () => {
            socket.close()
          }

          const call = (method, params, callback) => {
            sendMsg(method, params, (error, data) => {
              if (error) {
                callback(error, null)
              } else {
                callback(null, data)
              }
            })
          }
          socket.call = call

          connector(null, socket)
        })
      }
      client.connectWebsocket = connectWebsocket
    }

    if (currentConnection != null) return cb()

    client.connectWebsocket(connector)
  } catch (e) {
    lib.emit('error', e)
  }
}

function setNetworkDeprecationNotice() {
  console.log(
    "Deprecation notice: hazlib.setNetwork() isn't necessary on this server instance",
  )
}

lib.reconnect = () => {
  currentClient.connectWebsocket(async (err, result) => {
    if (err) {
      throw err
    } else if (!result) {
      throw new Error("Couldn't connect to endpoint")
    } else {
      currentConnection = result
      currentConnection.callAsync = promisify(currentConnection.call)

      const info = await currentConnection.callAsync('core.info', [])
      if (info.network) {
        generalServerInfo = info
        hazama.networks.mainnet = info.network
        if (hazNetworkSetupManually) {
          setNetworkDeprecationNotice()
        }
      }

      lib.emit('connected', result)

      currentConnection.on('close', () => {
        currentConnection = null
        lib.emit('close')
      })
    }
  })
}

lib.reconnectWithCallback = async (cb) => {
  return new Promise(async (resolve, reject) => {
    try {
      currentClient.connectWebsocket(async (err, result) => {
        try {
          if (err) {
            throw err
          } else if (!result) {
            throw new Error("Couldn't connect to endpoint")
          } else {
            currentConnection = result
            currentConnection.callAsync = promisify(currentConnection.call)

            const info = await currentConnection.callAsync('core.info', [])
            if (info.network) {
              generalServerInfo = info
              hazama.networks.mainnet = info.network
              if (hazNetworkSetupManually) {
                setNetworkDeprecationNotice()
              }
            }

            lib.emit('connected', result)

            currentConnection.on('close', () => {
              currentConnection = null
              lib.emit('close')
            })

            const res = await cb()
            resolve(res)
          }
        } catch (e) {
          reject(e)
        }
      })
    } catch (e) {
      reject(e)
    }
  })
}

lib.setNetwork = (network) => {
  if (generalServerInfo) {
    setNetworkDeprecationNotice()
  } else {
    hazNetworkSetupManually = true
    hazama.networks.mainnet = network
  }
}

lib.compile = (code) => {
  try {
    const result = hazama.asm.compile(code)

    return result
  } catch (e) {
    lib.emit('error', e)
  }
}

lib.envelope = (bytecode, wif) => {
  try {
    if (typeof bytecode === 'string') {
      bytecode = Buffer.from(bytecode, 'hex')
    }

    const ecpair = bitcoin.ECPair.fromWIF(wif)
    const result = hazama.buildEnvelope(ecpair, bytecode)

    return result
  } catch (e) {
    lib.emit('error', e)
  }
}

lib.envelopeWithWallet = async (
  bytecode,
  signer,
  clientKey = null,
  isCount = true,
) => {
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 36,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (typeof bytecode === 'string') {
      bytecode = Buffer.from(bytecode, 'hex')
    }

    const result = await hazama.buildEnvelopeWithWallet(signer, bytecode)

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 36,
      status: RequestTransactionStatus.SUCCEEDED,
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 36,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

lib.template = async (name, params, clientKey = null, isCount = true) => {
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 31,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    const result = await currentConnection.callAsync('core.template', [
      name,
      params,
    ])

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 31,
      status: RequestTransactionStatus.SUCCEEDED,
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 31,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

lib.get = async (key, clientKey = null, isCount = true) => {
  let result
  const targetClientKey = clientKey || currentClientKey || null
  const maxTries = 5
  let count = 0
  let isPassed = false
  const getFnc = async () => {
    return currentConnection.callAsync('core.get', [key])
  }

  await checkRequestTransactionStatus({
    functionId: 32,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    while (!isPassed && count < maxTries) {
      try {
        result = isObjectFull(currentConnection)
          ? await getFnc()
          : await lib.reconnectWithCallback(getFnc)
        isPassed = true
      } catch (e) {
        count++
      }
    }

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 32,
      status: RequestTransactionStatus.SUCCEEDED,
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 32,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

lib.iterate = async (options, amount) => {
  const iterId = await currentConnection.callAsync('core.inititerator', [
    options,
  ])
  let finished = false
  const next = async () => {
    if (finished) {
      return null
    } else {
      const response = await currentConnection.callAsync('core.fetchiterator', [
        iterId,
        amount,
      ])
      if (response) {
        const { hadValues, res: results } = response

        if (!hadValues) {
          // (!results/*.length < amount*/) {
          finished = true
        }

        return results
      } else {
        finished = true
        return null
      }
    }
  }

  return next
}

lib.tx = async (data, clientKey = null, isCount = true) => {
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 37,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (Buffer.isBuffer(data)) {
      data = data.toString('hex')
    }

    const result = await currentConnection.callAsync('core.tx', [data])

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 37,
      status: RequestTransactionStatus.SUCCEEDED,
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 37,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

lib.txwithoutmempool = async (data, clientKey = null, isCount = true) => {
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 38,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (Buffer.isBuffer(data)) {
      data = data.toString('hex')
    }

    const result = await currentConnection.callAsync('core.txwithoutmempool', [
      data,
    ])

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 38,
      status: RequestTransactionStatus.SUCCEEDED,
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 38,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

lib.block = async (data, clientKey = null, isCount = true) => {
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 39,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (Buffer.isBuffer(data)) {
      data = data.toString('hex')
    }

    const result = await currentConnection.callAsync('core.block', [data])

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 39,
      status: RequestTransactionStatus.SUCCEEDED,
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 39,
      reason: e?.message || null,
      isCount: true,
    })
    lib.emit('error', e)
  }
}

lib.watch = (regex, handler) => {
  try {
    const r = new RegExp(regex)
    lib.on('_raw_event', (key) => {
      if (r.test(key)) {
        handler(key)
      }
    })
    currentConnection.callAsync('core.subscribe', [regex])
  } catch (e) {
    lib.emit('error', e)
  }
}

lib.verifyEvents = (handler) => {
  try {
    lib.on('_verify_event', (msg) => {
      handler(msg)
    })
    currentConnection.callAsync('core.verifyevents', [])
  } catch (e) {
    lib.emit('error', e)
  }
}

lib.createWallet = () => {
  try {
    const ecpair = bitcoin.ECPair.makeRandom()
    const network = hazama.networks.mainnet
    const { address } = bitcoin.payments.p2pkh({
      pubkey: ecpair.publicKey,
      network,
    })

    return { address, wif: ecpair.toWIF(), ecpair }
  } catch (e) {
    lib.emit('error', e)
  }
}

lib.walletFromWif = (wif) => {
  try {
    const ecpair = bitcoin.ECPair.fromWIF(wif)
    const network = hazama.networks.mainnet
    const { address } = bitcoin.payments.p2pkh({
      pubkey: ecpair.publicKey,
      network,
    })

    return { address, wif: ecpair.toWIF(), ecpair }
  } catch (e) {
    lib.emit('error', e)
  }
}

lib.decodeCall = (envelope) => {
  try {
    const msg = hazama.openEnvelope(Buffer.from(envelope, 'hex'))
    const lines = hazama.asm
      .unpack(msg.data)
      .filter(
        (x) =>
          x.opName !== 'NOOP' && x.opName !== 'END' && x.opName !== 'REQUIRE',
      )
    const calls = []
    let params = []
    for (let i = 0; i < lines.length; i++) {
      const item = lines[i]
      if (item.opName.startsWith('PUSH')) {
        params.push(item.params[0])
      } else if (item.opName === 'ECALL') {
        calls.push({ [item.params[0]]: params })
        params = []
      }
    }

    return { from: msg.from, calls }
  } catch (e) {
    lib.emit('error', e)
  }
}

lib.makeFileFormat = (filename, cb) => {
  if (
    typeof process !== 'undefined' &&
    process.mainModule &&
    process.mainModule.path
  ) {
    const filetype = require('file-type')
    try {
      const path = require('path')
      const fname = process.mainModule.path + path.sep + filename

      fs.accessSync(fname, fs.constants.R_OK)
      fs.readFile(filename, 'base64', async (err, data) => {
        if (err) return cb(err, null)

        const fileType = await filetype.fromFile(fname)
        if (!fileType)
          return cb(new Error(`${filename} is an unsupported`), null)

        const dataFormat = {}
        dataFormat.size = data.length
        dataFormat.mime = fileType.mime
        dataFormat.ext = fileType.ext
        for (let i = 0; ; i++) {
          const splitData = data.substr(i * 50000, 50000)
          if (splitData.length <= 0) break
          dataFormat[fileType.mime + (i > 0 ? `/${i}` : '')] = splitData
        }

        cb(null, dataFormat)
      })
    } catch (e) {
      cb(e, null)
    }
  }
}

function getBase64(data) {
  let base64data = ''
  if (!data.mime) throw new Error('file format not found')

  for (let i = 0; ; i++) {
    const key = data.mime + (i > 0 ? `/${i}` : '')
    if (!(key in data)) break
    base64data += data[key]
  }

  if (base64data.length <= 0) throw new Error('empty file')

  return base64data
}

lib.getFileBase64 = (data) => {
  try {
    const result = getBase64(data)

    return result
  } catch (e) {
    lib.emit('error', e)
  }
}

lib.getFileBuffer = (data) => {
  try {
    const result = Buffer.from(getBase64(data), 'base64')

    return result
  } catch (e) {
    lib.emit('error', e)
  }
}

lib.waitBlock = () =>
  new Promise((resolve) => {
    lib.once('new block', () => {
      setTimeout(resolve, 10) // Needs a tiny bit of wait
    })
  })

lib.waitTx = (txid) =>
  new Promise((resolve) => {
    const watcher = (k) => {
      if (k.indexOf(txid) >= 0) {
        resolve(k)
      }
      lib.removeListener('new tx', watcher)
    }
    lib.on('new tx', watcher)
  })

lib.waitMinedTx = (txid) =>
  new Promise(async (resolve) => {
    const blockWatcher = async () => {
      const res = await lib.get(`/_/tx.${txid}.feds`, null, false)

      if (res) {
        const mined = Object.entries(res)
          .map(([k, v]) => v !== null && v.length > 0) // Check if it's mined by this each fed
          .reduce((p, x) => p && x, true) // All feds should be true to get it mined
        if (mined) {
          // lib.removeListener('new block', blockWatcher)
          resolve(txid)
        }
      }
    }

    const watcher = async (k) => {
      if (k.indexOf(txid) >= 0) {
        lib.on('new block', blockWatcher)

        const alreadyThere = await lib.get(`/_/tx.${txid}.block`, null, false)
        if (alreadyThere !== null) {
          blockWatcher()
        }
      }
      // lib.removeListener('new tx', watcher)
    }
    lib.on('new tx', watcher)

    const alreadyThere = await lib.get(`/_/tx.${txid}`, null, false)
    if (alreadyThere !== null) {
      await watcher(`/_/tx.${txid}`)
    }
  })

lib.waitFinishedTx = (lib, txid) => {
  return new Promise((resolve, reject) => {
    try {
      const timer = setInterval(async () => {
        const blockId = await lib.get(`/_/tx.${txid}.block`, null, false)
        if (blockId) {
          resolve(txid)
          clearInterval(timer)
        }
      }, 500)
    } catch (e) {
      reject(e)
    }
  })
}

lib.disconnect = () => {
  try {
    currentConnection.end()
  } catch (e) {
    lib.emit('error', e)
  }
}

lib.searchContractsByAbi = async (
  path,
  abi,
  clientKey = null,
  isCount = true,
) => {
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 48,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    const ret = []
    const replacedPath = path.replace(/\//g, '\\/')
    const matcher = '^' + replacedPath + '[a-zA-Z]+\\.meta$'
    const iter = await lib.iterate({ gt: path, match: matcher }, 10)
    let chunk = await iter()
    while (chunk !== null) {
      for (let i = 0; i < chunk.length; i++) {
        const { key, value } = chunk[i]
        const entrypointsKey = key.replace('.meta', '.entrypoints')
        const contractKey = key.replace('.meta', '')
        const entrypoints = await lib.get(entrypointsKey, null, false)

        if (isAbiCompatible(abi, entrypoints)) {
          ret.push([contractKey, value])
        }
      }

      chunk = await iter()
    }

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 48,
      status: RequestTransactionStatus.SUCCEEDED,
      isCount,
    })

    return ret
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 48,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * Check request transaction status
 *
 * @private
 *
 * @param {object} params parameters for creating request transaction
 * @param {string} [params.clientKey] client key
 * @param {number} params.functionId function id
 * @return {Promise<void>}}
 */
async function checkRequestTransactionStatus({
  clientKey,
  functionId,
  isCount = true,
}) {
  if (!isCount) return
  try {
    if (!functionId && functionId !== 0)
      throw new Error(messages.FUNCTION_ID_UNDEFINED)

    const data = {
      functionId,
      ...(clientKey ? { clientKey } : {}),
    }

    const httpService = new HttpService()
    const { data: dataForGetRequestTransactionStatus } = await httpService.get({
      url: '/api/app/request-transaction/check?' + encodeQueryData(data),
    })

    // update field thirdwebClientIdFromAPI
    thirdwebClientIdFromAPI =
      dataForGetRequestTransactionStatus?.data?.thirdwebClientId || null

    if (!thirdwebSecretKey && !thirdwebClientId && !thirdwebClientIdFromAPI)
      throw new Error(messages.THIRDWEB_API_KEY_NOTFOUND)
  } catch (e) {
    // reset value of field thirdwebClientIdFromAPI
    thirdwebClientIdFromAPI = null

    if (e?.constructor?.name === 'AxiosError') {
      const { response } = e
      throw new Error(response?.data?.errors?.[0]?.message || 'Bad request')
    } else {
      throw e
    }
  }
}

/**
 * Create failed request transaction
 *
 * @private
 *
 * @param {object} params parameters for creating failed request transaction
 * @param {string} [params.clientKey] client key
 * @param {number} params.functionId function id
 * @param {string} [params.walletAddress] wallet address
 * @param {string} [params.reason] reason
 * @return {Promise<void>}}
 */
async function createFailedRequestTransaction({
  clientKey,
  functionId,
  walletAddress,
  reason,
  isCount = true,
}) {
  if (!isCount) return
  try {
    if (!functionId && functionId !== 0)
      throw new Error(messages.FUNCTION_ID_UNDEFINED)

    await createRequestTransaction({
      clientKey,
      functionId,
      status: RequestTransactionStatus.FAILED,
      ...(reason ? { reason } : {}),
      ...(walletAddress ? { walletAddress } : {}),
      ...(clientKey ? { clientKey } : {}),
    })
  } catch (e) {
    // do nothing
  }
}

/**
 * Create request transaction
 *
 * @private
 *
 * @param {object} params parameters for creating request transaction
 * @param {string} [params.clientKey] client key
 * @param {number} params.functionId function id
 * @param {string} [params.walletAddress] wallet address
 * @param {string} [params.transactionHash] transaction hash
 * @param {Array.<ApiRequest>} [params.apiRequests] list of api request
 * @param {('succeeded'|'failed')} params.status status
 * @param {string} [params.reason] reason
 * @return {Promise<void>}}
 */
async function createRequestTransaction({
  clientKey,
  functionId,
  walletAddress,
  transactionHash,
  apiRequests = [],
  status,
  reason,
  isCount = true,
}) {
  if (!isCount) return
  try {
    if (!functionId && functionId !== 0)
      throw new Error(messages.FUNCTION_ID_UNDEFINED)

    if (!status) throw new Error(messages.STATUS_UNDEFINED)

    const httpService = new HttpService()
    await httpService.post({
      url: '/api/app/request-transaction',
      data: {
        clientKey,
        functionId,
        status,
        ...(walletAddress ? { walletAddress } : {}),
        ...(transactionHash ? { transactionHash } : {}),
        ...(apiRequests?.length ? { apiRequests } : {}),
        ...(reason ? { reason } : {}),
        ...(clientKey ? { clientKey } : {}),
      },
    })
  } catch (e) {
    // do nothing
  }
}

/**
 * Get caller
 *
 * @private
 *
 * @param {Object} params parameters
 * @param {string} [params.network] the chain name or rpc url
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions
 * @param {string} [params.privateKey] wallet's private key
 * @param {string} [params.wif] wallet import format
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @returns {Promise<string | null>} caller address
 */
async function getCaller({ network, signer, privateKey, wif, host }) {
  return new Promise(async (resolve, reject) => {
    try {
      let caller = null
      if (network !== 'hazama') {
        if (signer) {
          caller = signer?.getAddress() || null
        }

        if (privateKey) {
          const wallet = new ethers.Wallet(privateKey)
          caller = wallet.address
        }

        resolve(caller)
      } else {
        lib.connect(host, async () => {
          try {
            if (wif) {
              const { address } = await lib.walletFromWif(wif)
              caller = address
            }

            if (signer) {
              caller = (await signer?.getAddress()) || null
            }

            resolve(caller)
          } catch (e) {
            reject(e)
          }
        })
      }
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Format transaction hash
 *
 * @private
 *
 * @param {Object} params parameters
 * @param {(TransactionResult|TransactionResult[]|TransactionResultWithId|TransactionResultWithId[]|string)} params.input input
 * @returns {string} string
 */
function formatTransactionHash({ input }) {
  let transactionHash = null
  if (isStringFull(input)) {
    transactionHash = input
  } else if (isArrayFull(input)) {
    const txnHashList = []
    input.forEach((i) => {
      if (i?.receipt?.transactionHash) {
        txnHashList.push(i.receipt.transactionHash)
      }
    })
    transactionHash = txnHashList.join(',')
  } else if (isObjectFull(input)) {
    transactionHash = input?.receipt?.transactionHash || null
  }
  return transactionHash
}

/**
 * Get project id
 *
 * @private
 *
 * @param {Object} params parameters
 * @param {string} params.walletAddress wallet address
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {string} params.transactionHash transaction hash
 * @returns {Promise<string>} string
 */
async function getProjectId({ walletAddress, host = HOST, transactionHash }) {
  if (!walletAddress) throw new Error(messages.WALLET_ADDRESS_UNDEFINED)

  if (!transactionHash) throw new Error(messages.TRANSACTION_HASH_UNDEFINED)

  // eslint-disable-next-line no-async-promise-executor
  const projectId = await new Promise(async (resolve, reject) => {
    lib.connect(host, async () => {
      try {
        let projectId = null
        const history = await lib.getWallet({
          walletAddress,
          host,
          isCount: false,
        })
        if (history) {
          for (const path in history) {
            if (Object.prototype.hasOwnProperty.call(history, path)) {
              const val = history[path]

              if (
                val === transactionHash &&
                new RegExp(
                  `^/hazama/${walletAddress}/projects/[0-9a-z]+$`,
                ).test(path)
              ) {
                projectId = path.split('/').pop()
                break
              }
            }
          }
        }

        resolve(projectId)
      } catch (e) {
        reject(e)
      }
    })
  })
  return projectId
}

/**
 * Get deploy arguments
 *
 * @private
 * @param {object} params parameters for getting deploy arguments
 * @param {string} params.chainId chainId
 * @param {('nft-collection'|'nft-drop'|'edition-drop'|'token'|'multiwrap'|'BurnToRedeem1155'|'Vote')} params.contractType the contract type
 * @param {(NFTContractDeployMetadata|TokenContractDeployMetadata|BundleContractDeployMetadata)} params.contractMetadata contract metadata
 * @param {string} params.caller caller address
 * @param {string} params.contractURI contract uri
 * @param {string} [params.relayerUrl] relayer url (gasless transaction mode)
 * @return {Promise<any[]>} promise
 */
async function getDeployArguments({
  chainId,
  contractType,
  contractMetadata,
  caller,
  contractURI,
  relayerUrl,
}) {
  if (!contractType) throw new Error(messages.CONTRACT_TYPE_UNDEFINED)

  if (
    ![
      ContractType.DropERC721,
      ContractType.DropERC1155,
      ContractType.TokenERC20,
      ContractType.TokenERC721,
      ContractType.Multiwrap,
      CustomContract.ContractType.BurnToRedeem1155,
      CustomContract.ContractType.Vote,
    ].includes(contractType)
  )
    throw new Error(messages.CONTRACT_TYPE_IS_INVALID)

  if (!contractMetadata) throw new Error(messages.CONTRACT_METADATA_UNDEFINED)

  if (!caller) throw new Error(messages.CALLER_UNDEFINED)

  if (!contractURI) throw new Error(messages.CONTRACT_URI_UNDEFINED)

  const openzeppelinForwarder = getContractAddressByChainId(
    chainId,
    'openzeppelinForwarder',
  )
  const biconomyForwarder = getContractAddressByChainId(
    chainId,
    'biconomyForwarder',
  )
  const trustedForwarders = relayerUrl
    ? [relayerUrl, openzeppelinForwarder, biconomyForwarder]
    : [openzeppelinForwarder, biconomyForwarder]
  switch (contractType) {
    case NFTCollectionInitializer.contractType:
    case NFTDropInitializer.contractType: {
      const erc721metadata =
        NFTDropInitializer.schema.deploy.parse(contractMetadata)
      return [
        caller,
        erc721metadata.name,
        erc721metadata.symbol,
        contractURI,
        trustedForwarders,
        erc721metadata.primary_sale_recipient,
        erc721metadata.fee_recipient,
        erc721metadata.seller_fee_basis_points,
        erc721metadata.platform_fee_basis_points,
        erc721metadata.platform_fee_recipient,
      ]
    }
    case EditionDropInitializer.contractType: {
      const erc1155metadata =
        EditionDropInitializer.schema.deploy.parse(contractMetadata)
      return [
        caller,
        erc1155metadata.name,
        erc1155metadata.symbol,
        contractURI,
        trustedForwarders,
        erc1155metadata.primary_sale_recipient,
        erc1155metadata.fee_recipient,
        erc1155metadata.seller_fee_basis_points,
        erc1155metadata.platform_fee_basis_points,
        erc1155metadata.platform_fee_recipient,
      ]
    }
    case TokenInitializer.contractType: {
      const erc20metadata =
        TokenInitializer.schema.deploy.parse(contractMetadata)
      return [
        caller,
        erc20metadata.name,
        erc20metadata.symbol,
        contractURI,
        trustedForwarders,
        erc20metadata.primary_sale_recipient,
        erc20metadata.platform_fee_recipient,
        erc20metadata.platform_fee_basis_points,
      ]
    }
    case MultiwrapInitializer.contractType: {
      const multiwrapMetadata =
        MultiwrapInitializer.schema.deploy.parse(contractMetadata)
      return [
        caller,
        multiwrapMetadata.name,
        multiwrapMetadata.symbol,
        contractURI,
        trustedForwarders,
        multiwrapMetadata.fee_recipient,
        multiwrapMetadata.seller_fee_basis_points,
      ]
    }
    case CustomContract.ContractType.BurnToRedeem1155:
      return [
        caller,
        contractMetadata.name,
        contractMetadata.symbol || '',
        contractURI,
        trustedForwarders,
        contractMetadata.primary_sale_recipient ||
          contractMetadata.platform_fee_recipient,
        contractMetadata.seller_fee_basis_points ||
          contractMetadata.platform_fee_basis_points ||
          0,
      ]

    case CustomContract.ContractType.Vote:
      return [
        caller,
        contractMetadata.name,
        contractURI,
        trustedForwarders,
        contractMetadata.voting_token_address,
        ethers.utils.parseUnits(
          contractMetadata.proposal_token_threshold.toString(),
          18,
        ),
        ethers.utils.parseUnits(
          contractMetadata.voting_token_threshold.toString(),
          18,
        ),
        contractMetadata.voting_quorum_fraction,
        contractMetadata.voting_power,
        contractMetadata.only_owner_executeable,
      ]

    default:
      return []
  }
}

/**
 * Get contract abi
 *
 * @private
 * @param {object} params parameters for getting contract abi
 * @param {('nft-collection'|'nft-drop'|'edition-drop'|'token'|'multiwrap')} params.contractType the contract type
 * @return {Promise<any[]>} promise
 */
async function getContractAbi({ contractType }) {
  if (!contractType) throw new Error(messages.CONTRACT_TYPE_UNDEFINED)

  if (
    ![
      ContractType.DropERC721,
      ContractType.DropERC1155,
      ContractType.TokenERC20,
      ContractType.TokenERC721,
      ContractType.Multiwrap,
    ].includes(contractType)
  )
    throw new Error(messages.CONTRACT_TYPE_IS_INVALID)

  switch (contractType) {
    case NFTDropInitializer.contractType:
      return await NFTDropInitializer.getAbi()

    case EditionDropInitializer.contractType:
      return await EditionDropInitializer.getAbi()

    case TokenInitializer.contractType:
      return await TokenInitializer.getAbi()

    case NFTCollectionInitializer.contractType:
      return await NFTCollectionInitializer.getAbi()

    case MultiwrapInitializer.contractType:
      return await MultiwrapInitializer.getAbi()

    default:
      return []
  }
}

/**
 * Get contract version
 *
 * @private
 * @param {object} params parameters for getting contract version
 * @param {('nft-collection'|'nft-drop'|'edition-drop'|'token'|'multiwrap'|'BurnToRedeem1155'|'Vote')} params.contractType the contract type
 * @return {(number|string|undefined)}
 */
function getContractVersion({ contractType }) {
  if (!contractType) throw new Error(messages.CONTRACT_TYPE_UNDEFINED)

  const contractVersions = {
    [ContractType.DropERC721]: 4,
    [ContractType.TokenERC721]: 1,
    [ContractType.SignatureDrop]: 4,
    [ContractType.Multiwrap]: 1,
    [ContractType.DropERC1155]: 4,
    [ContractType.TokenERC1155]: 1,
    [ContractType.DropERC20]: 2,
    [ContractType.TokenERC20]: 1,
    [ContractType.VoteERC20]: 1,
    [ContractType.Split]: 1,
    [ContractType.Marketplace]: 2,
    [ContractType.Pack]: 2,
    // Custom Contract
    [CustomContract.ContractType.BurnToRedeem1155]: '1.0.6',
    [CustomContract.ContractType.Vote]: '1.0.6',
  }

  return contractVersions?.[contractType] || undefined
}

/**
 * Deploy a proxy contract of a given implementation via the given factory
 *
 * @private
 * @param {object} params parameters for deploying via factory
 * @param {string} params.network the chain name or rpc url
 * @param {ThirdwebSDK} params.sdk Thirdweb SDK
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions
 * @param {string} [params.privateKey] wallet's private key
 * @param {('nft-collection'|'nft-drop'|'edition-drop'|'token'|'multiwrap')} params.contractType the contract type
 * @param {(NFTContractDeployMetadata|TokenContractDeployMetadata|BundleContractDeployMetadata)} params.contractMetadata contract metadata
 * @param {number} [params.version]
 * @param {string} [params.relayerUrl] relayer url (gasless transaction mode)
 * @return {Promise<string>} promise
 */
async function deployViaFactory({
  network,
  sdk,
  signer,
  privateKey,
  contractType,
  contractMetadata,
  version,
  relayerUrl,
}) {
  if (!network) throw new Error(messages.NETWORK_UNDEFINED)

  if (!contractType) throw new Error(messages.CONTRACT_TYPE_UNDEFINED)

  if (!contractMetadata) throw new Error(messages.CONTRACT_METADATA_UNDEFINED)

  if (!contractMetadata.name) throw new Error(messages.CONTRACT_NAME_UNDEFINED)

  if (
    ![
      ContractType.DropERC721,
      ContractType.DropERC1155,
      ContractType.TokenERC20,
      ContractType.TokenERC721,
      ContractType.Multiwrap,
    ].includes(contractType)
  )
    throw new Error(messages.CONTRACT_TYPE_IS_INVALID)

  const contractVersion = version || getContractVersion({ contractType })
  const contractName = getContractName(contractType)
  const chainId = getChainId(network)
  const factoryAddress = getContractAddressByChainId(chainId, 'twFactory')
  const writeContract = new ethers.Contract(
    factoryAddress,
    twFactoryAbi,
    sdk.getSignerOrProvider(),
  ) // setup the read only contract
  const readContract = writeContract.connect(sdk.getProvider())
  const encodedType = ethers.utils.formatBytes32String(contractName)
  const implementationAddress = await readContract.getImplementation(
    encodedType,
    contractVersion,
  )
  const caller = await getCaller({ network, signer, privateKey })
  
  let options = {
    ...(isArrayFull(gatewayUrls) ? { gatewayUrls } : {}),
    ...(isNil(thirdwebSecretKey) &&
    (!isNil(thirdwebClientId) || !isNil(thirdwebClientIdFromAPI))
      ? { clientId: thirdwebClientId || thirdwebClientIdFromAPI }
      : {}),
    ...(!isNil(thirdwebSecretKey) ? { secretKey: thirdwebSecretKey } : {}),
  }

  const storage = new ThirdwebStorage(options)
  const contractURI = await storage.upload(contractMetadata)
  const initializerFunction = 'initialize'
  const initializerArgs = await getDeployArguments({
    chainId,
    contractType,
    contractMetadata,
    caller,
    contractURI,
    relayerUrl,
  })
  const contractAbi = await getContractAbi({ contractType })
  const result = await sdk.deployer.deployViaFactory(
    factoryAddress,
    implementationAddress,
    contractAbi,
    initializerFunction,
    initializerArgs,
  )
  return result
}

/**
 * Deploy released contract
 *
 * @private
 * @param {object} params parameters for deploying released contract
 * @param {string} params.network the chain name or rpc url
 * @param {ThirdwebSDK} params.sdk Thirdweb SDK
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions
 * @param {string} [params.privateKey] wallet's private key
 * @param {('BurnToRedeem1155'|'Vote')} params.contractType the contract type
 * @param {(NFTContractDeployMetadata|TokenContractDeployMetadata|BundleContractDeployMetadata)} params.contractMetadata contract metadata
 * @param {number|string} [params.version]
 * @param {string} [params.relayerUrl] relayer url (gasless transaction mode)
 * @return {Promise<string>} promise
 */
async function deployReleasedContract({
  network,
  sdk,
  signer,
  privateKey,
  contractType,
  contractMetadata,
  version,
  relayerUrl,
}) {
  if (!network) throw new Error(messages.NETWORK_UNDEFINED)

  if (!contractType) throw new Error(messages.CONTRACT_TYPE_UNDEFINED)

  if (!contractMetadata) throw new Error(messages.CONTRACT_METADATA_UNDEFINED)

  if (!contractMetadata.name) throw new Error(messages.CONTRACT_NAME_UNDEFINED)

  if (
    ![
      CustomContract.ContractType.BurnToRedeem1155,
      CustomContract.ContractType.Vote,
    ].includes(contractType)
  )
    throw new Error(messages.CONTRACT_TYPE_IS_INVALID)

  const contractVersion = version || getContractVersion({ contractType })

  const contractName = getContractName(contractType)
  const caller = await getCaller({ network, signer, privateKey })

  let options = {
    ...(isArrayFull(gatewayUrls) ? { gatewayUrls } : {}),
    ...(isNil(thirdwebSecretKey) &&
    (!isNil(thirdwebClientId) || !isNil(thirdwebClientIdFromAPI))
      ? { clientId: thirdwebClientId || thirdwebClientIdFromAPI }
      : {}),
    ...(!isNil(thirdwebSecretKey) ? { secretKey: thirdwebSecretKey } : {}),
  }

  const storage = new ThirdwebStorage(options)
  const contractURI = await storage.upload(contractMetadata)

  const chainId = getChainId(network)
  const initializerArgs = await getDeployArguments({
    chainId,
    contractType,
    contractMetadata,
    caller,
    contractURI,
    relayerUrl,
  })

  const result = await sdk.deployer.deployReleasedContract(
    CustomContract.releaserAddress,
    contractName,
    initializerArgs,
    contractVersion,
  )
  return result
}

/**
 * Fetch token metadata
 *
 * @private
 * @param {object} params parameters for fetching token metadata
 * @param {BigNumberish} params.tokenId the id of the token
 * @param {string} params.tokenUri token uri
 * @param {ThirdwebStorage} params.storage Thirdweb Storage
 * @return {Promise<any>} promise
 */
async function fetchTokenMetadata({ tokenId, tokenUri, storage }) {
  const FALLBACK_METADATA = {
    name: 'Failed to load NFT metadata',
  }

  // check for base64 encoded JSON
  if (
    tokenUri.startsWith('data:application/json;base64') &&
    typeof Buffer !== 'undefined'
  ) {
    const base64 = tokenUri.split(',')[1]
    const jsonMetadata = JSON.parse(
      Buffer.from(base64, 'base64').toString('utf-8'),
    )
    return {
      ...jsonMetadata,
      id: ethers.BigNumber.from(tokenId).toString(),
      uri: tokenUri,
    }
  }

  // handle dynamic id URIs (2 possible formats)
  const parsedUri = tokenUri.replace(
    '{id}',
    ethers.utils
      .hexZeroPad(ethers.BigNumber.from(tokenId).toHexString(), 32)
      .slice(2),
  )
  let jsonMetadata

  try {
    jsonMetadata = await storage.downloadJSON(parsedUri)
  } catch (err) {
    const unparsedTokenIdUri = tokenUri.replace(
      '{id}',
      ethers.BigNumber.from(tokenId).toString(),
    )
    try {
      jsonMetadata = await storage.downloadJSON(unparsedTokenIdUri)
    } catch (e) {
      console.warn(
        `failed to get token metadata: ${JSON.stringify({
          tokenId: tokenId.toString(),
          tokenUri,
        })} -- falling back to default metadata`,
      )
      jsonMetadata = FALLBACK_METADATA
    }
  }

  return {
    ...jsonMetadata,
    id: ethers.BigNumber.from(tokenId).toString(),
    uri: tokenUri,
  }
}

/**
 * Deploy bundle contract
 * @function
 * @name deployBundle
 * @memberof hazama-lib/Bundle
 *
 * @param {object} params parameters for deploying bundle contract
 * @param {string} params.network the chain name or rpc url
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions.<br/>If private key is not set, Signer is required.
 * @param {string} [params.privateKey] wallet's private key.<br/>If Signer is not set, private key is required.
 * @param {BundleContractDeployMetadata} params.metadata Options for deploying an bundle contract
 * @param {string} [params.relayerUrl] relayer url (gasless transaction mode)
 * @param {string} [params.clientKey] client key
 * @returns {string} the address of the deployed contract
 *
 * @example <caption>[Deploy bundle with privateKey]</caption>
 * import hazlib from hazama-lib
 * ...
 * const network = 'polygon'
 * const privateKey = 'dummy'
 * const metadata = {
 *   name: 'My bundle'
 * }
 * const clientKey = 'dummy'
 *
 * const contractAddress = hazlib.deployBundle({ network, privateKey, metadata, clientKey })
 * ...
 *
 * @example <caption>[Deploy bundle with signer and relayer]</caption>
 * import hazlib from hazama-lib
 * ...
 * const network = 'polygon'
 * const signer = <signer object>
 * const relayerUrl = 'https://api.defender.openzeppelin.com/autotasks/0f7cf308-.../runs/webhook/fd447953-...' // Openzeppelin webhook URI
 * const metadata = {
 *   name: 'My bundle'
 * }
 * const clientKey = 'dummy'
 *
 * const contractAddress = hazlib.deployBundle({ network, signer, metadata, relayerUrl, clientKey })
 * ...
 */
lib.deployBundle = async ({
  network,
  signer,
  privateKey,
  metadata,
  relayerUrl,
  clientKey,
  isCount = true,
}) => {
  let caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 0,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    let options = {
      readonlySettings: {
        rpcUrl: getRpcUrl(network),
      },
      ...(isArrayFull(gatewayUrls) ? { gatewayUrls } : {}),
      ...(isNil(thirdwebSecretKey) &&
      (!isNil(thirdwebClientId) || !isNil(thirdwebClientIdFromAPI))
        ? { clientId: thirdwebClientId || thirdwebClientIdFromAPI }
        : {}),
      ...(!isNil(thirdwebSecretKey) ? { secretKey: thirdwebSecretKey } : {}),
    }

    if (!network) throw new Error(messages.NETWORK_UNDEFINED)

    if (!signer && !privateKey)
      throw new Error(messages.SIGNER_PRIVATEKEY_UNDEFINED)

    if (!metadata) throw new Error(messages.METADATA_UNDEFINED)

    if (!metadata.name) throw new Error(messages.CONTRACT_NAME_UNDEFINED)

    if (relayerUrl) {
      options = {
        ...options,
        gasless: {
          openzeppelin: { relayerUrl },
        },
      }
    }

    let sdk = new ThirdwebSDK(network, options)
    if (signer) {
      sdk.updateSignerOrProvider(signer)
    }

    if (privateKey) {
      sdk = new ThirdwebSDK(
        new ethers.Wallet(
          privateKey,
          ethers.getDefaultProvider(getRpcUrl(network)),
        ),
        options,
      )
    }

    caller = (await sdk?.signer?.getAddress()) || null
    const result = await sdk.deployer.deployMultiwrap(metadata)

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 0,
      walletAddress: caller,
      status: RequestTransactionStatus.SUCCEEDED,
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 0,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * Wrap tokens (ERC20/ERC721/ERC1155) into a single wrapped token
 * @function
 * @name wrapTokens
 * @memberof hazama-lib/Bundle
 *
 *
 * @example <caption>[Wrap tokens with privateKey]</caption>
 * import hazlib from hazama-lib
 * ...
 * const network = 'polygon'
 * const privateKey = 'dummy'
 * const contractAddress = '0x...'
 * const contents = {
 *   erc20Tokens: [
 *     {
 *       contractAddress: '0x...',
 *       quantity: '0.8',
 *     },
 *   ],
 *   erc721Tokens: [
 *     {
 *       contractAddress: '0x...',
 *       tokenId: '0',
 *     },
 *   ],
 *   erc1155Tokens: [
 *     {
 *       contractAddress: '0x...',
 *       tokenId: '1',
 *       quantity: '2',
 *     },
 *   ],
 * }
 * const wrappedTokenMetadata = {
 *   name: 'wrapped token name'
 * }
 * const clientKey = 'dummy'
 *
 * const toNumber = hazlib.wrapTokens({ network, privateKey, contractAddress, contents, wrappedTokenMetadata, clientKey })
 * ...
 *
 * @example <caption>[Wrap tokens with signer and relayer]</caption>
 * import hazlib from hazama-lib
 * ...
 * const network = 'polygon'
 * const signer = <signer object>
 * const relayerUrl = 'https://api.defender.openzeppelin.com/autotasks/0f7cf308-.../runs/webhook/fd447953-...' // Openzeppelin webhook URI
 * const contractAddress = '0x...'
 * const contents = {
 *   erc20Tokens: [
 *     {
 *       contractAddress: '0x...',
 *       quantity: '0.8',
 *     },
 *   ],
 *   erc721Tokens: [
 *     {
 *       contractAddress: '0x...',
 *       tokenId: '0',
 *     },
 *   ],
 *   erc1155Tokens: [
 *     {
 *       contractAddress: '0x...',
 *       tokenId: '1',
 *       quantity: '2',
 *     },
 *   ],
 * }
 * const wrappedTokenMetadata = {
 *   name: 'wrapped token name'
 * }
 * const clientKey = 'dummy'
 *
 * const toNumber = hazlib.wrapTokens({ network, signer, contractAddress, contents, wrappedTokenMetadata, relayerUrl, clientKey })
 * ...
 *
 * @param {object} params parameters for wrapping tokens
 * @param {string} params.network the chain name or rpc url
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions.<br/>If private key is not set, Signer is required.
 * @param {string} [params.privateKey] wallet's private key.<br/>If Signer is not set, private key is required.
 * @param {TokensToWrap} params.contents the contents to wrap
 * @param {NFTMetadataOrUri} params.wrappedTokenMetadata metadata to represent the wrapped token bundle
 * @param {string} [params.recipientAddress] The address to send the wrapped token bundle to
 * @param {string} [params.relayerUrl] relayer url (gasless transaction mode)
 * @param {string} [params.clientKey] client key
 * @returns {[key: string]: any} transaction result
 */
lib.wrapTokens = async ({
  network,
  signer,
  privateKey,
  contractAddress,
  contents,
  wrappedTokenMetadata = {},
  recipientAddress,
  relayerUrl,
  clientKey,
  isCount = true,
}) => {
  let caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 6,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    let options = {
      readonlySettings: {
        rpcUrl: getRpcUrl(network),
      },
      ...(isArrayFull(gatewayUrls) ? { gatewayUrls } : {}),
      ...(isNil(thirdwebSecretKey) &&
      (!isNil(thirdwebClientId) || !isNil(thirdwebClientIdFromAPI))
        ? { clientId: thirdwebClientId || thirdwebClientIdFromAPI }
        : {}),
      ...(!isNil(thirdwebSecretKey) ? { secretKey: thirdwebSecretKey } : {}),
    }

    if (!network) throw new Error(messages.NETWORK_UNDEFINED)

    if (!signer && !privateKey)
      throw new Error(messages.SIGNER_PRIVATEKEY_UNDEFINED)

    if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

    if (!contents) throw new Error(messages.CONTENTS_TO_WRAP_UNDEFINED)

    if (relayerUrl) {
      options = {
        ...options,
        gasless: {
          openzeppelin: { relayerUrl },
        },
      }
    }

    let sdk = new ThirdwebSDK(network, options)
    if (signer) {
      sdk.updateSignerOrProvider(signer)
    }

    if (privateKey) {
      sdk = new ThirdwebSDK(
        new ethers.Wallet(
          privateKey,
          ethers.getDefaultProvider(getRpcUrl(network)),
        ),
        options,
      )
    }

    caller = (await sdk?.signer?.getAddress()) || null
    const contract = await sdk.getContract(
      contractAddress,
      ContractType.Multiwrap,
    )
    const tx = await contract.wrap(
      contents,
      wrappedTokenMetadata,
      recipientAddress,
    )

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 6,
      walletAddress: caller,
      status: RequestTransactionStatus.SUCCEEDED,
      transactionHash: formatTransactionHash({ input: tx }),
      isCount,
    })

    return tx
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 6,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 *
 * Get the contents of a wrapped token bundle
 * @function
 * @name getWrappedContents
 * @memberof hazama-lib/Bundle
 *
 * @example
 * import hazlib from hazama-lib
 * ...
 * const network = 'polygon'
 * const contractAddress = '0x...'
 * const tokenId = 0
 * const clientKey = 'dummy'
 *
 * const wrappedTokens = hazlib.getWrappedContents({ network, contractAddress, tokenId, clientKey })
 * ...
 *
 * @param {object} params parameters for getting contents of a wrapped token bundle
 * @param {string} params.network the chain name or rpc url
 * @param {string} params.contractAddress the address of the deployed contract
 * @param {BigNumberish} params.tokenId the id of the wrapped token bundle
 * @param {string} [params.clientKey] client key
 * @returns {WrappedTokensOutput} Output for the contents of a bundle wrapped token
 */
lib.getWrappedContents = async ({
  network,
  contractAddress,
  tokenId,
  clientKey,
  isCount = true,
}) => {
  const caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 2,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (!network) throw new Error(messages.NETWORK_UNDEFINED)

    if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

    if (!tokenId && tokenId !== 0) throw new Error(messages.TOKEN_ID_UNDEFINED)

    const sdk = new ThirdwebSDK(network, {
      ...(isNil(thirdwebSecretKey) &&
      (!isNil(thirdwebClientId) || !isNil(thirdwebClientIdFromAPI))
        ? { clientId: thirdwebClientId || thirdwebClientIdFromAPI }
        : {}),
      ...(!isNil(thirdwebSecretKey) ? { secretKey: thirdwebSecretKey } : {}),
    })

    const contract = await sdk.getContract(
      contractAddress,
      ContractType.Multiwrap,
    )
    const wrappedContents = await contract.getWrappedContents(tokenId)
    const promises = Object.keys(wrappedContents).map(async (key) => {
      let contractMetadata, token
      const tokenContents = wrappedContents[key]
      if (isArrayFull(tokenContents)) {
        const childPromises = tokenContents.map(async (tokenContent) => {
          const { contractAddress: tokenAddress, tokenId } = tokenContent
          const contractType = await sdk.resolveContractType(tokenAddress)
          const contract = await sdk.getContract(tokenAddress, contractType)
          switch (key) {
            case 'erc20Tokens':
              contractMetadata = await contract.metadata.get()
              tokenContent.contractName = contractMetadata?.name || ''
              token = await contract.get()
              tokenContent.tokenName = token?.name || ''
              break
            case 'erc721Tokens':
              contractMetadata = await contract.metadata.get()
              tokenContent.contractName = (await contractMetadata?.name) || ''
              token = await contract.get(tokenId)
              tokenContent.tokenName = token?.metadata?.name || ''
              break
            case 'erc1155Tokens':
              contractMetadata = await contract.metadata.get()
              tokenContent.contractName = (await contractMetadata?.name) || ''
              token = await contract.get(tokenId)
              tokenContent.tokenName = token?.metadata?.name || ''
              break
          }
        })
        return await Promise.all(childPromises)
      }
    })

    const result = await Promise.all(promises)

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 2,
      walletAddress: caller,
      status: RequestTransactionStatus.SUCCEEDED,
      isCount,
    })

    return result
      ? wrappedContents
      : {
          erc20Tokens: [],
          erc721Tokens: [],
          erc1155Tokens: [],
        }
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 2,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * Transfer a wrapped token bundle
 * @function
 * @name transfer
 * @memberof hazama-lib/Bundle
 *
 * @example <caption>Transfer a wrapped token bundle with privateKey</caption>
 * import hazlib from hazama-lib
 * ...
 * const network = 'polygon'
 * const privateKey = 'dummy'
 * const contractAddress = '0x...'
 * const tokenId = 0
 * const clientKey = 'dummy'
 *
 * const toNumber = hazlib.transfer({ network, privateKey, contractAddress, tokenId, clientKey })
 * ...
 *
 * @example <caption>Transfer a wrapped token bundle with signer and relayer</caption>
 * import hazlib from hazama-lib
 * ...
 * const network = 'polygon'
 * const signer = <signer object>
 * const relayerUrl = 'https://api.defender.openzeppelin.com/autotasks/0f7cf308-.../runs/webhook/fd447953-...' // Openzeppelin webhook URI
 * const contractAddress = '0x...'
 * const tokenId = 0
 * const clientKey = 'dummy'
 *
 * const toNumber = hazlib.transfer({ network, signer, contractAddress, tokenId, relayerUrl, clientKey })
 * ...
 *
 * @param {object} params parameters for transferring a wrapped token bundle
 * @param {string} params.network the chain name or rpc url
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions.<br/>If private key is not set, Signer is required.
 * @param {string} [params.privateKey] wallet's private key.<br/>If Signer is not set, private key is required.
 * @param {string} params.contractAddress the address of the deployed contract
 * @param {BigNumberish} params.tokenId the token ID to transfer
 * @param {string} [params.recipientAddress] address to transfer the tokens to
 * @param {string} [params.relayerUrl] relayer url (gasless transaction mode)
 * @param {string} [params.clientKey] client key
 * @returns {TransactionResult} TransactionResult
 */
lib.transfer = async ({
  network,
  signer,
  privateKey,
  contractAddress,
  tokenId,
  recipientAddress,
  relayerUrl,
  clientKey,
  isCount = true,
}) => {
  let caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 4,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    let options = {
      readonlySettings: {
        rpcUrl: getRpcUrl(network),
      },
      ...(isArrayFull(gatewayUrls) ? { gatewayUrls } : {}),
      ...(isNil(thirdwebSecretKey) &&
      (!isNil(thirdwebClientId) || !isNil(thirdwebClientIdFromAPI))
        ? { clientId: thirdwebClientId || thirdwebClientIdFromAPI }
        : {}),
      ...(!isNil(thirdwebSecretKey) ? { secretKey: thirdwebSecretKey } : {}),
    }

    if (!network) throw new Error(messages.NETWORK_UNDEFINED)

    if (!signer && !privateKey)
      throw new Error(messages.SIGNER_PRIVATEKEY_UNDEFINED)

    if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

    if (!tokenId && tokenId !== 0) throw new Error(messages.TOKEN_ID_UNDEFINED)

    if (!recipientAddress) throw new Error(messages.RECIPIENT_ADDRESS_UNDEFINED)

    if (relayerUrl) {
      options = {
        ...options,
        gasless: {
          openzeppelin: { relayerUrl },
        },
      }
    }

    let sdk = new ThirdwebSDK(network, options)
    if (signer) {
      sdk.updateSignerOrProvider(signer)
    }

    if (privateKey) {
      sdk = new ThirdwebSDK(
        new ethers.Wallet(
          privateKey,
          ethers.getDefaultProvider(getRpcUrl(network)),
        ),
        options,
      )
    }

    caller = (await sdk?.signer?.getAddress()) || null
    const contract = await sdk.getContract(
      contractAddress,
      ContractType.Multiwrap,
    )
    const result = await contract.transfer(recipientAddress, tokenId)

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 4,
      walletAddress: caller,
      status: RequestTransactionStatus.SUCCEEDED,
      transactionHash: formatTransactionHash({ input: result }),
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 4,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * Unwrap a wrapped token bundle, and retrieve its contents
 * @function
 * @name unwrap
 * @memberof hazama-lib/Bundle
 *
 * @example <caption>[Unwrap token with privateKey]</caption>
 * import hazlib from hazama-lib
 * ...
 * const network = 'polygon'
 * const privateKey = 'dummy'
 * const contractAddress = '0x...'
 * const tokenId = 0
 * const clientKey = 'dummy'
 *
 * const toNumber = hazlib.unwrap({ network, privateKey, contractAddress, tokenId, clientKey })
 * ...
 *
 * @example <caption>[Unwrap token with signer and relayer]</caption>
 * import hazlib from hazama-lib
 * ...
 * const network = 'polygon'
 * const signer = <signer object>
 * const relayerUrl = 'https://api.defender.openzeppelin.com/autotasks/0f7cf308-.../runs/webhook/fd447953-...' // Openzeppelin webhook URI
 * const contractAddress = '0x...'
 * const tokenId = 0
 * const clientKey = 'dummy'
 *
 * const toNumber = hazlib.unwrap({ network, signer, contractAddress, tokenId, relayerUrl, clientKey })
 * ...
 *
 * @param {object} params parameters for unwrapping a wrapped token bundle
 * @param {string} params.network the chain name or rpc url
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions.<br/>If private key is not set, Signer is required.
 * @param {string} [params.privateKey] wallet's private key.<br/>If Signer is not set, private key is required.
 * @param {string} params.contractAddress the address of the deployed contract
 * @param {BigNumberish} params.tokenId the token ID to unwrap
 * @param {string} [params.recipientAddress] address to transfer the tokens to
 * @param {string} [params.relayerUrl] relayer url (gasless transaction mode)
 * @param {string} [params.clientKey] client key
 * @returns {TransactionResult} TransactionResult
 */
lib.unwrap = async ({
  network,
  signer,
  privateKey,
  contractAddress,
  tokenId,
  recipientAddress,
  relayerUrl,
  clientKey,
  isCount = true,
}) => {
  let caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 5,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    let options = {
      readonlySettings: {
        rpcUrl: getRpcUrl(network),
      },
      ...(isArrayFull(gatewayUrls) ? { gatewayUrls } : {}),
      ...(isNil(thirdwebSecretKey) &&
      (!isNil(thirdwebClientId) || !isNil(thirdwebClientIdFromAPI))
        ? { clientId: thirdwebClientId || thirdwebClientIdFromAPI }
        : {}),
      ...(!isNil(thirdwebSecretKey) ? { secretKey: thirdwebSecretKey } : {}),
    }

    if (!network) throw new Error(messages.NETWORK_UNDEFINED)

    if (!signer && !privateKey)
      throw new Error(messages.SIGNER_PRIVATEKEY_UNDEFINED)

    if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

    if (!tokenId && tokenId !== 0) throw new Error(messages.TOKEN_ID_UNDEFINED)

    if (relayerUrl) {
      options = {
        ...options,
        gasless: {
          openzeppelin: { relayerUrl },
        },
      }
    }

    let sdk = new ThirdwebSDK(network, options)
    if (signer) {
      sdk.updateSignerOrProvider(signer)
    }

    if (privateKey) {
      sdk = new ThirdwebSDK(
        new ethers.Wallet(
          privateKey,
          ethers.getDefaultProvider(getRpcUrl(network)),
        ),
        options,
      )
    }

    caller = (await sdk?.signer?.getAddress()) || null
    const contract = await sdk.getContract(
      contractAddress,
      ContractType.Multiwrap,
    )

    const result = await contract.unwrap(tokenId, recipientAddress)

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 5,
      walletAddress: caller,
      status: RequestTransactionStatus.SUCCEEDED,
      transactionHash: formatTransactionHash({ input: result }),
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 5,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * Set approval for token (ERC20/ERC721/ERC1155)
 * @function
 * @name setApprovalForToken
 * @memberof hazama-lib/Bundle
 *
 * @example <caption>Set approval for token with privateKey</caption>
 * import hazlib from hazama-lib
 * ...
 * const network = 'polygon'
 * const privateKey = 'dummy'
 * const contractAddress = '0x...'
 * const tokenType = 'erc20'
 * const tokenWrappable = {
 *  contractAddress: '0x...',
 *  quantity: '0.8',
 * }
 * const clientKey = 'dummy'
 *
 * const toNumber = hazlib.setApprovalForToken({ network, privateKey, contractAddress, tokenType, tokenWrappable, clientKey })
 * ...
 *
 * @example <caption>Set approval for token with signer and relayer</caption>
 * import hazlib from hazama-lib
 * ...
 * const network = 'polygon'
 * const signer = <signer object>
 * const relayerUrl = 'https://api.defender.openzeppelin.com/autotasks/0f7cf308-.../runs/webhook/fd447953-...' // Openzeppelin webhook URI
 * const contractAddress = '0x...'
 * const tokenType = 'erc20'
 * const tokenWrappable = {
 *  contractAddress: '0x...',
 *  quantity: '0.8',
 * }
 * const clientKey = 'dummy'
 *
 * const toNumber = hazlib.setApprovalForToken({ network, signer, contractAddress, tokenType, tokenWrappable, relayerUrl, clientKey })
 * ...
 *
 * @param {object} params parameters for setting approval for token
 * @param {string} params.network the chain name or rpc url
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions.<br/>If private key is not set, Signer is required.
 * @param {string} [params.privateKey] wallet's private key.<br/>If Signer is not set, private key is required.
 * @param {string} params.contractAddress the operator/spender's address
 * @param {('erc20'|'erc721'|'erc1155')} params.tokenType token type
 * @param {(ERC20Wrappable|ERC721Wrappable|ERC1155Wrappable)} params.tokenWrappable Represents a token/NFT that can be wrapped
 * @param {string} [params.relayerUrl] relayer url (gasless transaction mode)
 * @param {string} [params.clientKey] client key
 * @returns {boolean} result
 */
lib.setApprovalForToken = async ({
  network,
  signer,
  privateKey,
  contractAddress,
  tokenType,
  tokenWrappable,
  relayerUrl,
  clientKey,
  isCount = true,
}) => {
  let caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 3,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    let result = false
    let options = {
      readonlySettings: {
        rpcUrl: getRpcUrl(network),
      },
      ...(isArrayFull(gatewayUrls) ? { gatewayUrls } : {}),
      ...(isNil(thirdwebSecretKey) &&
      (!isNil(thirdwebClientId) || !isNil(thirdwebClientIdFromAPI))
        ? { clientId: thirdwebClientId || thirdwebClientIdFromAPI }
        : {}),
      ...(!isNil(thirdwebSecretKey) ? { secretKey: thirdwebSecretKey } : {}),
    }
    let transactionHash = null

    if (!network) throw new Error(messages.NETWORK_UNDEFINED)

    if (!signer && !privateKey)
      throw new Error(messages.SIGNER_PRIVATEKEY_UNDEFINED)

    if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

    if (!tokenType) throw new Error(messages.TOKEN_TYPE_UNDEFINED)

    if (!tokenWrappable) throw new Error(messages.TOKEN_WRAPPABLE_UNDEFINED)

    if (typeof tokenWrappable.contractAddress === 'undefined')
      throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

    if (
      tokenType === TokenType.ERC20 &&
      typeof tokenWrappable.quantity === 'undefined'
    )
      throw new Error(messages.TOKEN_QUANTITY_UNDEFINED)

    if (
      tokenType === TokenType.ERC721 &&
      typeof tokenWrappable.tokenId === 'undefined'
    )
      throw new Error(messages.TOKEN_ID_UNDEFINED)

    if (relayerUrl) {
      options = {
        ...options,
        gasless: {
          openzeppelin: { relayerUrl },
        },
      }
    }

    let sdk = new ThirdwebSDK(network, options)
    if (signer) {
      sdk.updateSignerOrProvider(signer)
    }

    if (privateKey) {
      sdk = new ThirdwebSDK(
        new ethers.Wallet(
          privateKey,
          ethers.getDefaultProvider(getRpcUrl(network)),
        ),
        options,
      )
    }

    caller = (await sdk?.signer?.getAddress()) || null
    const {
      contractAddress: tokenAddress,
      tokenId,
      quantity,
      approved,
    } = tokenWrappable
    const contractType = await sdk.resolveContractType(tokenAddress)
    const contract = await sdk.getContract(tokenAddress, contractType)
    switch (tokenType) {
      case TokenType.ERC20: {
        const allowance = await contract.allowance(contractAddress)
        if (quantity > (allowance?.displayValue || 0)) {
          const transactionResult = await contract.setAllowance(
            contractAddress,
            quantity,
          )
          transactionHash = transactionResult?.receipt?.transactionHash || null
        }
        result = true
        break
      }
      case TokenType.ERC721: {
        const erc721Owner = await contract.ownerOf(tokenId)
        const erc721ApprovedStatus =
          typeof approved === 'boolean' ? approved : true
        if (erc721ApprovedStatus) {
          const isERC721Approved = await contract.isApproved(
            erc721Owner,
            contractAddress,
          )
          if (!isERC721Approved) {
            const transactionResult = await contract.setApprovalForAll(
              contractAddress,
              true,
            )
            transactionHash =
              transactionResult?.receipt?.transactionHash || null
          }
        } else {
          const transactionResult = await contract.setApprovalForAll(
            contractAddress,
            false,
          )
          transactionHash = transactionResult?.receipt?.transactionHash || null
        }
        result = true
        break
      }
      case TokenType.ERC1155: {
        const erc1155ApprovedStatus =
          typeof approved === 'boolean' ? approved : true
        if (erc1155ApprovedStatus) {
          if (!signer) signer = sdk.signer
          const walletAddress = await signer.getAddress()
          const isERC1155Approved = await contract.isApproved(
            walletAddress,
            contractAddress,
          )
          if (!isERC1155Approved) {
            const transactionResult = await contract.setApprovalForAll(
              contractAddress,
              true,
            )
            transactionHash =
              transactionResult?.receipt?.transactionHash || null
          }
        } else {
          const transactionResult = await contract.setApprovalForAll(
            contractAddress,
            false,
          )
          transactionHash = transactionResult?.receipt?.transactionHash || null
        }
        result = true
        break
      }
    }

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 3,
      walletAddress: caller,
      transactionHash,
      status: RequestTransactionStatus.SUCCEEDED,
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 3,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * Get Wrapped token bundles loop
 * @private
 *
 * @param {object} params parameters for getting contents of a wrapped token bundle
 * @param {Multiwrap} params.contract the deployed contract
 * @param {QueryAllParams} params.queryParams Pagination Parameters
 * @return {Promise<NFTMetadataOwner[]>} promise
 */
function getWrappedBundlesLoop({
  contract,
  queryParams = { start: 0, count: 100 },
}) {
  return contract.getAll(queryParams)
}

/**
 * Get all Wrapped token bundles
 * @function
 * @name getWrappedBundles
 * @memberof hazama-lib/Bundle
 *
 * @example
 * import hazlib from hazama-lib
 * ...
 * const network = 'polygon'
 * const contractAddress = '0x...'
 * const clientKey = 'dummy'
 *
 * const wrappedTokens = hazlib.getWrappedBundles({ network, contractAddress, clientKey })
 * ...
 *
 * @param {object} params parameters for getting contents of a wrapped token bundle
 * @param {string} params.network the chain name or rpc url
 * @param {string} params.contractAddress the address of the deployed contract
 * @param {QueryAllParams} [params.queryParams] Pagination Parameters
 * @param {string} [params.clientKey] client key
 * @returns {NFTMetadataOwner[]} The NFT metadata for all tokens/NFTs queried.
 */
lib.getWrappedBundles = async ({
  network,
  contractAddress,
  queryParams = { start: 0, count: DEFAULT_QUERY_ALL_COUNT },
  clientKey,
  isCount = true,
}) => {
  const caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 1,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (!network) throw new Error(messages.NETWORK_UNDEFINED)

    if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

    const sdk = new ThirdwebSDK(network, {
      ...(isNil(thirdwebSecretKey) &&
      (!isNil(thirdwebClientId) || !isNil(thirdwebClientIdFromAPI))
        ? { clientId: thirdwebClientId || thirdwebClientIdFromAPI }
        : {}),
      ...(!isNil(thirdwebSecretKey) ? { secretKey: thirdwebSecretKey } : {}),
    })

    const contract = await sdk.getContract(
      contractAddress,
      ContractType.Multiwrap,
    )

    let wrappedBundlesAll = []

    let newQueryParams = { ...queryParams }

    while (true) {
      const wrappedBundles = await getWrappedBundlesLoop({
        contract,
        queryParams: newQueryParams,
      })
      wrappedBundlesAll = wrappedBundlesAll.concat(
        wrappedBundles.map((el) => Object.assign({}, el)),
      )

      if (wrappedBundles.length < DEFAULT_QUERY_ALL_COUNT) break

      newQueryParams = {
        ...newQueryParams,
        start: newQueryParams.start + DEFAULT_QUERY_ALL_COUNT,
      }
    }

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 1,
      walletAddress: caller,
      status: RequestTransactionStatus.SUCCEEDED,
      isCount,
    })

    return wrappedBundlesAll
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 1,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * Compile and broadcast
 *
 * @private
 * @param {object} params parameters for compiling and broadcasting
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {string} params.asm asm
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions
 * @param {string} [params.wif] wallet import format
 * @returns {Promise<string>} promise
 */
async function compileAndBroadcast({ host = HOST, asm, signer, wif }) {
  // eslint-disable-next-line no-async-promise-executor
  return new Promise(async (resolve, reject) => {
    try {
      if (!/^#asm/.test(asm)) {
        asm = '#asm\n' + asm
      }

      if (!signer && !wif) throw new Error(messages.SIGNER_WIF_UNDEFINED)

      const code = lib.compile(asm)
      const envelope =
        typeof wif !== 'undefined'
          ? lib.envelope(code, wif)
          : await lib.envelopeWithWallet(code, signer, null, false)

      lib.connect(host, async () => {
        try {
          const txid = await lib.tx(envelope, null, false)
          await lib.waitFinishedTx(lib, txid)
          resolve(txid)
        } catch (e) {
          reject(e)
        }
      })
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Deploy contract on EVM network
 *
 * @private
 * @param {object} params parameters for deploying contract
 * @param {string} params.network the chain name or rpc url
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions.<br/>If private key is not set, Signer is required.
 * @param {string} [params.privateKey] wallet's private key.<br/>If Signer is not set, private key is required.
 * @param {('nft-collection'|'nft-drop'|'edition-drop'|'token'|'multiwrap')} params.contractType the contract type
 * @param {(NFTContractDeployMetadata|TokenContractDeployMetadata|BundleContractDeployMetadata)} params.contractMetadata contract metadata
 * @param {string} [params.relayerUrl] relayer url (gasless transaction mode)
 * @return {Promise<string>} promise
 */
async function deployContractOnEVMNetwork({
  network,
  signer,
  privateKey,
  contractType,
  contractMetadata,
  relayerUrl,
}) {
  // eslint-disable-next-line no-async-promise-executor
  return new Promise(async (resolve, reject) => {
    try {
      let result = null
      let options = {
        readonlySettings: {
          rpcUrl: getRpcUrl(network),
        },
        ...(isArrayFull(gatewayUrls) ? { gatewayUrls } : {}),
        ...(isNil(thirdwebSecretKey) &&
        (!isNil(thirdwebClientId) || !isNil(thirdwebClientIdFromAPI))
          ? { clientId: thirdwebClientId || thirdwebClientIdFromAPI }
          : {}),
        ...(!isNil(thirdwebSecretKey) ? { secretKey: thirdwebSecretKey } : {}),
      }

      if (!network) throw new Error(messages.NETWORK_UNDEFINED)

      if (!signer && !privateKey)
        throw new Error(messages.SIGNER_PRIVATEKEY_UNDEFINED)

      if (!contractType) throw new Error(messages.CONTRACT_TYPE_UNDEFINED)

      if (
        ![
          ContractType.TokenERC721,
          ContractType.DropERC721,
          ContractType.DropERC1155,
          ContractType.TokenERC20,
          ContractType.Multiwrap,
          CustomContract.ContractType.BurnToRedeem1155,
          CustomContract.ContractType.Vote,
        ].includes(contractType)
      )
        throw new Error(messages.CONTRACT_TYPE_IS_INVALID)

      if (!contractMetadata)
        throw new Error(messages.CONTRACT_METADATA_UNDEFINED)

      if (!contractMetadata.name)
        throw new Error(messages.CONTRACT_NAME_UNDEFINED)

      if (
        contractType !== ContractType.Multiwrap &&
        contractType !== CustomContract.ContractType.Vote
      ) {
        if (!contractMetadata.primary_sale_recipient)
          throw new Error(messages.PRIMARY_SALE_RECIPIENT_UNDEFINED)
      }

      if (relayerUrl) {
        options = {
          ...options,
          gasless: {
            openzeppelin: { relayerUrl },
          },
        }
      }

      let sdk = new ThirdwebSDK(network, options)
      if (signer) {
        sdk.updateSignerOrProvider(signer)
      }

      if (privateKey) {
        sdk = new ThirdwebSDK(
          new ethers.Wallet(
            privateKey,
            ethers.getDefaultProvider(getRpcUrl(network)),
          ),
          options,
        )
      }

      switch (contractType) {
        case ContractType.Multiwrap:
        case ContractType.TokenERC20:
        case ContractType.TokenERC721:
        case ContractType.DropERC721:
        case ContractType.DropERC1155:
          result = await deployViaFactory({
            network,
            sdk,
            signer,
            privateKey,
            contractType,
            contractMetadata,
            relayerUrl,
            version: getContractVersion({ contractType }),
          })
          break
        case CustomContract.ContractType.BurnToRedeem1155:
        case CustomContract.ContractType.Vote:
          result = await deployReleasedContract({
            network,
            sdk,
            signer,
            privateKey,
            contractType,
            contractMetadata,
            relayerUrl,
            version: getContractVersion({ contractType }),
          })
          break
        default:
          throw new Error(messages.CONTRACT_TYPE_IS_NOT_SUPPORTED)
      }
      resolve(result)
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Deploy contract on HAZAMA network
 *
 * @private
 * @param {object} params parameters for deploying contract
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions.<br/>If wif is not set, Signer is required.
 * @param {string} [params.wif] wallet import format.<br/>If Signer is not set, wif is required.
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {('nft_extensible'|'nft_nontransferable'|'fungible_token')} params.contractTemplate the contract template
 * @param {HazamaTokenContractDeployMetadata} params.contractMetadata contract metadata
 * @return {Promise<string>} promise
 */
async function deployContractOnHazamaNetwork({
  signer,
  wif,
  host = HOST,
  contractTemplate,
  contractMetadata,
}) {
  // eslint-disable-next-line no-async-promise-executor
  return new Promise(async (resolve, reject) => {
    try {
      if (!signer && !wif) throw new Error(messages.SIGNER_WIF_UNDEFINED)

      if (!host) throw new Error(messages.HOST_UNDEFINED)

      if (!contractTemplate)
        throw new Error(messages.CONTRACT_TEMPLATE_UNDEFINED)

      if (
        !['nft_extensible', 'nft_nontransferable', 'fungible_token'].includes(
          contractTemplate,
        )
      )
        throw new Error(messages.CONTRACT_TEMPLATE_IS_INVALID)

      if (!contractMetadata)
        throw new Error(messages.CONTRACT_METADATA_UNDEFINED)

      if (!contractMetadata.name)
        throw new Error(messages.CONTRACT_NAME_UNDEFINED)

      if (!contractMetadata.primarySaleRecipient)
        throw new Error(messages.PRIMARY_SALE_RECIPIENT_UNDEFINED)

      if (!contractMetadata.path)
        throw new Error(messages.CONTRACT_PATH_UNDEFINED)

      if (!contractMetadata.author)
        throw new Error(messages.CONTRACT_AUTHOR_UNDEFINED)

      if (!contractMetadata.tokenVersion)
        throw new Error(messages.TOKEN_VERSION_UNDEFINED)

      if (!contractMetadata.decimals)
        throw new Error(messages.TOKEN_DECIMALS_UNDEFINED)

      lib.connect(host, async () => {
        try {
          const asm = await lib.template(
            contractTemplate,
            contractMetadata,
            null,
            false,
          )
          await compileAndBroadcast({ host, asm, signer, wif })

          const data = await lib.get(
            `${contractMetadata.path}.meta`,
            null,
            false,
          )
          resolve(data.ContractHash)
        } catch (e) {
          reject(e)
        }
      })
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Deploy contract for token (ERC20/ERC721/ERC1155)
 * @function
 * @name deployContract
 * @memberof hazama-lib/Others
 *
 * @example <caption>[Deploy contract with privateKey on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const privateKey = 'dummy'
 * const contractAddress = '0x...'
 * const contractType = 'nft-collection'
 * const contractMetadata = {
 *  name: 'contract name',
 *  primary_sale_recipient: '0x...',
 *  ...
 * }
 * const clientKey = 'dummy'
 *
 * const contract = hazlib.deployContract({ network, privateKey, contractType, contractMetadata, clientKey })
 * ...
 *
 * @example <caption>[Deploy contract with signer and relayer on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const signer = <signer object>
 * const relayerUrl = 'https://api.defender.openzeppelin.com/autotasks/0f7cf308-.../runs/webhook/fd447953-...' // Openzeppelin webhook URI
 * const contractAddress = '0x...'
 * const contractType = 'nft-collection'
 * const contractMetadata = {
 *  name: 'contract name',
 *  primary_sale_recipient: '0x...',
 *  ...
 * }
 * const clientKey = 'dummy'
 *
 * const contract = hazlib.deployContract({ network, signer, contractType, contractMetadata, relayerUrl, clientKey })
 * ...
 *
 * @example <caption>[Deploy contract on HAZAMA network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'hazama'
 * const wif = 'dummny'
 * const host = 'wss://n.hazbase.com'
 * const contractTemplate = 'nft_extensible'
 * const contractMetadata = {
 *  path: '/hazama/...',
 *  name: 'contract name',
 *  primarySaleRecipient: '0x...',
 *  author: 'author name',
 *  tokenVersion: '0.0.1',
 *  decimals: 18,
 *  ...
 * }
 * const clientKey = 'dummy'
 *
 * const contract = hazlib.deployContract({ network, wif, host, contractTemplate, contractMetadata, clientKey })
 * ...
 *
 * @param {object} params parameters for deploying contract
 * @param {string} params.network the chain name or rpc url
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions
 * @param {string} [params.privateKey] wallet's private key
 * @param {string} [params.wif] wallet import format
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {('nft-collection'|'nft-drop'|'edition-drop'|'token'|'multiwrap')} [params.contractType] the contract type
 * @param {('nft_extensible'|'nft_nontransferable'|'fungible_token')} [params.contractTemplate] the contract template (use by Hazama network)
 * @param {(NFTContractDeployMetadata|TokenContractDeployMetadata|BundleContractDeployMetadata|HazamaTokenContractDeployMetadata)} params.contractMetadata contract metadata
 * @param {string} [params.relayerUrl] relayer url (gasless transaction mode)
 * @param {string} [params.clientKey] client key
 * @return {string} contract id or transaction id (Hazama)
 */
lib.deployContract = async ({
  network,
  signer,
  privateKey,
  wif,
  host = HOST,
  contractType,
  contractTemplate,
  contractMetadata,
  relayerUrl,
  clientKey,
  isCount = true,
}) => {
  let caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 12,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (!network) throw new Error(messages.NETWORK_UNDEFINED)

    caller = await getCaller({ network, host, signer, privateKey, wif })
    const isHazamaNetwork = network === 'hazama'

    const result = !isHazamaNetwork
      ? await deployContractOnEVMNetwork({
          network,
          signer,
          privateKey,
          contractType,
          contractMetadata,
          relayerUrl,
        })
      : await deployContractOnHazamaNetwork({
          signer,
          wif,
          host,
          contractTemplate,
          contractMetadata,
        })

    let payload = {
      chainId: getChainId(network),
      contractAddress: result,
      walletAddress: caller,
    }

    if (isHazamaNetwork) {
      const contractInfo = getContractInfoFromPath(contractMetadata?.path || '')
      payload = {
        ...payload,
        contractId: contractInfo?.contractId || null,
        projectId: contractInfo?.projectId || null,
      }
    }

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 12,
      walletAddress: caller,
      status: RequestTransactionStatus.SUCCEEDED,
      transactionHash: network === 'hazama' ? result : null,
      apiRequests: [
        {
          apiCallStatus: ApiRequestStatus.NOT_CALLED_YET,
          apiBody: JSON.stringify(payload),
          apiURI: '/api/app/contracts',
          apiMethod: 'post',
        },
      ],
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 12,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * Mint token on EVM network
 *
 * @private
 * @param {object} params parameters for minting token
 * @param {string} params.network the chain name or rpc url
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions.<br/>If private key is not set, Signer is required.
 * @param {string} [params.privateKey] wallet's private key.<br/>If Signer is not set, private key is required.
 * @param {string} params.contractAddress the token contract address
 * @param {string} params.recipientAddress address to transfer the tokens to
 * @param {Amount} params.quantity number of tokens to transfer
 * @param {(NFTMetadataOrUri[]|NFTMetadataOrUri)} [params.metadata] token metadata
 * @param {string} [params.relayerUrl] relayer url (gasless transaction mode)
 * @return {Promise<TransactionResult|TransactionResultWithId[]>} promise
 */
async function mintTokenOnEVMNetwork({
  network,
  signer,
  privateKey,
  contractAddress,
  recipientAddress,
  quantity,
  metadata,
  relayerUrl,
}) {
  // eslint-disable-next-line no-async-promise-executor
  return new Promise(async (resolve, reject) => {
    try {
      let result = null
      let options = {
        readonlySettings: {
          rpcUrl: getRpcUrl(network),
        },
        ...(isArrayFull(gatewayUrls) ? { gatewayUrls } : {}),
        ...(isNil(thirdwebSecretKey) &&
        (!isNil(thirdwebClientId) || !isNil(thirdwebClientIdFromAPI))
          ? { clientId: thirdwebClientId || thirdwebClientIdFromAPI }
          : {}),
        ...(!isNil(thirdwebSecretKey) ? { secretKey: thirdwebSecretKey } : {}),
      }

      if (!network) throw new Error(messages.NETWORK_UNDEFINED)

      if (!signer && !privateKey)
        throw new Error(messages.SIGNER_PRIVATEKEY_UNDEFINED)

      if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

      if (relayerUrl) {
        options = {
          ...options,
          gasless: {
            openzeppelin: { relayerUrl },
          },
        }
      }

      let sdk = new ThirdwebSDK(network, options)
      if (signer) {
        sdk.updateSignerOrProvider(signer)
      }

      if (privateKey) {
        sdk = new ThirdwebSDK(
          new ethers.Wallet(
            privateKey,
            ethers.getDefaultProvider(getRpcUrl(network)),
          ),
          options,
        )
      }

      let contractType = await resolveContractTypeWithCustom({
        sdk,
        contractAddress,
      })

      if (contractType === ContractType.TokenERC1155) {
        // BurnToRedeem/ to use lazyMint
        contractType = ContractType.DropERC1155
      }

      if ([ContractType.TokenERC20].includes(contractType) && !recipientAddress)
        throw new Error(messages.RECIPIENT_ADDRESS_UNDEFINED)

      if (contractType === ContractType.TokenERC20 && !quantity)
        throw new Error(messages.TOKEN_QUANTITY_UNDEFINED)

      if (
        [
          ContractType.DropERC1155,
          ContractType.DropERC721,
          ContractType.TokenERC721,
        ].includes(contractType) &&
        !metadata
      )
        throw new Error(messages.METADATA_UNDEFINED)

      const contract = await sdk.getContract(contractAddress, contractType)
      switch (contractType) {
        case ContractType.TokenERC20:
          result = await contract.mintTo(recipientAddress, quantity)
          break
        case ContractType.DropERC1155:
        case ContractType.DropERC721:
          result = await contract.createBatch(
            isArray(metadata) ? metadata : [metadata],
          )
          break
        case ContractType.TokenERC721:
          result = await contract.mintBatch(
            isArray(metadata) ? metadata : [metadata],
          )
          break
        default:
          throw new Error(messages.TOKEN_TYPE_IS_NOT_SUPPORTED)
      }
      resolve(result)
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Mint token on HAZAMA network
 *
 * @private
 * @param {object} params parameters for minting token
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions.<br/>If wif is not set, Signer is required.
 * @param {string} [params.wif] wallet import format.<br/>If Signer is not set, wif is required.
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {('fungible_token_issuance'|'nft_extensible_mint'|'nft_multi_extensible_mint')} params.tokenTemplate the token template
 * @param {(HazamaFungibleTokenMintMetadata|HazamaNFTMintMetadata|HazamaMultiNFTMintMetadata)} params.metadata token metadata
 * @return {Promise<string>} promise
 */
async function mintTokenOnHazamaNetwork({
  signer,
  wif,
  host = HOST,
  tokenTemplate,
  metadata,
}) {
  // eslint-disable-next-line no-async-promise-executor
  return new Promise(async (resolve, reject) => {
    try {
      if (!signer && !wif) throw new Error(messages.SIGNER_WIF_UNDEFINED)

      if (!host) throw new Error(messages.HOST_UNDEFINED)

      if (!tokenTemplate) throw new Error(messages.TOKEN_TEMPLATE_UNDEFINED)

      if (
        ![
          'fungible_token_issuance',
          'nft_extensible_mint',
          'nft_multi_extensible_mint',
        ].includes(tokenTemplate)
      )
        throw new Error(messages.TOKEN_TEMPLATE_IS_INVALID)

      if (!metadata) throw new Error(messages.METADATA_UNDEFINED)

      if (!metadata.path) throw new Error(messages.TOKEN_PATH_UNDEFINED)

      if (tokenTemplate === 'fungible_token_issuance') {
        if (!metadata.amount) throw new Error(messages.TOKEN_QUANTITY_UNDEFINED)

        if (!metadata.mintto)
          throw new Error(messages.RECIPIENT_ADDRESS_UNDEFINED)
      } else if (tokenTemplate === 'nft_extensible_mint') {
        if (!metadata.name) throw new Error(messages.TOKEN_NAME_UNDEFINED)
      } else if (tokenTemplate === 'nft_multi_extensible_mint') {
        if (!metadata.mints || (metadata.mints && !isArrayFull(metadata.mints)))
          throw new Error(messages.MINTS_UNDEFINED)

        if (metadata.mints.some((el) => typeof el.name === 'undefined'))
          throw new Error(messages.TOKEN_NAME_UNDEFINED)
      }

      lib.connect(host, async () => {
        try {
          const asm = await lib.template(tokenTemplate, metadata, null, false)
          const result = await compileAndBroadcast({ host, asm, signer, wif })
          resolve(result)
        } catch (e) {
          reject(e)
        }
      })
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Mint token (ERC20/ERC721/ERC1155)
 * @function
 * @name mint
 * @memberof hazama-lib/Others
 * 
 * @example <caption>[Mint ERC20 with privateKey on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const privateKey = 'dummy'
 * const contractAddress = '0x...'
 * const recipientAddress = '0x...'
 * const quantity = '0.8'
 * const clientKey = 'dummy'
 * 
 * const result = hazlib.mint({ network, privateKey, contractAddress, recipientAddress, quantity, clientKey })
 * ...
 * 
 * @example <caption>[Mint ERC20 with signer and relayer on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const signer = <signer object>
 * const relayerUrl = 'https://api.defender.openzeppelin.com/autotasks/0f7cf308-.../runs/webhook/fd447953-...' // Openzeppelin webhook URI
 * const contractAddress = '0x...'
 * const recipientAddress = '0x...'
 * const quantity = '0.8'
 * const clientKey = 'dummy'
 * 
 * const result = hazlib.mint({ network, signer, contractAddress, recipientAddress, quantity, relayerUrl, clientKey })
 * ...
 * 
 * @example <caption>[Mint ERC721 with privateKey on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const privateKey = 'dummy'
 * const contractAddress = '0x...'
 * const recipientAddress = '0x...'
 * const metadata = [{
 *  name: 'token name',
 *  description: 'token description',
 *  ...
 * }]
 * const clientKey = 'dummy'
 * 
 * const result = hazlib.mint({ network, privateKey, contractAddress, recipientAddress, metadata, clientKey })
 * ...
 * 
 * @example <caption>[Mint ERC721 with signer and relayer on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const signer = <signer object>
 * const relayerUrl = 'https://api.defender.openzeppelin.com/autotasks/0f7cf308-.../runs/webhook/fd447953-...' // Openzeppelin webhook URI
 * const contractAddress = '0x...'
 * const recipientAddress = '0x...'
 * const metadata = [{
 *  name: 'token name',
 *  description: 'token description',
 *  ...
 * }]
 * const clientKey = 'dummy'
 * 
 * const result = hazlib.mint({ network, signer, contractAddress, recipientAddress, metadata, relayerUrl, clientKey })
 * ...
 * 
 * @example <caption>[Mint DropERC721 with privateKey on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const privateKey = 'dummy'
 * const contractAddress = '0x...'
 * const metadata = [{
 *  name: 'token name',
 *  description: 'token description',
 *  ...
 * }]
 * const clientKey = 'dummy'
 * 
 * const result = hazlib.mint({ network, privateKey, contractAddress, metadata, clientKey })
 * ...
 * 
 * @example <caption>[Mint DropERC721 with signer and relayer on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const signer = <signer object>
 * const relayerUrl = 'https://api.defender.openzeppelin.com/autotasks/0f7cf308-.../runs/webhook/fd447953-...' // Openzeppelin webhook URI
 * const contractAddress = '0x...'
 * const metadata = [{
 *  name: 'token name',
 *  description: 'token description',
 *  ...
 * }]
 * const clientKey = 'dummy'
 * 
 * const result = hazlib.mint({ network, signer, contractAddress, metadata, relayerUrl, clientKey })
 * ...
 * 
 * @example <caption>[Mint DropERC1155 with privateKey on EVM network]</caption>
 * import hazlib from 'hazama-lib'

 * ...
 * const network = 'polygon'
 * const privateKey = 'dummy'
 * const contractAddress = '0x...'
 * const metadata = [{
 *  name: 'token name',
 *  description: 'token description',
 *  ...
 * }]
 * const clientKey = 'dummy'
 * 
 * const result = hazlib.mint({ network, privateKey, contractAddress, metadata, clientKey })
 * ...
 * 
 * @example <caption>[Mint DropERC1155 with signer and relayer on EVM network]</caption>
 * import hazlib from 'hazama-lib'

 * ...
 * const network = 'polygon'
 * const signer = <signer object>
 * const relayerUrl = 'https://api.defender.openzeppelin.com/autotasks/0f7cf308-.../runs/webhook/fd447953-...' // Openzeppelin webhook URI
 * const contractAddress = '0x...'
 * const metadata = [{
 *  name: 'token name',
 *  description: 'token description',
 *  ...
 * }]
 * const clientKey = 'dummy'
 * 
 * const result = hazlib.mint({ network, signer, contractAddress, metadata, relayerUrl, clientKey })
 * ...
 * 
 * @example <caption>[Mint fungible token on HAZAMA network]</caption>
 * import hazlib from 'hazama-lib'

 * ...
 * const network = 'hazama'
 * const wif = 'dummny'
 * const host = 'wss://n.hazbase.com'
 * const tokenTemplate = 'fungible_token_issuance'
 * const metadata = {
 *  path: '/hazama/...',
 *  amount: '2000000000000000000',
 *  mintto: '0x...'
 * }
 * const clientKey = 'dummy'
 * 
 * const result = hazlib.mint({ network, wif, host, tokenTemplate, metadata, clientKey })
 * ...
 * 
 * @example <caption>[Mint NFT on HAZAMA network]</caption>
 * import hazlib from 'hazama-lib'

 * ...
 * const network = 'hazama'
 * const wif = 'dummny'
 * const host = 'wss://n.hazbase.com'
 * const tokenTemplate = 'nft_extensible_mint'
 * const metadata = {
 *  path: '/hazama/...',
 *  name: 'token name',
 *  description: 'token description',
 *  ...
 * }
 * const clientKey = 'dummy'
 * 
 * const result = hazlib.mint({ network, wif, host, tokenTemplate, metadata, clientKey })
 * ...
 * 
 * @example <caption>[Mint multiple NFTs on HAZAMA network]</caption>
 * import hazlib from 'hazama-lib'

 * ...
 * const network = 'hazama'
 * const wif = 'dummny'
 * const host = 'wss://n.hazbase.com'
 * const tokenTemplate = 'nft_multi_extensible_mint'
 * const metadata = {
 *  path: '/hazama/...',
 *  mints: [
 *   {
 *    name: 'token name 1',
 *    description: 'token description',
 *    ...
 *   },
 *   {
 *    name: 'token name 2',
 *    description: 'token description',
 *    ...
 *   }
 *  ]
 * }
 * const clientKey = 'dummy'
 * 
 * const result = hazlib.mint({ network, wif, host, tokenTemplate, metadata, clientKey })
 * ...
 * 
 * @param {object} params parameters for minting token
 * @param {string} params.network the chain name or rpc url
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions
 * @param {string} [params.privateKey] wallet's private key
 * @param {string} [params.wif] wallet import format
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {string} [params.contractAddress] the token contract address
 * @param {('fungible_token_issuance'|'nft_extensible_mint'|'nft_multi_extensible_mint')} [params.tokenTemplate] the token template (use by Hazama network)
 * @param {string} [params.recipientAddress] address to transfer the tokens to
 * @param {Amount} [params.quantity] number of tokens to transfer
 * @param {(NFTMetadataOrUri[]|NFTMetadataOrUri|HazamaFungibleTokenMintMetadata|HazamaNFTMintMetadata|HazamaMultiNFTMintMetadata)} [params.metadata] token metadata
 * @param {string} [params.relayerUrl] relayer url (gasless transaction mode)
 * @param {string} [params.clientKey] client key
 * @return {(TransactionResult|TransactionResultWithId[]|string)} transaction id (Hazama) or transaction result (other networks)
 */
lib.mint = async ({
  network,
  signer,
  privateKey,
  wif,
  host = HOST,
  contractAddress,
  tokenTemplate,
  recipientAddress,
  quantity,
  metadata,
  relayerUrl,
  clientKey,
  isCount = true,
}) => {
  let caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 24,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (!network) throw new Error(messages.NETWORK_UNDEFINED)

    caller = await getCaller({ network, host, signer, privateKey, wif })
    const isHazamaNetwork = network === 'hazama'

    const result = !isHazamaNetwork
      ? await mintTokenOnEVMNetwork({
          network,
          signer,
          privateKey,
          contractAddress,
          recipientAddress,
          quantity,
          metadata,
          relayerUrl,
        })
      : await mintTokenOnHazamaNetwork({
          signer,
          wif,
          host,
          tokenTemplate,
          metadata,
        })

    const apiRequests = []
    if (!isHazamaNetwork) {
      const sdk = new ThirdwebSDK(network, {
        ...(isNil(thirdwebSecretKey) &&
        (!isNil(thirdwebClientId) || !isNil(thirdwebClientIdFromAPI))
          ? { clientId: thirdwebClientId || thirdwebClientIdFromAPI }
          : {}),
        ...(!isNil(thirdwebSecretKey) ? { secretKey: thirdwebSecretKey } : {}),
      })
      const contractType = await sdk.resolveContractType(contractAddress)
      if (
        [ContractType.DropERC1155, ContractType.DropERC721].includes(
          contractType,
        ) &&
        isArrayFull(result)
      ) {
        const tokens = result.map((token) => {
          return {
            chainId: getChainId(network),
            tokenId: token.id.toString(),
            contractAddress,
            walletAddress: caller,
          }
        })
        apiRequests.push({
          apiCallStatus: ApiRequestStatus.NOT_CALLED_YET,
          apiBody: JSON.stringify({ tokens }),
          apiURI: '/api/app/token/save-multiple-tokens',
          apiMethod: 'post',
        })
      }
    }

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 24,
      walletAddress: caller,
      status: RequestTransactionStatus.SUCCEEDED,
      transactionHash: formatTransactionHash({ input: result }),
      ...(apiRequests?.length ? { apiRequests } : {}),
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 24,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * Create proposal on EVM Network
 *
 * @private
 * @param {object} params parameters for creating proposal
 * @param {string} params.network the chain name or rpc url
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions.<br/>If private key is not set, Signer is required.
 * @param {string} [params.privateKey] wallet's private key.<br/>If Signer is not set, private key is required.
 * @param {string} params.contractAddress the token contract address
 * @param {string} params.description description
 * @param {number} params.starttime start time
 * @param {number} params.endtime end time
 * @param {ProposalExecution[]} params.executions executions
 * @param {string[]} [params.choices] choices
 * @param {string} [params.relayerUrl] relayer url (gasless transaction mode)
 * @return {Promise<TransactionResult>} promise
 */
async function proposalOnEVMNetwork({
  network,
  signer,
  privateKey,
  contractAddress,
  description,
  starttime,
  endtime,
  executions,
  choices,
  relayerUrl,
}) {
  // eslint-disable-next-line no-async-promise-executor
  return new Promise(async (resolve, reject) => {
    try {
      let result = null
      let options = {
        readonlySettings: {
          rpcUrl: getRpcUrl(network),
        },
        ...(isArrayFull(gatewayUrls) ? { gatewayUrls } : {}),
        ...(isNil(thirdwebSecretKey) &&
        (!isNil(thirdwebClientId) || !isNil(thirdwebClientIdFromAPI))
          ? { clientId: thirdwebClientId || thirdwebClientIdFromAPI }
          : {}),
        ...(!isNil(thirdwebSecretKey) ? { secretKey: thirdwebSecretKey } : {}),
      }

      if (!network) throw new Error(messages.NETWORK_UNDEFINED)

      if (!signer && !privateKey)
        throw new Error(messages.SIGNER_PRIVATEKEY_UNDEFINED)

      if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

      if (!description) throw new Error(messages.DESCRIPTION_UNDEFINED)

      if (!starttime) throw new Error(messages.START_TIME_UNDEFINED)

      if (!endtime) throw new Error(messages.END_TIME_UNDEFINED)

      if (starttime > endtime) throw new Error(messages.TIME_INVALID)

      if (!isArrayFull(executions))
        throw new Error(messages.EXECUTIONS_UNDEFINED)

      if (relayerUrl) {
        options = {
          ...options,
          gasless: {
            openzeppelin: { relayerUrl },
          },
        }
      }

      let sdk = new ThirdwebSDK(network, options)
      if (signer) {
        sdk.updateSignerOrProvider(signer)
      }

      if (privateKey) {
        sdk = new ThirdwebSDK(
          new ethers.Wallet(
            privateKey,
            ethers.getDefaultProvider(getRpcUrl(network)),
          ),
          options,
        )
      }

      const contractType = await resolveContractTypeWithCustom({
        sdk,
        contractAddress,
      })
      const contract = await sdk.getContract(contractAddress, contractType)

      const tos = executions.map((p) => p.toAddress)
      const values = executions.map((p) => p.nativeTokenValue)
      const datas = executions.map((p) => p.transactionData)

      result = await contract.call('propose', [
        tos,
        values,
        datas,
        description,
        starttime,
        endtime,
        choices || [],
      ])
      const event = contract.contractWrapper.parseLogs(
        'ProposalCreated',
        result?.receipt?.logs,
      )
      const proposalId = event?.[0]?.args?.proposalId?.toString() || null
      const state = proposalId
        ? await contract.call('state', [proposalId])
        : null

      resolve({
        id: proposalId,
        state,
        receipt: result?.receipt || null,
      })
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Create proposal
 * @function
 * @name proposal
 * @memberof hazama-lib/Others
 *
 * @example <caption>[Create proposal on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const privateKey = 'dummy'
 * const contractAddress = '0x...'
 * const description = 'dummy'
 * const starttime = 1680200000
 * const endtime = 1680250000
 * const executions = []
 * const clientKey = 'dummy'
 *
 * const result = hazlib.proposal({ network, privateKey, contractAddress, description, starttime, endtime, executions, clientKey })
 * ...
 *
 * @example <caption>[Create proposal with signer and relayer on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const signer = <signer object>
 * const relayerUrl = 'https://api.defender.openzeppelin.com/autotasks/0f7cf308-.../runs/webhook/fd447953-...' // Openzeppelin webhook URI
 * const contractAddress = '0x...'
 * const description = 'dummy'
 * const starttime = 1680200000
 * const endtime = 1680250000
 * const executions = []
 * const clientKey = 'dummy'
 *
 * const result = hazlib.proposal({ network, signer, contractAddress, description, starttime, endtime, executions, relayerUrl, clientKey })
 * ...
 *
 * @param {object} params parameters for creating proposal
 * @param {string} params.network the chain name or rpc url
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions
 * @param {string} [params.privateKey] wallet's private key
 * @param {string} [params.wif] wallet import format
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {string} [params.contractAddress] the token contract address
 * @param {string} params.description description
 * @param {number} params.starttime start time
 * @param {number} params.endtime end time
 * @param {ProposalExecution[]} params.executions executions
 * @param {string[]} [params.choices] choices
 * @param {string} [params.relayerUrl] relayer url (gasless transaction mode)
 * @param {string} [params.clientKey] client key
 * @return {TransactionResult}
 */
lib.proposal = async ({
  network,
  signer,
  privateKey,
  wif,
  host = HOST,
  contractAddress,
  tokenTemplate,
  description,
  starttime,
  endtime,
  executions,
  choices,
  relayerUrl,
  clientKey,
  isCount = true,
}) => {
  let caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 55,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (!network) throw new Error(messages.NETWORK_UNDEFINED)

    caller = await getCaller({ network, host, signer, privateKey, wif })
    const isHazamaNetwork = network === 'hazama'

    const result = !isHazamaNetwork
      ? await proposalOnEVMNetwork({
          network,
          signer,
          privateKey,
          contractAddress,
          description,
          starttime,
          endtime,
          executions,
          relayerUrl,
          choices,
        })
      : null // await proposalOnHazamaNetwork({signer, wif, host, tokenTemplate, metadata})

    const payload = {
      chainId: getChainId(network),
      contractAddress,
      walletAddress: caller,
      proposalId: result?.id || null,
    }

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 55,
      walletAddress: caller,
      status: RequestTransactionStatus.SUCCEEDED,
      transactionHash: formatTransactionHash({ input: result }),
      apiRequests: [
        {
          apiCallStatus: ApiRequestStatus.NOT_CALLED_YET,
          apiBody: JSON.stringify({ proposals: [payload] }),
          apiURI: '/api/app/proposal/save-multiple-proposals',
          apiMethod: 'post',
        },
      ],
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 55,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * Transfer token on EVM network
 *
 * @private
 * @param {object} params parameters for transfering token
 * @param {string} params.network the chain name or rpc url
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions
 * @param {string} [params.privateKey] wallet's private key
 * @param {string} params.contractAddress the token contract address
 * @param {string} params.recipientAddress address to transfer the tokens to
 * @param {BigNumberish} [params.tokenId] token id
 * @param {Amount} [params.quantity] number of tokens to transfer
 * @param {string} [params.relayerUrl] relayer url (gasless transaction mode)
 * @return {Promise<TransactionResult>} promise
 */
async function transferTokenOnEVMNetwork({
  network,
  signer,
  privateKey,
  contractAddress,
  recipientAddress,
  tokenId,
  quantity,
  relayerUrl,
}) {
  // eslint-disable-next-line no-async-promise-executor
  return new Promise(async (resolve, reject) => {
    try {
      let result = null
      let options = {
        readonlySettings: {
          rpcUrl: getRpcUrl(network),
        },
        ...(isArrayFull(gatewayUrls) ? { gatewayUrls } : {}),
        ...(isNil(thirdwebSecretKey) &&
        (!isNil(thirdwebClientId) || !isNil(thirdwebClientIdFromAPI))
          ? { clientId: thirdwebClientId || thirdwebClientIdFromAPI }
          : {}),
        ...(!isNil(thirdwebSecretKey) ? { secretKey: thirdwebSecretKey } : {}),
      }

      if (!network) throw new Error(messages.NETWORK_UNDEFINED)

      if (!signer && !privateKey)
        throw new Error(messages.SIGNER_PRIVATEKEY_UNDEFINED)

      if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

      if (!recipientAddress)
        throw new Error(messages.RECIPIENT_ADDRESS_UNDEFINED)

      if (relayerUrl) {
        options = {
          ...options,
          gasless: {
            openzeppelin: { relayerUrl },
          },
        }
      }

      let sdk = new ThirdwebSDK(network, options)
      if (signer) {
        sdk.updateSignerOrProvider(signer)
      }

      if (privateKey) {
        sdk = new ThirdwebSDK(
          new ethers.Wallet(
            privateKey,
            ethers.getDefaultProvider(getRpcUrl(network)),
          ),
          options,
        )
      }

      const contractType = await resolveContractTypeWithCustom({
        sdk,
        contractAddress,
      })

      if (
        ![
          ContractType.TokenERC721,
          ContractType.DropERC721,
          ContractType.DropERC1155,
          ContractType.TokenERC20,
          ContractType.Multiwrap,
          ContractType.TokenERC1155,
        ].includes(contractType)
      )
        throw new Error(messages.CONTRACT_TYPE_IS_INVALID)

      if (
        [
          ContractType.TokenERC721,
          ContractType.DropERC721,
          ContractType.Multiwrap,
        ].includes(contractType) &&
        !tokenId &&
        tokenId !== 0
      )
        throw new Error(messages.TOKEN_ID_UNDEFINED)

      if (contractType === ContractType.TokenERC20 && !quantity)
        throw new Error(messages.TOKEN_QUANTITY_UNDEFINED)

      if (
        (contractType === ContractType.DropERC1155 ||
          contractType === ContractType.TokenERC1155) &&
        (!quantity || (!tokenId && tokenId !== 0))
      )
        throw new Error(messages.BOTH_TOKEN_ID_AND_QUANTITY_ARE_REQUIRED)

      const contract = await sdk.getContract(contractAddress, contractType)
      switch (contractType) {
        case ContractType.TokenERC721:
          result = await contract.transfer(recipientAddress, tokenId)
          break
        case ContractType.DropERC721:
          result = await contract.transfer(recipientAddress, tokenId)
          break
        case ContractType.TokenERC20:
          result = await contract.transfer(recipientAddress, quantity)
          break
        case ContractType.DropERC1155:
        case ContractType.TokenERC1155:
          result = await contract.transfer(recipientAddress, tokenId, quantity)
          break
        case ContractType.Multiwrap:
          result = await contract.transfer(recipientAddress, tokenId)
          break
        default:
          throw new Error(messages.TOKEN_TYPE_IS_NOT_SUPPORTED)
      }
      resolve(result)
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Transfer token on HAZAMA network
 *
 * @private
 * @param {object} params parameters for transfering token
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions
 * @param {string} [params.wif] wallet import format
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {('fungible_token_send'|'nft_send')} params.tokenTemplate the token template (use by Hazama network)
 * @param {HazamaTransferTokenMetadata} params.metadata
 * @return {Promise<string>} promise
 */
async function transferTokenOnHazamaNetwork({
  signer,
  wif,
  host = HOST,
  tokenTemplate,
  metadata,
}) {
  return new Promise(async (resolve, reject) => {
    try {
      if (!signer && !wif) throw new Error(messages.SIGNER_WIF_UNDEFINED)

      if (!host) throw new Error(messages.HOST_UNDEFINED)

      if (!tokenTemplate) throw new Error(messages.TOKEN_TEMPLATE_UNDEFINED)

      if (!['fungible_token_send', 'nft_send'].includes(tokenTemplate))
        throw new Error(messages.TOKEN_TEMPLATE_IS_INVALID)

      if (!metadata) throw new Error(messages.METADATA_UNDEFINED)

      if (!metadata.path) throw new Error(messages.TOKEN_PATH_UNDEFINED)

      if (!metadata.destination)
        throw new Error(messages.RECIPIENT_ADDRESS_UNDEFINED)

      if (tokenTemplate === 'fungible_token_send') {
        if (!metadata.amount) throw new Error(messages.TOKEN_QUANTITY_UNDEFINED)
      } else if (tokenTemplate === 'nft_send') {
        if (!metadata.name) throw new Error(messages.NFT_ID_QUANTITY_UNDEFINED)
      }

      lib.connect(host, async () => {
        try {
          const asm = await lib.template(tokenTemplate, metadata, null, false)
          const result = await compileAndBroadcast({ host, asm, signer, wif })
          resolve(result)
        } catch (e) {
          reject(e)
        }
      })
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Transfer token (ERC20/ERC721/ERC1155)
 * @function
 * @name transferToken
 * @memberof hazama-lib/Others
 *
 * @example <caption>[Transfer ERC20 with privateKey on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const privateKey = 'dummy'
 * const contractAddress = '0x...'
 * const recipientAddress = '0x...'
 * const quantity = '0.8'
 * const clientKey = 'dummy'
 *
 * const result = hazlib.transferToken({ network, privateKey, contractAddress, recipientAddress, quantity, clientKey })
 * ...
 *
 * @example <caption>[Transfer ERC20 with signer and relayer on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const signer = <signer object>
 * const relayerUrl = 'https://api.defender.openzeppelin.com/autotasks/0f7cf308-.../runs/webhook/fd447953-...' // Openzeppelin webhook URI
 * const contractAddress = '0x...'
 * const recipientAddress = '0x...'
 * const quantity = '0.8'
 * const clientKey = 'dummy'
 *
 * const result = hazlib.transferToken({ network, signer, contractAddress, recipientAddress, quantity, relayerUrl, clientKey })
 * ...
 *
 * @example <caption>[Transfer ERC721 with privateKey on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const privateKey = 'dummy'
 * const contractAddress = '0x...'
 * const recipientAddress = '0x...'
 * const tokenId = 1
 * const clientKey = 'dummy'
 *
 * const result = hazlib.transferToken({ network, privateKey, contractAddress, recipientAddress, tokenId, clientKey })
 * ...
 *
 * @example <caption>[Transfer ERC721 with signer and relayer on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const signer = <signer object>
 * const relayerUrl = 'https://api.defender.openzeppelin.com/autotasks/0f7cf308-.../runs/webhook/fd447953-...' // Openzeppelin webhook URI
 * const contractAddress = '0x...'
 * const recipientAddress = '0x...'
 * const tokenId = 1
 * const clientKey = 'dummy'
 *
 * const result = hazlib.transferToken({ network, signer, contractAddress, recipientAddress, tokenId, relayerUrl, clientKey })
 * ...
 *
 * @example <caption>[Transfer DropERC721 with privateKey on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const privateKey = 'dummy'
 * const contractAddress = '0x...'
 * const recipientAddress = '0x...'
 * const tokenId = 1
 * const clientKey = 'dummy'
 *
 * const result = hazlib.transferToken({ network, privateKey, contractAddress, recipientAddress, tokenId, clientKey })
 * ...
 *
 * @example <caption>[Transfer DropERC721 with signer and relayer on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const signer = <signer object>
 * const relayerUrl = 'https://api.defender.openzeppelin.com/autotasks/0f7cf308-.../runs/webhook/fd447953-...' // Openzeppelin webhook URI
 * const contractAddress = '0x...'
 * const recipientAddress = '0x...'
 * const tokenId = 1
 * const clientKey = 'dummy'
 *
 * const result = hazlib.transferToken({ network, signer, contractAddress, recipientAddress, tokenId, relayerUrl, clientKey })
 * ...
 *
 * @example <caption>[Transfer DropERC1155 with privateKey on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const privateKey = 'dummy'
 * const contractAddress = '0x...'
 * const recipientAddress = '0x...'
 * const tokenId = 1
 * const quantity = '0.8'
 * const clientKey = 'dummy'
 *
 * const result = hazlib.transferToken({ network, privateKey, contractAddress, recipientAddress, tokenId, quantity, clientKey })
 * ...
 *
 * @example <caption>[Transfer DropERC1155 with signer and relayer on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const signer = <signer object>
 * const relayerUrl = 'https://api.defender.openzeppelin.com/autotasks/0f7cf308-.../runs/webhook/fd447953-...' // Openzeppelin webhook URI
 * const contractAddress = '0x...'
 * const recipientAddress = '0x...'
 * const tokenId = 1
 * const quantity = '0.8'
 * const clientKey = 'dummy'
 *
 * const result = hazlib.transferToken({ network, signer, contractAddress, recipientAddress, tokenId, quantity, relayerUrl, clientKey })
 * ...
 *
 * @example <caption>[Transfer Multiwrap with privateKey on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const privateKey = 'dummy'
 * const contractAddress = '0x...'
 * const recipientAddress = '0x...'
 * const tokenId = 1
 * const clientKey = 'dummy'
 *
 * const result = hazlib.transferToken({ network, privateKey, contractAddress, recipientAddress, tokenId, clientKey })
 * ...
 *
 * @example <caption>[Transfer Multiwrap with signer and relayer on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const signer = <signer object>
 * const relayerUrl = 'https://api.defender.openzeppelin.com/autotasks/0f7cf308-.../runs/webhook/fd447953-...' // Openzeppelin webhook URI
 * const contractAddress = '0x...'
 * const recipientAddress = '0x...'
 * const tokenId = 1
 * const clientKey = 'dummy'
 *
 * const result = hazlib.transferToken({ network, signer, contractAddress, recipientAddress, tokenId, relayerUrl, clientKey })
 * ...
 *
 * @example <caption>[Transfer fungible token on HAZAMA network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'hazama'
 * const wif = 'dummy'
 * const host = 'wss://n.hazbase.com'
 * const tokenTemplate = 'fungible_token_send'
 * const metadata = {
 *  path: '/hazama/...'
 *  amount: '2000000000000000000',
 *  destination: '0x...',
 *  ...
 * }
 * const clientKey = 'dummy'
 *
 * const result = hazlib.transferToken({ network, wif, host, tokenTemplate, metadata, clientKey })
 * ...
 *
 * @example <caption>[Transfer NFT on HAZAMA network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'hazama'
 * const wif = 'dummy'
 * const host = 'wss://n.hazbase.com'
 * const tokenTemplate = 'nft_send'
 * const metadata = {
 *  path: '/hazama/...'
 *  name: '2000000000000000000',  // idOrAmount
 *  destination: '0x...',
 *  ...
 * }
 * const clientKey = 'dummy'
 *
 * const result = hazlib.transferToken({ network, wif, host, tokenTemplate, metadata, clientKey })
 * ...
 *
 * @param {object} params parameters for transfering token
 * @param {string} params.network the chain name or rpc url
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions
 * @param {string} [params.privateKey] wallet's private key
 * @param {string} [params.wif] wallet import format
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {string} [params.contractAddress] the token contract address
 * @param {('fungible_token_send'|'nft_send')} [params.tokenTemplate] the token template (use by Hazama network)
 * @param {string} [params.recipientAddress] address to transfer the tokens to
 * @param {BigNumberish} [params.tokenId] token id
 * @param {Amount} [params.quantity] number of tokens to transfer
 * @param {HazamaTransferTokenMetadata} [params.metadata] token metadata
 * @param {string} [params.relayerUrl] relayer url (gasless transaction mode)
 * @param {string} [params.clientKey] client key
 * @return {TransactionResult|string} transaction id (Hazama) or transaction result
 */
lib.transferToken = async ({
  network,
  signer,
  privateKey,
  wif,
  host = HOST,
  contractAddress,
  tokenTemplate,
  recipientAddress,
  tokenId,
  quantity,
  metadata,
  relayerUrl,
  clientKey,
  isCount = true,
}) => {
  let caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 29,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (!network) throw new Error(messages.NETWORK_UNDEFINED)

    caller = await getCaller({ network, host, signer, privateKey, wif })

    const result =
      network !== 'hazama'
        ? await transferTokenOnEVMNetwork({
            network,
            signer,
            privateKey,
            contractAddress,
            recipientAddress,
            tokenId,
            quantity,
            relayerUrl,
          })
        : await transferTokenOnHazamaNetwork({
            signer,
            wif,
            host,
            tokenTemplate,
            metadata,
          })

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 29,
      walletAddress: caller,
      status: RequestTransactionStatus.SUCCEEDED,
      transactionHash: formatTransactionHash({ input: result }),
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 29,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * Set claim conditions on EVM Network
 *
 * @private
 * @param {object} params parameters for setting claim conditions
 * @param {string} params.network the chain name or rpc url
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions.<br/>If private key is not set, Signer is required.
 * @param {string} [params.privateKey] wallet's private key.<br/>If Signer is not set, private key is required.
 * @param {string} params.contractAddress the token contract address
 * @param {BigNumberish} [params.tokenId] token id
 * @param {ClaimConditionInput[]} params.claimConditions claim conditions
 * @param {string} [params.relayerUrl] relayer url (gasless transaction mode)
 * @return {Promise<TransactionResult>} promise
 */
async function setClaimConditionsOnEVMNetwork({
  network,
  signer,
  privateKey,
  contractAddress,
  tokenId,
  claimConditions,
  relayerUrl,
}) {
  return new Promise(async (resolve, reject) => {
    try {
      let result = null
      let options = {
        readonlySettings: {
          rpcUrl: getRpcUrl(network),
        },
        ...(isArrayFull(gatewayUrls) ? { gatewayUrls } : {}),
        ...(isNil(thirdwebSecretKey) &&
        (!isNil(thirdwebClientId) || !isNil(thirdwebClientIdFromAPI))
          ? { clientId: thirdwebClientId || thirdwebClientIdFromAPI }
          : {}),
        ...(!isNil(thirdwebSecretKey) ? { secretKey: thirdwebSecretKey } : {}),
      }

      if (!network) throw new Error(messages.NETWORK_UNDEFINED)

      if (!signer && !privateKey)
        throw new Error(messages.SIGNER_PRIVATEKEY_UNDEFINED)

      if (!claimConditions) throw new Error(messages.CLAIM_CONDITIONS_UNDEFINED)

      if (relayerUrl) {
        options = {
          ...options,
          gasless: {
            openzeppelin: { relayerUrl },
          },
        }
      }

      let sdk = new ThirdwebSDK(network, options)
      if (signer) {
        sdk.updateSignerOrProvider(signer)
      }

      if (privateKey) {
        sdk = new ThirdwebSDK(
          new ethers.Wallet(
            privateKey,
            ethers.getDefaultProvider(getRpcUrl(network)),
          ),
          options,
        )
      }

      const contractType = await sdk.resolveContractType(contractAddress)

      if (
        ![ContractType.DropERC721, ContractType.DropERC1155].includes(
          contractType,
        )
      )
        throw new Error(messages.CONTRACT_TYPE_IS_INVALID)

      if (
        contractType === ContractType.DropERC1155 &&
        !tokenId &&
        tokenId !== 0
      )
        throw new Error(messages.TOKEN_ID_UNDEFINED)

      const contract = await sdk.getContract(contractAddress, contractType)
      switch (contractType) {
        case ContractType.DropERC1155:
          result = await contract.claimConditions.set(tokenId, claimConditions)
          break
        case ContractType.DropERC721:
          result = await contract.claimConditions.set(claimConditions)
          break
        default:
          throw new Error(messages.TOKEN_TYPE_IS_NOT_SUPPORTED)
      }
      resolve(result)
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Set claim conditions on Hazama Network
 *
 * @private
 * @param {object} params parameters for setting claim conditions
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions.<br/>If wif is not set, Signer is required.
 * @param {string} [params.wif] wallet import format.<br/>If Signer is not set, wif is required.
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {('nft_set_claim_conditions')} params.tokenTemplate the token template
 * @param {HazamaClaimConditionInput} params.claimConditions claim conditions
 * @return {Promise<string>} promise
 */
async function setClaimConditionsOnHazamaNetwork({
  signer,
  wif,
  host = HOST,
  tokenTemplate,
  claimConditions,
}) {
  return new Promise(async (resolve, reject) => {
    try {
      if (!signer && !wif) throw new Error(messages.SIGNER_WIF_UNDEFINED)

      if (!host) throw new Error(messages.HOST_UNDEFINED)

      if (!tokenTemplate) throw new Error(messages.TOKEN_TEMPLATE_UNDEFINED)

      if (!['nft_set_claim_conditions'].includes(tokenTemplate))
        throw new Error(messages.TOKEN_TEMPLATE_IS_INVALID)

      if (!claimConditions) throw new Error(messages.CLAIM_CONDITIONS_UNDEFINED)

      if (!claimConditions.path)
        throw new Error(messages.CONTRACT_PATH_UNDEFINED)

      if (!claimConditions.token && claimConditions.token !== 0)
        throw new Error(messages.TOKEN_ID_UNDEFINED)

      const permissions = []
      const existsSnap = await iterate({
        lib,
        path: `${claimConditions.path}/${claimConditions.token}/permissions/claim/`,
        match: '0x[0-9a-zA-Z]+',
      })

      const revokeSnaps = {}
      existsSnap?.forEach((snap) => {
        const address = snap[0].split('/').pop()
        revokeSnaps[address] = true
      })
      if (claimConditions.snapshot?.length) {
        claimConditions.snapshot.forEach((permit) => {
          permissions.push({
            token: claimConditions.token,
            address: permit.address,
            maxClaimable:
              permit.maxClaimable == null ? 0 : Number(permit.maxClaimable),
            permission: 'grant',
          })
          delete revokeSnaps[permit.address]
        })
      }
      for (const address in revokeSnaps) {
        permissions.push({
          token: claimConditions.token,
          address,
          maxClaimable: 0,
          permission: 'revoke',
        })
      }
      const permissionAsm = await lib.template(
        'nft_set_claim_permissions',
        {
          path: claimConditions.path,
          permissions,
        },
        null,
        false,
      )

      lib.connect(host, async () => {
        try {
          const asm = await lib.template(
            tokenTemplate,
            claimConditions,
            null,
            false,
          )
          const result = await compileAndBroadcast({
            host,
            asm: jointAsm(asm, permissionAsm),
            signer,
            wif,
          })
          resolve(result)
        } catch (e) {
          reject(e)
        }
      })
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Set claim conditions for DropERC721/DropERC1155
 * @function
 * @name setClaimConditions
 * @memberof hazama-lib/Others
 *
 * @example <caption>[Set claim conditions for DropERC721 with privateKey on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const privateKey = 'dummy'
 * const contractAddress = '0x...'
 * const claimConditions = [{
 *  startTime: new Date(),
 *  maxQuantity: 2,
 *  ...
 * }]
 * const clientKey = 'dummy'
 *
 * const result = hazlib.setClaimConditions({ network, privateKey, contractAddress, claimConditions, clientKey })
 * ...
 *
 * @example <caption>[Set claim conditions for DropERC721 with signer and relayer on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const signer = <signer object>
 * const relayerUrl = 'https://api.defender.openzeppelin.com/autotasks/0f7cf308-.../runs/webhook/fd447953-...' // Openzeppelin webhook URI
 * const contractAddress = '0x...'
 * const claimConditions = [{
 *  startTime: new Date(),
 *  maxQuantity: 2,
 *  ...
 * }]
 * const clientKey = 'dummy'
 *
 * const result = hazlib.setClaimConditions({ network, signer, contractAddress, claimConditions, relayerUrl, clientKey })
 * ...
 *
 * @example <caption>[Set claim conditions for DropERC1155 with privateKey on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const privateKey = 'dummy'
 * const contractAddress = '0x...'
 * const tokenId = 1
 * const claimConditions = [{
 *  startTime: new Date(),
 *  maxQuantity: 2,
 *  ...
 * }]
 * const clientKey = 'dummy'
 *
 * const result = hazlib.setClaimConditions({ network, privateKey, contractAddress, tokenId, claimConditions, clientKey })
 * ...
 *
 * @example <caption>[Set claim conditions for DropERC1155 with signer and relayer on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const signer = <signer object>
 * const relayerUrl = 'https://api.defender.openzeppelin.com/autotasks/0f7cf308-.../runs/webhook/fd447953-...' // Openzeppelin webhook URI
 * const contractAddress = '0x...'
 * const tokenId = 1
 * const claimConditions = [{
 *  startTime: new Date(),
 *  maxQuantity: 2,
 *  ...
 * }]
 * const clientKey = 'dummy'
 *
 * const result = hazlib.setClaimConditions({ network, signer, contractAddress, tokenId, claimConditions, relayerUrl, clientKey })
 * ...
 *
 * @example <caption>[Set claim conditions for NFT on HAZAMA network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'hazama'
 * const wif = 'dummy'
 * const host = 'wss://n.hazbase.com'
 * const tokenTemplate = 'nft_set_claim_conditions'
 * const claimConditions = {
 *  path: '/hazama/...',
 *  token: 1,
 *  startTime: new Date(),
 *  maxQuantity: 1,
 *  ...
 * }
 * const clientKey = 'dummy'
 *
 * const result = hazlib.setClaimConditions({ network, wif, host, tokenTemplate, claimConditions, clientKey })
 * ...
 *
 * @param {object} params parameters for setting claim conditions
 * @param {string} params.network the chain name or rpc url
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions
 * @param {string} [params.privateKey] wallet's private key
 * @param {string} [params.wif] wallet import format
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {string} [params.contractAddress] the token contract address
 * @param {('nft_set_claim_conditions')} [params.tokenTemplate] the token template (use by Hazama network)
 * @param {BigNumberish} [params.tokenId] token id
 * @param {ClaimConditionInput[]|HazamaClaimConditionInput} params.claimConditions claim conditions
 * @param {string} [params.relayerUrl] relayer url (gasless transaction mode)
 * @param {string} [params.clientKey] client key
 * @param {string} [params.paymentMode] payment mode
 * @param {boolean} [params.sellByCash] sell by cash
 * @return {TransactionResult|string} transaction id (Hazama) or transaction result
 */
lib.setClaimConditions = async ({
  network,
  signer,
  privateKey,
  wif,
  host = HOST,
  contractAddress,
  tokenTemplate,
  tokenId,
  claimConditions,
  relayerUrl,
  clientKey,
  sellByCash,
  paymentMode,
  isCount = true,
}) => {
  let caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 28,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (!network) throw new Error(messages.NETWORK_UNDEFINED)

    caller = await getCaller({ network, host, signer, privateKey, wif })
    const isHazamaNetwork = network === 'hazama'

    const result = !isHazamaNetwork
      ? await setClaimConditionsOnEVMNetwork({
          network,
          signer,
          privateKey,
          contractAddress,
          tokenId,
          claimConditions,
          relayerUrl,
        })
      : await setClaimConditionsOnHazamaNetwork({
          signer,
          wif,
          host,
          tokenTemplate,
          claimConditions,
        })

    let payload = {
      chainId: getChainId(network),
      tokenId,
      contractAddress,
      walletAddress: caller,
      price: isArray(claimConditions)
        ? claimConditions?.[0]?.price || 0
        : claimConditions?.price || 0,
      ...(typeof sellByCash !== 'undefined' ? { sellByCash } : {}),
      ...(paymentMode ? { paymentMode } : {}),
    }

    if (isHazamaNetwork) {
      const contractInfo = getContractInfoFromPath(claimConditions?.path || '')
      payload = {
        ...payload,
        tokenId: claimConditions?.token || null,
        contractId: contractInfo?.contractId || null,
        projectId: contractInfo?.projectId || null,
      }
    } else {
      const provider = ethers.getDefaultProvider(getRpcUrl(network))
      const currencyMetadata = claimConditions?.[0]?.currencyAddress
        ? await fetchCurrencyMetadata(
            provider,
            claimConditions?.[0]?.currencyAddress,
          )
        : null
      if (currencyMetadata) {
        payload = {
          ...payload,
          currency: currencyMetadata?.symbol || '',
        }
      }
    }

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 28,
      walletAddress: caller,
      status: RequestTransactionStatus.SUCCEEDED,
      transactionHash: formatTransactionHash({ input: result }),
      apiRequests: [
        {
          apiCallStatus: ApiRequestStatus.NOT_CALLED_YET,
          apiBody: JSON.stringify(payload),
          apiURI: '/api/app/token/save-token-details',
          apiMethod: 'post',
        },
      ],
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 28,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * Set access key tokens on EVM Network
 *
 * @private
 * @param {object} params parameters for setting access key tokens
 * @param {string} params.network the chain name or rpc url
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions.<br/>If private key is not set, Signer is required.
 * @param {string} [params.privateKey] wallet's private key.<br/>If Signer is not set, private key is required.
 * @param {string} params.contractAddress the token contract address
 * @param {BigNumberish} [params.tokenId] token id
 * @param {string} params.burnContractAddress the burn contract address
 * @param {any} params.burnTokens
 * @param {string} [params.relayerUrl] relayer url (gasless transaction mode)
 * @return {Promise<any>} promise
 */
async function setAccessKeyTokensOnEVMNetwork({
  network,
  signer,
  privateKey,
  contractAddress,
  tokenId,
  burnContractAddress,
  burnTokens,
  relayerUrl,
}) {
  return new Promise(async (resolve, reject) => {
    try {
      let result = null
      let options = {
        readonlySettings: {
          rpcUrl: getRpcUrl(network),
        },
        ...(isArrayFull(gatewayUrls) ? { gatewayUrls } : {}),
        ...(isNil(thirdwebSecretKey) &&
        (!isNil(thirdwebClientId) || !isNil(thirdwebClientIdFromAPI))
          ? { clientId: thirdwebClientId || thirdwebClientIdFromAPI }
          : {}),
        ...(!isNil(thirdwebSecretKey) ? { secretKey: thirdwebSecretKey } : {}),
      }

      if (!network) throw new Error(messages.NETWORK_UNDEFINED)

      if (!signer && !privateKey)
        throw new Error(messages.SIGNER_PRIVATEKEY_UNDEFINED)

      if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

      if (!burnContractAddress)
        throw new Error(messages.CLAIM_CONDITIONS_UNDEFINED)

      if (!burnTokens) throw new Error(messages.CLAIM_CONDITIONS_UNDEFINED)

      if (relayerUrl) {
        options = {
          ...options,
          gasless: {
            openzeppelin: { relayerUrl },
          },
        }
      }

      let sdk = new ThirdwebSDK(network, options)
      if (signer) {
        sdk.updateSignerOrProvider(signer)
      }

      if (privateKey) {
        sdk = new ThirdwebSDK(
          new ethers.Wallet(
            privateKey,
            ethers.getDefaultProvider(getRpcUrl(network)),
          ),
          options,
        )
      }

      const contractTypeOfBurnContractAddress = await sdk.resolveContractType(
        burnContractAddress,
      )
      if (contractTypeOfBurnContractAddress !== ContractType.DropERC1155)
        throw new Error(messages.CONTRACT_TYPE_IS_INVALID)

      const contractType = await sdk.resolveContractType(contractAddress)
      if (
        ![ContractType.TokenERC1155, ContractType.Custom].includes(contractType)
      )
        throw new Error(messages.CONTRACT_TYPE_IS_INVALID)

      if (
        contractType === ContractType.TokenERC1155 &&
        !tokenId &&
        tokenId !== 0
      )
        throw new Error(messages.TOKEN_ID_UNDEFINED)

      const contract = await sdk.getContract(contractAddress, contractType)
      switch (contractType) {
        case ContractType.TokenERC1155:
        case ContractType.Custom:
          result = await contract.call('setAccessKeyTokens', [
            burnContractAddress,
            burnTokens,
          ])
          break
        default:
          throw new Error(messages.TOKEN_TYPE_IS_NOT_SUPPORTED)
      }
      resolve(result)
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Set access key tokens
 * @function
 * @name setAccessKeyTokens
 * @memberof hazama-lib/Others
 *
 * @example <caption>[Set access key tokens on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const privateKey = 'dummy'
 * const contractAddress = '0x...'
 * const tokenId = 1
 * const burnContractAddress = '0x...'
 * const burnTokens = [{
 *  tokenId: 1,
 *  quantity: 1
 * }]
 * const clientKey = 'dummy'
 *
 * const result = hazlib.setAccessKeyTokens({ network, privateKey, contractAddress, tokenId, burnContractAddress, burnTokens, clientKey })
 * ...
 *
 * @example <caption>[Set access key tokens with signer and relayer on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const signer = <signer object>
 * const relayerUrl = 'https://api.defender.openzeppelin.com/autotasks/0f7cf308-.../runs/webhook/fd447953-...' // Openzeppelin webhook URI
 * const contractAddress = '0x...'
 * const tokenId = 1
 * const burnContractAddress = '0x...'
 * const burnTokens = [{
 *  tokenId: 1,
 *  quantity: 1
 * }]
 * const clientKey = 'dummy'
 *
 * const result = hazlib.setAccessKeyTokens({ network, signer, contractAddress, tokenId, burnContractAddress, burnTokens, relayerUrl, clientKey })
 * ...
 *
 * @param {object} params parameters for setting access key tokens
 * @param {string} params.network the chain name or rpc url
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions
 * @param {string} [params.privateKey] wallet's private key
 * @param {string} [params.wif] wallet import format
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {string} [params.contractAddress] the token contract address
 * @param {BigNumberish} [params.tokenId] token id
 * @param {string} params.burnContractAddress the burn contract address
 * @param {any} params.burnTokens
 * @param {string} [params.relayerUrl] relayer url (gasless transaction mode)
 * @param {string} [params.clientKey] client key
 * @return {any}
 */
lib.setAccessKeyTokens = async ({
  network,
  signer,
  privateKey,
  wif,
  host = HOST,
  contractAddress,
  tokenId,
  burnContractAddress,
  burnTokens,
  relayerUrl,
  clientKey,
  isCount = true,
}) => {
  let caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 28,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (!network) throw new Error(messages.NETWORK_UNDEFINED)

    if (network === 'hazama') throw new Error(messages.NETWORK_IS_NOT_SUPPORTED)

    caller = await getCaller({ network, host, signer, privateKey, wif })

    const result = await setAccessKeyTokensOnEVMNetwork({
      network,
      signer,
      privateKey,
      contractAddress,
      tokenId,
      burnContractAddress,
      burnTokens,
      relayerUrl,
    })

    const payload = {
      chainId: getChainId(network),
      tokenId,
      contractAddress,
      walletAddress: caller,
      burnContractAddress,
      burnTokens,
    }

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 28,
      walletAddress: caller,
      status: RequestTransactionStatus.SUCCEEDED,
      transactionHash: formatTransactionHash({ input: result }),
      apiRequests: [
        {
          apiCallStatus: ApiRequestStatus.NOT_CALLED_YET,
          apiBody: JSON.stringify(payload),
          apiURI: '/api/app/token/save-token-details',
          apiMethod: 'post',
        },
      ],
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 28,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * Claim on EVM Network
 *
 * @private
 * @param {object} params parameters for claiming
 * @param {string} params.network the chain name or rpc url
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions.<br/>If private key is not set, Signer is required.
 * @param {string} [params.privateKey] wallet's private key.<br/>If Signer is not set, private key is required.
 * @param {string} params.contractAddress the token contract address
 * @param {string} params.recipientAddress address to transfer the tokens to
 * @param {BigNumberish} [params.tokenId] token id
 * @param {Amount} params.quantity number of tokens to transfer
 * @param {string} [params.relayerUrl] relayer url (gasless transaction mode)
 * @return {Promise<TransactionResult|TransactionResultWithId[]>} promise
 */
async function claimOnEVMNetwork({
  network,
  signer,
  privateKey,
  contractAddress,
  recipientAddress,
  tokenId,
  quantity,
  relayerUrl,
}) {
  return new Promise(async (resolve, reject) => {
    try {
      let result = null
      let options = {
        readonlySettings: {
          rpcUrl: getRpcUrl(network),
        },
        ...(isArrayFull(gatewayUrls) ? { gatewayUrls } : {}),
        ...(isNil(thirdwebSecretKey) &&
        (!isNil(thirdwebClientId) || !isNil(thirdwebClientIdFromAPI))
          ? { clientId: thirdwebClientId || thirdwebClientIdFromAPI }
          : {}),
        ...(!isNil(thirdwebSecretKey) ? { secretKey: thirdwebSecretKey } : {}),
      }

      if (!network) throw new Error(messages.NETWORK_UNDEFINED)

      if (!signer && !privateKey)
        throw new Error(messages.SIGNER_PRIVATEKEY_UNDEFINED)

      if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

      if (!recipientAddress)
        throw new Error(messages.RECIPIENT_ADDRESS_UNDEFINED)

      if (!quantity) throw new Error(messages.TOKEN_QUANTITY_UNDEFINED)

      if (relayerUrl) {
        options = {
          ...options,
          gasless: {
            openzeppelin: { relayerUrl },
          },
        }
      }

      let sdk = new ThirdwebSDK(network, options)
      if (signer) {
        sdk.updateSignerOrProvider(signer)
      }

      if (privateKey) {
        sdk = new ThirdwebSDK(
          new ethers.Wallet(
            privateKey,
            ethers.getDefaultProvider(getRpcUrl(network)),
          ),
          options,
        )
      }

      const contractType = await resolveContractTypeWithCustom({
        sdk,
        contractAddress,
      })

      if (
        ![
          ContractType.DropERC1155,
          ContractType.DropERC721,
          ContractType.TokenERC1155,
        ].includes(contractType)
      )
        throw new Error(messages.CONTRACT_TYPE_IS_INVALID)

      if (
        (contractType === ContractType.DropERC1155 ||
          contractType === ContractType.TokenERC1155) &&
        !tokenId &&
        tokenId !== 0
      )
        throw new Error(messages.TOKEN_ID_UNDEFINED)

      const contract = await sdk.getContract(contractAddress, contractType)
      switch (contractType) {
        case ContractType.DropERC721:
          result = await contract.claimTo(recipientAddress, quantity)
          break
        case ContractType.DropERC1155:
          result = await contract.claimTo(recipientAddress, tokenId, quantity)
          break
        case ContractType.TokenERC1155:
          try {
            const [burnContractAddress] = await contract.call(
              'currentAccessKeyTokens',
            )
            const burnContract = await sdk.getContract(
              burnContractAddress,
              'edition-drop',
            )

            const walletAddress = await signer.getAddress()
            const isERC1155Approved = await burnContract.isApproved(
              walletAddress,
              contractAddress,
            )

            if (!isERC1155Approved) {
              await burnContract.setApprovalForAll(contractAddress, true)
            }
          } catch (e) {}

          result = await contract.call('claim', [
            recipientAddress,
            tokenId,
            quantity,
          ])
          break
        default:
          throw new Error(messages.TOKEN_TYPE_IS_NOT_SUPPORTED)
      }
      resolve(result)
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Claim on Hazama Network
 *
 * @private
 * @param {object} params parameters for claiming
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions.<br/>If wif is not set, Signer is required.
 * @param {string} [params.wif] wallet import format.<br/>If Signer is not set, wif is required.
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {('nft_claimto')} params.tokenTemplate the token template (use by Hazama network)
 * @param {HazamaClaimMetadata} params.metadata claim metadata
 * @return {Promise<string>} transaction id (Hazama) or transaction result
 */
async function claimOnHazamaNetwork({
  signer,
  wif,
  host = HOST,
  tokenTemplate,
  metadata,
}) {
  return new Promise(async (resolve, reject) => {
    try {
      if (!signer && !wif) throw new Error(messages.SIGNER_WIF_UNDEFINED)

      if (!host) throw new Error(messages.HOST_UNDEFINED)

      if (!tokenTemplate) throw new Error(messages.TOKEN_TEMPLATE_UNDEFINED)

      if (!['nft_claimto'].includes(tokenTemplate))
        throw new Error(messages.TOKEN_TEMPLATE_IS_INVALID)

      if (!metadata) throw new Error(messages.METADATA_UNDEFINED)

      if (!metadata.path) throw new Error(messages.TOKEN_PATH_UNDEFINED)

      if (!metadata.address)
        throw new Error(messages.RECIPIENT_ADDRESS_UNDEFINED)

      if (!metadata.amount) throw new Error(messages.TOKEN_QUANTITY_UNDEFINED)

      if (!metadata.token && metadata.token !== 0)
        throw new Error(messages.TOKEN_ID_UNDEFINED)

      lib.connect(host, async () => {
        try {
          const asm = await lib.template(tokenTemplate, metadata, null, false)
          const result = await compileAndBroadcast({ host, asm, signer, wif })
          resolve(result)
        } catch (e) {
          reject(e)
        }
      })
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Claim DropERC721/DropERC1155
 * @function
 * @name claim
 * @memberof hazama-lib/Others
 *
 * @example <caption>[Claim DropERC721 with privateKey on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const privateKey = 'dummy'
 * const contractAddress = '0x...'
 * const recipientAddress = '0x...'
 * const quantity = 2
 * const clientKey = 'dummy'
 *
 * const result = hazlib.claim({ network, privateKey, contractAddress, recipientAddress, quantity, clientKey })
 * ...
 *
 * @example <caption>[Claim DropERC721 with signer and relayer on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const signer = <signer object>
 * const relayerUrl = 'https://api.defender.openzeppelin.com/autotasks/0f7cf308-.../runs/webhook/fd447953-...' // Openzeppelin webhook URI
 * const contractAddress = '0x...'
 * const recipientAddress = '0x...'
 * const quantity = 2
 * const clientKey = 'dummy'
 *
 * const result = hazlib.claim({ network, signer, contractAddress, recipientAddress, quantity, relayerUrl, clientKey })
 * ...
 *
 * @example <caption>[Claim DropERC1155 with privateKey on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const privateKey = 'dummy'
 * const contractAddress = '0x...'
 * const recipientAddress = '0x...'
 * const tokenId = 1
 * const quantity = 2
 * const clientKey = 'dummy'
 *
 * const result = hazlib.claim({ network, privateKey, contractAddress, recipientAddress, tokenId, quantity, clientKey })
 * ...
 *
 * @example <caption>[Claim DropERC1155 with signer and relayer on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const signer = <signer object>
 * const relayerUrl = 'https://api.defender.openzeppelin.com/autotasks/0f7cf308-.../runs/webhook/fd447953-...' // Openzeppelin webhook URI
 * const contractAddress = '0x...'
 * const recipientAddress = '0x...'
 * const tokenId = 1
 * const quantity = 2
 * const clientKey = 'dummy'
 *
 * const result = hazlib.claim({ network, signer, contractAddress, recipientAddress, tokenId, quantity, relayerUrl, clientKey })
 * ...
 *
 * @example <caption>[Claim NFT on HAZAMA network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'hazama'
 * const wif = 'dummy'
 * const host = 'wss://n.hazbase.com'
 * const tokenTemplate = 'nft_claimto'
 * const metadata = {
 *  path: '/hazama/...'
 *  amount: 1,
 *  address: '0x...',
 *  token: 1
 * }
 * const clientKey = 'dummy'
 *
 * const result = hazlib.claim({ network, wif, host, tokenTemplate, metadata, clientKey })
 * ...
 *
 * @param {object} params parameters for claiming
 * @param {string} params.network the chain name or rpc url
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions
 * @param {string} [params.privateKey] wallet's private key
 * @param {string} [params.wif] wallet import format
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {string} [params.contractAddress] the token contract address
 * @param {('nft_claimto')} [params.tokenTemplate] the token template (use by Hazama network)
 * @param {string} [params.recipientAddress] address to transfer the tokens to
 * @param {BigNumberish} [params.tokenId] token id
 * @param {Amount} [params.quantity] number of tokens to transfer
 * @param {HazamaClaimMetadata} [params.metadata] claim metadata
 * @param {string} [params.relayerUrl] relayer url (gasless transaction mode)
 * @param {string} [params.clientKey] client key
 * @return {TransactionResult|string} transaction id (Hazama) or transaction result
 */
lib.claim = async ({
  network,
  signer,
  privateKey,
  wif,
  host = HOST,
  contractAddress,
  tokenTemplate,
  recipientAddress,
  tokenId,
  quantity,
  metadata,
  relayerUrl,
  clientKey,
  isCount = true,
}) => {
  let caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 10,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (!network) throw new Error(messages.NETWORK_UNDEFINED)

    caller = await getCaller({ network, host, signer, privateKey, wif })
    const isHazamaNetwork = network === 'hazama'

    const result = !isHazamaNetwork
      ? await claimOnEVMNetwork({
          network,
          signer,
          privateKey,
          contractAddress,
          recipientAddress,
          tokenId,
          quantity,
          relayerUrl,
        })
      : await claimOnHazamaNetwork({
          signer,
          wif,
          host,
          tokenTemplate,
          metadata,
        })

    let payload = {
      chainId: getChainId(network),
      tokenId,
      contractAddress,
      isClaimed: true,
      walletAddress: caller,
      transactionHash: formatTransactionHash(result),
    }

    if (isHazamaNetwork) {
      const contractInfo = getContractInfoFromPath(metadata?.path || '')
      payload = {
        ...payload,
        tokenId: metadata?.token || null,
        contractId: contractInfo?.contractId || null,
        projectId: contractInfo?.projectId || null,
      }
    }

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 10,
      walletAddress: caller,
      status: RequestTransactionStatus.SUCCEEDED,
      transactionHash: formatTransactionHash({ input: result }),
      apiRequests: [
        {
          apiCallStatus: ApiRequestStatus.NOT_CALLED_YET,
          apiBody: JSON.stringify({ tokens: [payload] }),
          apiURI: '/api/app/token/save-multiple-tokens',
          apiMethod: 'post',
        },
      ],
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 10,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * grant roles on EVM Network
 *
 * @private
 * @param {object} params parameters for granting roles
 * @param {string} params.network the chain name or rpc url
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions
 * @param {string} [params.privateKey] wallet's private key
 * @param {string} params.contractAddress the token contract address
 * @param {RolesObject} params.roles roles of contract
 * @param {string} [params.relayerUrl] relayer url (gasless transaction mode)
 * @return {Promise<TransactionResult[]>} promise
 */
async function grantRolesOnEVMNetwork({
  network,
  signer,
  privateKey,
  contractAddress,
  roles,
  relayerUrl,
}) {
  return new Promise(async (resolve, reject) => {
    try {
      let contract = null
      let options = {
        readonlySettings: {
          rpcUrl: getRpcUrl(network),
        },
        ...(isArrayFull(gatewayUrls) ? { gatewayUrls } : {}),
        ...(isNil(thirdwebSecretKey) &&
        (!isNil(thirdwebClientId) || !isNil(thirdwebClientIdFromAPI))
          ? { clientId: thirdwebClientId || thirdwebClientIdFromAPI }
          : {}),
        ...(!isNil(thirdwebSecretKey) ? { secretKey: thirdwebSecretKey } : {}),
      }

      if (!roles) throw new Error(messages.ROLES_UNDEFINED)

      if (!signer && !privateKey)
        throw new Error(messages.SIGNER_PRIVATEKEY_UNDEFINED)

      if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

      if (!ethers.utils.isAddress(contractAddress))
        throw new Error(messages.CONTRACT_ADDRESS_INVALID)

      const rolesArr = []

      Object.keys(roles).forEach((roleName) => {
        if (!['admin', 'minter', 'transfer'].includes(roleName)) {
          return
        }

        const addressByRoleName = roles[roleName]

        addressByRoleName.forEach((address) => {
          if (!ethers.utils.isAddress(address))
            throw new Error(messages.ROLE_CONTRACT_ADDRESS_INVALID)

          rolesArr.push({
            name: roleName,
            address,
          })
        })
      })

      if (relayerUrl) {
        options = {
          ...options,
          gasless: {
            openzeppelin: { relayerUrl },
          },
        }
      }

      let sdk = new ThirdwebSDK(network, options)
      if (signer) {
        sdk.updateSignerOrProvider(signer)
      }

      if (privateKey) {
        sdk = new ThirdwebSDK(
          new ethers.Wallet(
            privateKey,
            ethers.getDefaultProvider(getRpcUrl(network)),
          ),
          options,
        )
      }

      const contractType = await sdk.resolveContractType(contractAddress)
      if (
        ![
          ContractType.DropERC721,
          ContractType.TokenERC721,
          ContractType.DropERC1155,
          ContractType.TokenERC20,
          ContractType.Multiwrap,
        ].includes(contractType)
      )
        throw new Error(messages.CONTRACT_TYPE_IS_INVALID)

      switch (contractType) {
        case ContractType.DropERC721:
        case ContractType.TokenERC721:
        case ContractType.DropERC1155:
        case ContractType.TokenERC20:
        case ContractType.Multiwrap:
          contract = await sdk.getContract(contractAddress, contractType)
          break
        default:
          throw new Error(messages.TOKEN_TYPE_IS_NOT_SUPPORTED)
      }

      const results = []
      for (let index = 0; index < rolesArr.length; index++) {
        const { name, address } = rolesArr[index]

        const role = await contract.roles.get(name)

        if (
          role.includes(address) ||
          (name === 'transfer' && role.includes(ethers.constants.AddressZero))
        ) {
          continue
        }

        const result = await contract.roles.grant(name, address)
        results.push(result)
      }

      resolve(results)
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Add permissions<br/><br/>
 * Important Note:
 * 1. If wallet address exists in role, it will be skipped
 * 2. If zero address(0x0000000000000000000000000000000000000000) exists in role, all wallet address of role will be skipped
 * @function
 * @name addPermissions
 * @memberof hazama-lib/ERC20/ERC721/ERC1155
 *
 * @example <caption>[add permissions with privateKey on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const privateKey = 'dummy'
 * const contractAddress = '0x...'
 * const roles = {
 *   admin: ['0x...'],
 *   minter: ['0x...'],
 *   transfer: ['0x...'],
 * }
 * const clientKey = 'dummy'
 *
 * const result = hazlib.addPermissions({ network, privateKey, contractAddress, roles, clientKey })
 * ...
 *
 * @example <caption>[add permissions with signer and relayer on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const signer = <signer object>
 * const relayerUrl = 'https://api.defender.openzeppelin.com/autotasks/0f7cf308-.../runs/webhook/fd447953-...' // Openzeppelin webhook URI
 * const contractAddress = '0x...'
 * const roles = {
 *   admin: ['0x...'],
 *   minter: ['0x...'],
 *   transfer: ['0x...'],
 * }
 * const clientKey = 'dummy'
 *
 * const result = hazlib.addPermissions({ network, signer, contractAddress, roles, relayerUrl, clientKey })
 * ...
 *
 * @param {object} params parameters for adding permissions
 * @param {string} params.network the chain name or rpc url
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions
 * @param {string} [params.privateKey] wallet's private key
 * @param {string} params.contractAddress the token contract address
 * @param {RolesObject} params.roles roles of contract
 * @param {string} [params.relayerUrl] relayer url (gasless transaction mode)
 * @param {string} [params.clientKey] client key
 * @return {TransactionResult[]} transaction id (Hazama) or transaction result
 */
lib.addPermissions = async ({
  network,
  signer,
  privateKey,
  contractAddress,
  roles,
  relayerUrl,
  clientKey,
  isCount = true,
}) => {
  let caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 7,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (!network) throw new Error(messages.NETWORK_UNDEFINED)

    if (network === 'hazama') throw new Error(messages.NETWORK_IS_NOT_SUPPORTED)

    caller = await getCaller({ network, signer, privateKey })

    const result = await grantRolesOnEVMNetwork({
      network,
      signer,
      privateKey,
      contractAddress,
      roles,
      relayerUrl,
    })

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 7,
      walletAddress: caller,
      status: RequestTransactionStatus.SUCCEEDED,
      transactionHash: formatTransactionHash({ input: result }),
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 7,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * revoke roles on EVM Network
 *
 * @private
 * @param {object} params parameters for revoking roles
 * @param {string} params.network the chain name or rpc url
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions
 * @param {string} [params.privateKey] wallet's private key
 * @param {string} params.contractAddress the token contract address
 * @param {RolesObject} params.roles roles of contract
 * @param {string} [params.relayerUrl] relayer url (gasless transaction mode)
 * @return {Promise<TransactionResult[]>} promise
 */
async function revokeRolesOnEVMNetwork({
  network,
  signer,
  privateKey,
  contractAddress,
  roles,
  relayerUrl,
}) {
  return new Promise(async (resolve, reject) => {
    try {
      let contract = null
      let options = {
        readonlySettings: {
          rpcUrl: getRpcUrl(network),
        },
        ...(isArrayFull(gatewayUrls) ? { gatewayUrls } : {}),
        ...(isNil(thirdwebSecretKey) &&
        (!isNil(thirdwebClientId) || !isNil(thirdwebClientIdFromAPI))
          ? { clientId: thirdwebClientId || thirdwebClientIdFromAPI }
          : {}),
        ...(!isNil(thirdwebSecretKey) ? { secretKey: thirdwebSecretKey } : {}),
      }

      if (!roles) throw new Error(messages.ROLES_UNDEFINED)

      if (!signer && !privateKey)
        throw new Error(messages.SIGNER_PRIVATEKEY_UNDEFINED)

      if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

      if (!ethers.utils.isAddress(contractAddress))
        throw new Error(messages.CONTRACT_ADDRESS_INVALID)

      const rolesArr = []

      Object.keys(roles).forEach((roleName) => {
        if (!['admin', 'minter', 'transfer'].includes(roleName)) {
          return
        }

        const addressByRoleName = roles[roleName]

        addressByRoleName.forEach((address) => {
          // if (!ethers.utils.isAddress(address)) throw new Error(messages.ROLE_CONTRACT_ADDRESS_INVALID)

          rolesArr.push({
            name: roleName,
            address,
          })
        })
      })

      if (relayerUrl) {
        options = {
          ...options,
          gasless: {
            openzeppelin: { relayerUrl },
          },
        }
      }

      let sdk = new ThirdwebSDK(network, options)
      if (signer) {
        sdk.updateSignerOrProvider(signer)
      }

      if (privateKey) {
        sdk = new ThirdwebSDK(
          new ethers.Wallet(
            privateKey,
            ethers.getDefaultProvider(getRpcUrl(network)),
          ),
          options,
        )
      }

      const contractType = await sdk.resolveContractType(contractAddress)
      if (
        ![
          ContractType.DropERC721,
          ContractType.TokenERC721,
          ContractType.DropERC1155,
          ContractType.TokenERC20,
          ContractType.Multiwrap,
        ].includes(contractType)
      )
        throw new Error(messages.CONTRACT_TYPE_IS_INVALID)

      switch (contractType) {
        case ContractType.DropERC721:
        case ContractType.TokenERC721:
        case ContractType.DropERC1155:
        case ContractType.TokenERC20:
        case ContractType.Multiwrap:
          contract = await sdk.getContract(contractAddress, contractType)
          break
        default:
          throw new Error(messages.TOKEN_TYPE_IS_NOT_SUPPORTED)
      }

      const results = []
      for (let index = 0; index < rolesArr.length; index++) {
        const { name, address } = rolesArr[index]

        const role = await contract.roles.get(name)

        if (!role.includes(address)) {
          continue
        }

        const result = await contract.roles.revoke(name, address)
        results.push(result)
      }

      resolve(results)
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Remove permissions<br/><br/>
 * @function
 * @name removePermissions
 * @memberof hazama-lib/ERC20/ERC721/ERC1155
 *
 * @example <caption>[remove permissions with privateKey on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const privateKey = 'dummy'
 * const contractAddress = '0x...'
 * const roles = {
 *   admin: ['0x...'],
 *   minter: ['0x...'],
 *   transfer: ['0x...'],
 * }
 * const clientKey = 'dummy'
 *
 * const result = hazlib.removePermissions({ network, privateKey, contractAddress, roles, clientKey })
 * ...
 *
 * @example <caption>[remove permissions with signer and relayer on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const signer = <signer object>
 * const relayerUrl = 'https://api.defender.openzeppelin.com/autotasks/0f7cf308-.../runs/webhook/fd447953-...' // Openzeppelin webhook URI
 * const contractAddress = '0x...'
 * const roles = {
 *   admin: ['0x...'],
 *   minter: ['0x...'],
 *   transfer: ['0x...'],
 * }
 * const clientKey = 'dummy'
 *
 * const result = hazlib.removePermissions({ network, signer, contractAddress, roles, relayerUrl, clientKey })
 * ...
 *
 * @param {object} params parameters for removing permissions
 * @param {string} params.network the chain name or rpc url
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions
 * @param {string} [params.privateKey] wallet's private key
 * @param {string} params.contractAddress the token contract address
 * @param {RolesObject} params.roles roles of contract
 * @param {string} [params.relayerUrl] relayer url (gasless transaction mode)
 * @param {string} [params.clientKey] client key
 * @return {TransactionResult[]} transaction id (Hazama) or transaction result
 */
lib.removePermissions = async ({
  network,
  signer,
  privateKey,
  contractAddress,
  roles,
  relayerUrl,
  clientKey,
  isCount = true,
}) => {
  let caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 9,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (!network) throw new Error(messages.NETWORK_UNDEFINED)

    if (network === 'hazama') throw new Error(messages.NETWORK_IS_NOT_SUPPORTED)

    caller = await getCaller({ network, signer, privateKey })

    const result = await revokeRolesOnEVMNetwork({
      network,
      signer,
      privateKey,
      contractAddress,
      roles,
      relayerUrl,
    })

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 9,
      walletAddress: caller,
      status: RequestTransactionStatus.SUCCEEDED,
      transactionHash: formatTransactionHash({ input: result }),
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 9,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * Check whether token is native or not
 *
 * @private
 * @param {string} tokenAddress token address
 * @return {boolean} native token: return true, otherwise: return false
 */
function isNativeToken(tokenAddress) {
  return (
    tokenAddress.toLowerCase() === NATIVE_TOKEN_ADDRESS ||
    tokenAddress.toLowerCase() === ethers.constants.AddressZero
  )
}

/**
 * Fetch currency meta data from abi
 *
 * @private
 * @param {any} provider provider
 * @param {string} asset asset
 * @return {Promise<Currency>} curency metadata
 */
async function fetchCurrencyMetadata(provider, asset) {
  if (isNativeToken(asset)) {
    const network = await provider.getNetwork()

    const nativeToken = getNativeTokenByChainId(network.chainId)

    return {
      name: nativeToken.name,
      symbol: nativeToken.symbol,
      decimals: nativeToken.decimals,
    }
  } else {
    const erc20 = new ethers.Contract(asset, ERC20MetadataAbi, provider)

    const [name, symbol, decimals] = await Promise.all([
      erc20.name(),
      erc20.symbol(),
      erc20.decimals(),
    ])

    return {
      name,
      symbol,
      decimals,
    }
  }
}

/**
 * Get currency metadata<br/><br/>
 * @function
 * @name getCurrencyMetadata
 * @memberof hazama-lib/ERC20/ERC721/ERC1155
 *
 * @example
 * import hazlib from 'hazama-lib'
 *
 * const network = 'polygon'
 * const currencyAddress = '0x...'
 * const clientKey = 'dummy'
 *
 * const metadata = hazlib.getCurrencyMetadata({ network, currencyAddress, clientKey })
 * ...
 *
 * @param {object} params parameters for getting currency metadatas
 * @param {string} params.network the chain name or rpc url
 * @param {string} params.currencyAddress contract address of currency
 * @param {string} [params.clientKey] client key
 * @return {Promise<Currency>} curency metadata
 */
lib.getCurrencyMetadata = async ({
  network,
  currencyAddress,
  clientKey,
  isCount = true,
}) => {
  const caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 8,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (!network) throw new Error(messages.NETWORK_UNDEFINED)

    if (!currencyAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

    const provider = ethers.getDefaultProvider(getRpcUrl(network))

    const result = await fetchCurrencyMetadata(provider, currencyAddress)

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 8,
      status: RequestTransactionStatus.SUCCEEDED,
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 8,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * Check and return the network that token exists
 *
 * @private
 * @param {object} params parameters
 * @param {string} [params.contractNetwork] the chain name or rpc url
 * @param {string} params.contractAddress contract address
 * @return {object} result
 */
async function resolveContractTypeWithMultiNetwork({
  contractNetwork,
  contractAddress,
}) {
  if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

  const networks = [
    {
      network: 'ethereum',
      sdk: new ThirdwebSDK(getRpcUrlsByNetWork('ethereum'), {
        ...(isNil(thirdwebSecretKey) &&
        (!isNil(thirdwebClientId) || !isNil(thirdwebClientIdFromAPI))
          ? { clientId: thirdwebClientId || thirdwebClientIdFromAPI }
          : {}),
        ...(!isNil(thirdwebSecretKey) ? { secretKey: thirdwebSecretKey } : {}),
      }),
    },
    {
      network: 'polygon',
      sdk: new ThirdwebSDK(getRpcUrlsByNetWork('polygon'), {
        ...(isNil(thirdwebSecretKey) &&
        (!isNil(thirdwebClientId) || !isNil(thirdwebClientIdFromAPI))
          ? { clientId: thirdwebClientId || thirdwebClientIdFromAPI }
          : {}),
        ...(!isNil(thirdwebSecretKey) ? { secretKey: thirdwebSecretKey } : {}),
      }),
    },
    {
      network: 'goerli',
      sdk: new ThirdwebSDK(getRpcUrlsByNetWork('goerli'), {
        ...(isNil(thirdwebSecretKey) &&
        (!isNil(thirdwebClientId) || !isNil(thirdwebClientIdFromAPI))
          ? { clientId: thirdwebClientId || thirdwebClientIdFromAPI }
          : {}),
        ...(!isNil(thirdwebSecretKey) ? { secretKey: thirdwebSecretKey } : {}),
      }),
    },
    {
      network: 'mumbai',
      sdk: new ThirdwebSDK(getRpcUrlsByNetWork('mumbai'), {
        ...(isNil(thirdwebSecretKey) &&
        (!isNil(thirdwebClientId) || !isNil(thirdwebClientIdFromAPI))
          ? { clientId: thirdwebClientId || thirdwebClientIdFromAPI }
          : {}),
        ...(!isNil(thirdwebSecretKey) ? { secretKey: thirdwebSecretKey } : {}),
      }),
    },
  ]

  const filteredNetworks = (
    await Promise.all(
      networks.map((data) => {
        return new Promise(async (resolve, reject) => {
          const contractType = await resolveContractTypeWithCustom({
            sdk: data.sdk,
            contractAddress,
          }).catch((e) => {})
          const result = { ...data, contractType }
          resolve(result)
        })
      }),
    )
  ).filter((data) => data.contractType !== ContractType.Custom)

  if (filteredNetworks.length === 1) {
    return filteredNetworks.pop()
  } else if (filteredNetworks.length <= 0) {
    throw new Error(messages.CONTRACT_NOT_EXIST_OR_UNSUPPORT_NETWORK)
  } else {
    if (!contractNetwork) {
      throw new Error(messages.CONTRACT_NETWORK_UNDEFINED)
    }
    if (!!contractNetwork && !SUPPORTED_CHAIN_NAME.includes(contractNetwork)) {
      throw new Error(messages.CONTRACT_NETWORK_NOTSUPPORT)
    }

    return filteredNetworks
      .filter((data) => !!data.contractType && data.network === contractNetwork)
      .pop()
  }
}

/**
 * Check and return the contract type
 *
 * @private
 * @param {object} params parameters
 * @param {ThirdwebSDK} params.sdk Thirdweb SDK
 * @param {string} params.contractAddress contract address
 * @return {string} contractType
 */
async function resolveContractTypeWithCustom({ sdk, contractAddress }) {
  if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

  let contractType = await sdk.resolveContractType(contractAddress)
  if (
    typeof contractType !== 'string' ||
    contractType === ContractType.Custom
  ) {
    try {
      const contractMetadata = await sdk
        .getPublisher()
        .fetchCompilerMetadataFromAddress(contractAddress)
      contractType = contractMetadata?.name
      switch (contractType) {
        case CustomContract.ContractType.BurnToRedeem1155:
        case CustomContract.ContractType.Vote:
          contractType = getParentContractType(contractType)
          break
        default:
          contractType = ContractType.Custom
      }
    } catch (e) {}
  }
  return contractType
}

/**
 * Get token detail on EVM network
 *
 * @private
 * @param {object} params parameters
 * @param {string} params.contractNetwork the chain name or rpc url
 * @param {string} params.contractAddress contract address
 * @param {BigNumberish} params.tokenId the id of the token
 * @return {TokenDetail} token detail
 */
async function getTokenDetailOnEVMNetwork({
  contractNetwork,
  contractAddress,
  tokenId,
}) {
  if (!contractNetwork) throw new Error(messages.NETWORK_UNDEFINED)

  if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

  if (!tokenId && tokenId !== 0) throw new Error(messages.TOKEN_ID_UNDEFINED)

  const res = await resolveContractTypeWithMultiNetwork({
    contractNetwork,
    contractAddress,
  })

  if (res) {
    const { contractType, network, sdk } = res

    if (!!contractNetwork && contractNetwork !== network) {
      throw new Error(messages.TOKEN_NOTFOUND)
    }

    let sellerFeeBasisPoints
    let owner
    let token
    let royaltyInfo
    let isTransferable = false
    let transerRole = []

    const contract = await sdk.getContract(contractAddress, contractType)
    switch (contractType) {
      case ContractType.DropERC1155:
      case ContractType.TokenERC1155: {
        const settings = {
          apiKey: getAlchemyApiKey(network),
          network: getAlchemyNetwork(network),
        }
        const alchemy = new Alchemy(settings)

        let alchemyOwners
        ;[royaltyInfo, token, alchemyOwners, transerRole] =
          await Promise.all([
            contract.royalties.getDefaultRoyaltyInfo(),
            contract.erc1155.get(tokenId),
            alchemy.nft
              .getOwnersForNft(contractAddress, tokenId)
              .catch((reason) => console.log('getOwnersForNft failed', reason)),
            contract.roles.get('transfer'),
          ])
        owner = alchemyOwners?.owners || []

        sellerFeeBasisPoints = royaltyInfo.seller_fee_basis_points
        if (transerRole.includes(ethers.constants.AddressZero))
          isTransferable = true
        break
      }
      case ContractType.DropERC721:
      case ContractType.TokenERC721:
        ;[royaltyInfo, token, transerRole] = await Promise.all([
          contract.royalties.getDefaultRoyaltyInfo(),
          contract.erc721.get(tokenId),
          contract.roles.get('transfer'),
        ])

        owner = token.owner
        sellerFeeBasisPoints = royaltyInfo.seller_fee_basis_points
        if (transerRole.includes(ethers.constants.AddressZero))
          isTransferable = true
        break
      default:
        throw new Error(messages.CONTRACT_TYPE_IS_NOT_SUPPORTED)
    }

    if (token?.metadata?.name.includes('Failed to load NFT metadata'))
      throw new Error(messages.TOKEN_NOTFOUND)

    let tokenUri = token?.metadata?.uri
    const isIPFS = tokenUri?.includes('ipfs://')
    tokenUri = !isIPFS ? tokenUri + `?t=${moment().unix()}` : ''
    const tokenMetadata = !isIPFS ? await fetchTokenMetadata({
      tokenId,
      tokenUri,
      storage: sdk.storage,
    }) : {}
  
    return {
      ...token?.metadata,
      ...tokenMetadata,
      tokenId: Number(token?.metadata.id),
      contractId: contractAddress,
      type: contractType,
      tokenStandard: TokenStandard[contractType],
      isTransferable,
      isThirdweb: true,
      sellerFeeBasisPoints: sellerFeeBasisPoints / 100,
      tokenProperties: tokenMetadata?.attributes ? tokenMetadata?.attributes : token?.metadata?.attributes,
      owner,
      network,
      supply: parseInt(token?.supply, 10),
    }
  }
}

/**
 * Get token detail on HAZAMA network
 *
 * @private
 * @param {object} params parameters
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {string} params.contractAddress contract address
 * @param {BigNumberish} params.tokenId the id of the token
 * @return {Promise<TokenDetail>} token detail
 */
async function getTokenDetailOnHazamaNetwork({
  host = HOST,
  contractAddress,
  tokenId,
}) {
  return new Promise(async (resolve, reject) => {
    try {
      if (!host) throw new Error(messages.HOST_UNDEFINED)

      if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

      if (!tokenId && tokenId !== 0)
        throw new Error(messages.TOKEN_ID_UNDEFINED)

      lib.connect(host, async () => {
        const path = await lib.get(
          `/_/contract.${contractAddress}.namespace`,
          null,
          false,
        )

        if (path === null) reject(new Error(messages.TOKEN_NOTFOUND))

        const meta = await lib.get(`${path}.meta`, null, false)

        if (meta.Type === ContractType.FUNGIBLE) {
          reject(new Error(messages.CONTRACT_TYPE_IS_NOT_SUPPORTED))
        } else {
          const data = await lib.get(`${path}/${tokenId}`, null, false)

          if (!data) reject(new Error(messages.TOKEN_NOTFOUND))

          const owners = []
          if (meta.Type !== ContractType.NFT) {
            const history = await lib.get(`/_/addr.${data.owner}`, null, false)

            for (const x in history) {
              if (new RegExp(`^${path}/${tokenId}/[0-9]+$`).test(x)) {
                const meta = await lib.get(x, null, false)
                !!meta?.owner &&
                  !owners.includes(meta?.owner) &&
                  owners.push(meta.owner)
              }
            }
          }

          const tokenDetail = {
            title: data.name,
            description: data.description,
            tokenId,
            image: data.image.replace(
              'gateway.ipfscdn.io',
              'hzbs.infura-ipfs.io',
            ),
            isTransferable: meta.Type === ContractType.NFT,
            sellerFeeBasisPoints: meta.SellerFeeBasisPoints,
            isThirdweb: false,
            type: meta.Type,
            tokenStandard: meta.Type,
            ...(meta.Type === ContractType.NFT
              ? { owners: [data.owner] }
              : { owners }),
          }

          resolve(tokenDetail)
        }
      })
    } catch (e) {
      reject(e)
    }
  })
}

/**
 *
 * Get detail information of token
 * @function
 * @name getTokenDetail
 * @memberof hazama-lib/Others
 *
 * @example <caption>[get detail information of token on EVM network]</caption>
 * import hazlib from hazama-lib
 * ...
 * const network = 'polygon'
 * const contractAddress = '0x...'
 * const tokenId = 0
 * const clientKey = 'dummy'
 *
 * const tokenDetail = hazlib.getTokenDetail({ network, contractAddress, tokenId, clientKey })
 * ...
 *
 * @example <caption>[get detail information of token on HAZAMA network]</caption>
 * import hazlib from hazama-lib
 * ...
 * const network = 'hazama'
 * const host = 'ws://...'
 * const contractAddress = '0x...'
 * const tokenId = 0
 * const clientKey = 'dummy'
 *
 * const tokenDetail = await hazlib.getTokenDetail({ network, host, contractAddress, tokenId, clientKey })
 * ...
 *
 * @param {object} params parameters for getting contents of a wrapped token bundle
 * @param {string} params.network the chain name or rpc url
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {string} params.contractAddress the address of the deployed contract
 * @param {BigNumberish} params.tokenId the id of the wrapped token bundle
 * @param {string} [params.clientKey] client key
 * @returns {TokenDetail} Output for the contents of a bundle wrapped token
 */
lib.getTokenDetail = async ({
  network,
  host = HOST,
  contractAddress,
  tokenId,
  clientKey,
  isCount = true,
}) => {
  const caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 19,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (!network) throw new Error(messages.NETWORK_UNDEFINED)

    if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

    if (!tokenId && tokenId !== 0) throw new Error(messages.TOKEN_ID_UNDEFINED)

    const result =
      network !== 'hazama'
        ? await getTokenDetailOnEVMNetwork({
            contractNetwork: network,
            contractAddress,
            tokenId,
          })
        : await getTokenDetailOnHazamaNetwork({
            host,
            contractAddress,
            tokenId,
          })

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 19,
      walletAddress: caller,
      status: RequestTransactionStatus.SUCCEEDED,
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 19,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 *
 * Get detail information of token by Alchemy
 * @function
 * @name getTokenDetailByAlchemy
 * @memberof hazama-lib/Others
 *
 * @example
 * import hazlib from hazama-lib
 * ...
 * const network = 'polygon'
 * const contractAddress = '0x...'
 * const tokenId = 0
 * const clientKey = 'dummy'
 *
 * const tokenDetail = await hazlib.getTokenDetailByAlchemy({ network, contractAddress, tokenId, clientKey })
 * ...
 *
 * @param {object} params parameters for getting contents of a wrapped token bundle
 * @param {string} params.network the chain name or rpc url
 * @param {string} params.contractAddress the address of the deployed contract
 * @param {BigNumberish} params.tokenId the id of the wrapped token bundle
 * @param {string} [params.clientKey] client key
 * @returns {TokenDetail} Output for the contents of a bundle wrapped token
 */
lib.getTokenDetailByAlchemy = async ({
  network,
  contractAddress,
  tokenId,
  clientKey,
  isCount = true,
}) => {
  const caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 20,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (!network) throw new Error(messages.NETWORK_UNDEFINED)

    if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

    if (!tokenId && tokenId !== 0) throw new Error(messages.TOKEN_ID_UNDEFINED)

    let result
    const settings = {
      apiKey: getAlchemyApiKey(network),
      network: getAlchemyNetwork(network),
    }
    const breakTimeout = new Date().getTime() + WaitingAlchemyPeriod
    const alchemy = new Alchemy(settings)
    while (new Date().getTime() < breakTimeout) {
      try {
        let nft = {}
        try {
          nft = await alchemy.nft.getNftMetadata(contractAddress, tokenId)
        } catch (error) {
          throw new Error(messages.TOKEN_NOTFOUND)
        }

        let responseForGetOwners = []

        try {
          responseForGetOwners = await alchemy.nft.getOwnersForNft(
            contractAddress,
            tokenId,
          )
        } catch (error) {}

        if (!nft.metadataError) {
          const metadata = nft?.rawMetadata || {}
          const media = nft?.media || []

          result = {
            title: metadata?.name || '',
            description: metadata?.description || '',
            image: media?.[0]?.gateway || '',
            tokenId,
            tokenProperties: metadata?.attributes || null,
            tokenStandard: nft?.tokenType || null,
            isThirdweb: false,
            owner: responseForGetOwners?.owners || [],
          }

          createRequestTransaction({
            clientKey: targetClientKey,
            functionId: 20,
            walletAddress: caller,
            status: RequestTransactionStatus.SUCCEEDED,
            isCount,
          })

          return result
        } else {
          await sleep(5000)
        }
      } catch (err) {
        await sleep(5000)
      }
    }

    if (!result) throw new Error(messages.TOKEN_NOTFOUND)
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 20,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * Update contract on EVM network
 *
 * @private
 * @param {object} params parameters for updating contract
 * @param {string} params.network the chain name or rpc url
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions.<br/>If private key is not set, Signer is required.
 * @param {string} [params.privateKey] wallet's private key.<br/>If Signer is not set, private key is required.
 * @param {string} params.contractAddress contract address
 * @param {ContractUpdateMetadata} params.contractMetadata contract metadata
 * @param {string} [params.relayerUrl] relayer url (gasless transaction mode)
 * @return {Promise<TransactionResult>} promise
 */
async function updateContractOnEVMNetwork({
  network,
  signer,
  privateKey,
  contractAddress,
  contractMetadata,
  relayerUrl,
}) {
  return new Promise(async (resolve, reject) => {
    try {
      let options = {
        readonlySettings: {
          rpcUrl: getRpcUrl(network),
        },
        ...(isArrayFull(gatewayUrls) ? { gatewayUrls } : {}),
        ...(isNil(thirdwebSecretKey) &&
        (!isNil(thirdwebClientId) || !isNil(thirdwebClientIdFromAPI))
          ? { clientId: thirdwebClientId || thirdwebClientIdFromAPI }
          : {}),
        ...(!isNil(thirdwebSecretKey) ? { secretKey: thirdwebSecretKey } : {}),
      }

      if (!network) throw new Error(messages.NETWORK_UNDEFINED)

      if (!signer && !privateKey)
        throw new Error(messages.SIGNER_PRIVATEKEY_UNDEFINED)

      if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

      if (!contractMetadata)
        throw new Error(messages.CONTRACT_METADATA_UNDEFINED)

      if (relayerUrl) {
        options = {
          ...options,
          gasless: {
            openzeppelin: { relayerUrl },
          },
        }
      }

      let sdk = new ThirdwebSDK(network, options)
      if (signer) {
        sdk.updateSignerOrProvider(signer)
      }

      if (privateKey) {
        sdk = new ThirdwebSDK(
          new ethers.Wallet(
            privateKey,
            ethers.getDefaultProvider(getRpcUrl(network)),
          ),
          options,
        )
      }

      const contractType = await sdk.resolveContractType(contractAddress)
      const contract = await sdk.getContract(contractAddress, contractType)
      const result = await contract.metadata.update(contractMetadata)
      resolve(result)
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Update contract on HAZAMA network
 *
 * @private
 * @param {object} params parameters for updating contract
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions.<br/>If wif is not set, Signer is required.
 * @param {string} [params.wif] wallet import format.<br/>If Signer is not set, wif is required.
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {('nft_extensible'|'nft_nontransferable'|'fungible_token')} params.contractTemplate the contract template
 * @param {HazamaContractUpdateMetadata} params.contractMetadata contract metadata
 * @return {Promise<string>} promise
 */
async function updateContractOnHazamaNetwork({
  signer,
  wif,
  host = HOST,
  contractTemplate,
  contractMetadata,
}) {
  return deployContractOnHazamaNetwork({
    signer,
    wif,
    host,
    contractTemplate,
    contractMetadata,
  })
}

/**
 * Update contract
 * @function
 * @name updateContract
 * @memberof hazama-lib/Others
 *
 * @example <caption>[Update contract with privateKey on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const privateKey = 'dummy'
 * const contractAddress = '0x...'
 * const contractMetadata = {
 *   name: 'contract name',
 * }
 * const clientKey = 'dummy'
 *
 * const result = await hazlib.updateContract({ network, privateKey, contractAddress, contractMetadata, clientKey })
 * ...
 *
 * @example <caption>[Update contract with signer and relayer on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const signer = <signer object>
 * const relayerUrl = 'https://api.defender.openzeppelin.com/autotasks/0f7cf308-.../runs/webhook/fd447953-...' // Openzeppelin webhook URI
 * const contractAddress = '0x...'
 * const contractMetadata = {
 *   name: 'contract name',
 * }
 * const clientKey = 'dummy'
 *
 * const result = await hazlib.updateContract({ network, signer, contractAddress, contractMetadata, relayerUrl, clientKey })
 * ...
 *
 * @example <caption>[Update contract on HAZAMA network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'hazama'
 * const wif = 'dummny'
 * const host = 'wss://n.hazbase.com'
 * const contractTemplate = 'nft_extensible'
 * const contractMetadata = {
 *  path: '/hazama/...',
 *  name: 'contract name',
 *  primarySaleRecipient: '0x...',
 *  author: 'author name',
 *  tokenVersion: '0.0.1',
 *  decimals: 18,
 *  ...
 * }
 * const clientKey = 'dummy'
 *
 * const result = hazlib.updateContract({ network, wif, host, contractTemplate, contractMetadata, clientKey })
 * ...
 *
 * @param {object} params parameters for deploying contract
 * @param {string} params.network the chain name or rpc url
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions
 * @param {string} [params.privateKey] wallet's private key
 * @param {string} [params.wif] wallet import format
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {string} [params.contractAddress] contract address
 * @param {('nft_extensible'|'nft_nontransferable'|'fungible_token')} [params.contractTemplate] the contract template
 * @param {ContractUpdateMetadata|HazamaContractUpdateMetadata} params.contractMetadata contract metadata
 * @param {string} [params.relayerUrl] relayer url (gasless transaction mode)
 * @param {string} [params.clientKey] client key
 * @return {(TransactionResult|string)} transaction id (Hazama) or transaction result (other networks)
 */
lib.updateContract = async ({
  network,
  host = HOST,
  signer,
  privateKey,
  wif,
  contractAddress,
  contractTemplate,
  contractMetadata,
  relayerUrl,
  clientKey,
  isCount = true,
}) => {
  let caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 30,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (!network) throw new Error(messages.NETWORK_UNDEFINED)

    if (!contractMetadata) throw new Error(messages.CONTRACT_METADATA_UNDEFINED)

    if (network === 'hazama' && !host) throw new Error(messages.HOST_UNDEFINED)

    if (network === 'hazama' && !contractTemplate)
      throw new Error(messages.CONTRACT_TEMPLATE_UNDEFINED)

    if (network === 'hazama' && !signer && !wif)
      throw new Error(messages.SIGNER_WIF_UNDEFINED)

    if (network !== 'hazama' && !contractAddress)
      throw new Error(messages.WALLET_ADDRESS_UNDEFINED)

    if (network !== 'hazama' && !signer && !privateKey)
      throw new Error(messages.SIGNER_PRIVATEKEY_UNDEFINED)

    caller = await getCaller({ network, host, signer, privateKey, wif })
    const isHazamaNetwork = network === 'hazama'

    const result = !isHazamaNetwork
      ? await updateContractOnEVMNetwork({
          network,
          signer,
          privateKey,
          contractAddress,
          contractMetadata,
          relayerUrl,
        })
      : await updateContractOnHazamaNetwork({
          signer,
          wif,
          host,
          contractTemplate,
          contractMetadata,
        })

    let payload = {
      chainId: getChainId(network),
      contractAddress,
      walletAddress: caller,
    }

    if (isHazamaNetwork) {
      const contractInfo = getContractInfoFromPath(contractMetadata?.path || '')
      payload = {
        ...payload,
        contractId: contractInfo?.contractId || null,
        projectId: contractInfo?.projectId || null,
      }
    }

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 30,
      walletAddress: caller,
      status: RequestTransactionStatus.SUCCEEDED,
      transactionHash: formatTransactionHash({ input: result }),
      apiRequests: [
        {
          apiCallStatus: ApiRequestStatus.NOT_CALLED_YET,
          apiBody: JSON.stringify(payload),
          apiURI: '/api/app/contracts',
          apiMethod: 'post',
        },
      ],
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 30,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * Get contract type on EVM Network
 *
 * @private
 * @param {object} params parameters for getting contract type
 * @param {string} params.network the chain name or rpc url
 * @param {string} params.contractAddress the contract address
 * @return {Promise<ContractTypeResult>} promise
 */
async function getContractTypeOnEVMNetwork({ network, contractAddress }) {
  return new Promise(async (resolve, reject) => {
    try {
      if (!network) throw new Error(messages.NETWORK_UNDEFINED)

      if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

      const sdk = new ThirdwebSDK(network, {
        ...(isNil(thirdwebSecretKey) &&
        (!isNil(thirdwebClientId) || !isNil(thirdwebClientIdFromAPI))
          ? { clientId: thirdwebClientId || thirdwebClientIdFromAPI }
          : {}),
        ...(!isNil(thirdwebSecretKey) ? { secretKey: thirdwebSecretKey } : {}),
      })

      const contractType = await resolveContractTypeWithCustom({
        sdk,
        contractAddress,
      })
      resolve({ type: StandardContractType?.[contractType] || '' })
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Get contract type on Hazama Network
 *
 * @private
 * @param {object} params parameters for getting contract type
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {string} params.walletAddress wallet address
 * @param {string} params.projectId project id
 * @param {string} params.contractAddress contract address
 * @return {Promise<ContractTypeResult>} promise
 */
async function getContractTypeOnHazamaNetwork({
  host = HOST,
  walletAddress,
  projectId,
  contractAddress,
}) {
  return new Promise(async (resolve, reject) => {
    try {
      if (!host) throw new Error(messages.HOST_UNDEFINED)

      if (!walletAddress) throw new Error(messages.WALLET_ADDRESS_UNDEFINED)

      if (!projectId) throw new Error(messages.PROJECT_ID_UNDEFINED)

      if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

      lib.connect(host, async () => {
        try {
          const path = generatePath({
            walletAddress,
            projectId,
            contractAddress,
          })
          const data = await lib.get(`${path}.meta`, null, false)
          resolve({ type: StandardContractType?.[data?.Type] || '' })
        } catch (e) {
          reject(e)
        }
      })
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Get contract type
 * @function
 * @name getContractType
 * @memberof hazama-lib/Others
 *
 * @example <caption>[Get contract type on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const contractAddress = '0x...'
 * const clientKey = 'dummy'
 *
 * const contractType = await hazlib.getContractType({ network, contractAddress, clientKey })
 * ...
 *
 * @example <caption>[Get contract type on HAZAMA network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'hazama'
 * const host = 'wss://n.hazbase.com'
 * const walletAddress = '0x...'
 * const projectId = 'project id'
 * const contractAddress = '0x...'
 * const clientKey = 'dummy'
 *
 * const contractType = await hazlib.getContractType({ network, host, walletAddress, projectId, contractAddress, clientKey })
 * ...
 *
 * @param {object} params parameters for getting contract type
 * @param {string} params.network the chain name or rpc url
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {string} params.contractAddress contract address
 * @param {string} [params.walletAddress] wallet address
 * @param {string} [params.projectId] project id
 * @param {string} [params.clientKey] client key
 * @return {ContractTypeResult}
 */
lib.getContractType = async ({
  network,
  host = HOST,
  contractAddress,
  projectId,
  walletAddress,
  clientKey,
  isCount = true,
}) => {
  const caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 15,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (!network) throw new Error(messages.NETWORK_UNDEFINED)

    if (network === 'hazama' && !host) throw new Error(messages.HOST_UNDEFINED)

    if (network === 'hazama' && !walletAddress)
      throw new Error(messages.WALLET_ADDRESS_UNDEFINED)

    if (network === 'hazama' && !projectId)
      throw new Error(messages.PROJECT_ID_UNDEFINED)

    if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

    const result =
      network !== 'hazama'
        ? await getContractTypeOnEVMNetwork({ network, contractAddress })
        : await getContractTypeOnHazamaNetwork({
            host,
            walletAddress,
            projectId,
            contractAddress,
          })

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 15,
      walletAddress: caller,
      status: RequestTransactionStatus.SUCCEEDED,
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 15,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * Get Wallet on HAZAMA network
 *
 * @private
 * @param {object} params parameters
 * @param {string} params.walletAddress wallet address
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @returns {Promise<WalletInfo>} wallet
 */
async function getWalletOnHazamaNetwork({ walletAddress, host = HOST }) {
  return new Promise(async (resolve, reject) => {
    try {
      if (!host) throw new Error(messages.HOST_UNDEFINED)

      if (!walletAddress) throw new Error(messages.WALLET_ADDRESS_UNDEFINED)

      lib.connect(host, async () => {
        try {
          const wallet = await lib.get(`/_/addr.${walletAddress}`, null, false)
          resolve(wallet)
        } catch (e) {
          reject(e)
        }
      })
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Get wallet (HAZAMA network only)
 *
 * @function
 * @name getWallet
 * @memberof hazama-lib/Others
 *
 * @example <caption>Get wallet on HAZAMA network</caption>
 *
 * import hazlib from 'hazama-lib'
 * ...
 * const walletAddress = '0x...'
 * const host = 'wss://n.hazbase.com'
 * const clientKey = 'dummy'
 *
 * const wallet = await hazlib.getWallet({walletAddress, host, clientKey})
 * ...
 * @param {object} params parameters
 * @param {string} params.walletAddress wallet address
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {string} [params.clientKey] client key
 * @param {boolean} [params.clientKey] client key
 * @returns {WalletInfo} wallet
 */
lib.getWallet = async ({
  walletAddress,
  host = HOST,
  clientKey,
  isCount = true,
}) => {
  const caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 21,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (!host) throw new Error(messages.HOST_UNDEFINED)

    if (!walletAddress) throw new Error(messages.WALLET_ADDRESS_UNDEFINED)

    const result = await getWalletOnHazamaNetwork({ walletAddress, host })

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 21,
      walletAddress: caller,
      status: RequestTransactionStatus.SUCCEEDED,
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 21,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * Create project on HAZAMA network
 *
 * @private
 * @param {object} params parameters
 * @param {string} params.walletAddress wallet address
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {string} params.name project name
 * @param {string} params.network project network
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions.<br/>If wif is not set, Signer is required.
 * @param {string} [params.wif] wallet import format.<br/>If Signer is not set, wif is required.
 * @returns {Promise<string>} transaction id
 */
async function createProjectOnHazamaNetwork({
  walletAddress,
  host = HOST,
  wif,
  signer,
  name,
  network,
}) {
  return new Promise(async (resolve, reject) => {
    try {
      if (!network) throw new Error(messages.NETWORK_UNDEFINED)

      if (!host) throw new Error(messages.HOST_UNDEFINED)

      if (!signer && !wif) throw new Error(messages.SIGNER_WIF_UNDEFINED)

      if (!walletAddress) throw new Error(messages.WALLET_ADDRESS_UNDEFINED)

      if (!name) throw new Error(messages.PROJECT_NAME_UNDEFINED)

      const uuid = () => uuidv4().replace(/-/g, '')
      lib.connect(host, async () => {
        try {
          const path = await lib.get(
            `/hazama/${walletAddress}/projects.meta`,
            null,
            false,
          )
          if (path == null) {
            const asm = await lib.template(
              'dao_projects',
              {
                path: `/hazama/${walletAddress}/projects`,
                name: 'HAZAMA BASE Projects',
                version: '0.0.1',
                author: walletAddress,
              },
              null,
              false,
            )
            if (asm) {
              const txid = await compileAndBroadcast({
                host,
                asm,
                signer,
                wif,
              })
            }
          }

          const asm = await lib.template(
            'dao_projects_create',
            {
              path: `/hazama/${walletAddress}/projects`,
              id: uuid(),
              name,
              network,
            },
            null,
            false,
          )
          if (asm) {
            const txid = await compileAndBroadcast({ host, asm, signer, wif })
            resolve(txid)
          }
        } catch (e) {
          reject(e)
        }
      })
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Create project
 *
 * @function
 * @name createProject
 * @memberof hazama-lib/Others
 *
 * @example
 *
 * import hazlib from 'hazama-lib'
 * ...
 * const walletAddress = '0x...'
 * const host = 'wss://n.hazbase.com'
 * const wif = 'dummny'
 * const signer = <signer object>
 * const name = 'project name'
 * const network = 'hazama'
 * const clientKey = 'dummy'
 *
 * const transactionId = await hazlib.createProject({walletAddress, host, wif, signer, name, network, clientKey})
 * ...
 *
 * @param {object} params parameters
 * @param {string} params.walletAddress wallet address
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {string} params.name project name
 * @param {string} params.network project network
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions.<br/>If wif is not set, Signer is required.
 * @param {string} [params.wif] wallet import format.<br/>If Signer is not set, wif is required.
 * @param {string} [params.clientKey] client key
 * @returns {Promise<string>} transaction id
 */
lib.createProject = async ({
  walletAddress,
  host = HOST,
  wif,
  signer,
  name,
  network,
  clientKey,
  isCount = true,
}) => {
  let caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 11,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (!network) throw new Error(messages.NETWORK_UNDEFINED)

    if (!host) throw new Error(messages.HOST_UNDEFINED)

    if (!signer && !wif) throw new Error(messages.SIGNER_WIF_UNDEFINED)

    if (!walletAddress) throw new Error(messages.WALLET_ADDRESS_UNDEFINED)

    if (!name) throw new Error(messages.PROJECT_NAME_UNDEFINED)

    caller = await getCaller({ network, host, signer, wif })

    const result = await createProjectOnHazamaNetwork({
      walletAddress,
      host,
      wif,
      signer,
      name,
      network,
    })
    const projectId = await getProjectId({
      walletAddress: caller,
      host,
      transactionHash: result,
    })

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 11,
      walletAddress: caller,
      status: RequestTransactionStatus.SUCCEEDED,
      transactionHash: result,
      apiRequests: [
        {
          apiCallStatus: ApiRequestStatus.NOT_CALLED_YET,
          apiBody: JSON.stringify({
            walletAddress: caller,
            projectId,
          }),
          apiURI: '/api/app/project',
          apiMethod: 'post',
        },
      ],
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 11,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * Search projects on HAZAMA network
 *
 * @private
 * @param {object} params parameters
 * @param {string} params.walletAddress wallet address
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @returns {Promise<Project[]>} list of projects
 */
async function searchProjectsOnHazamaNetwork({ walletAddress, host = HOST }) {
  return new Promise(async (resolve, reject) => {
    try {
      if (!walletAddress) throw new Error(messages.WALLET_ADDRESS_UNDEFINED)

      if (!host) throw new Error(messages.HOST_UNDEFINED)

      lib.connect(host, async () => {
        const projects = []
        const iter = await lib.iterate(
          {
            gte: `/hazama/${walletAddress}/projects/`,
            lte: `/hazama/${walletAddress}/projects/~`,
            match: `/hazama/${walletAddress}/projects/[^/]+$`,
          },
          10,
        )

        let chunk = await iter()
        while (chunk !== null) {
          for (let i = 0; i < chunk.length; i++) {
            const { key, value } = chunk[i]
            projects.push([key, value])
          }

          chunk = await iter()
        }
        resolve(projects)
      })
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Search projects (HAZAMA network only)
 *
 * @function
 * @name searchProjects
 * @memberof hazama-lib/Others
 *
 * @example <caption>Search projects on HAZAMA network</caption>
 *
 * import hazlib from 'hazama-lib'
 * ...
 * const walletAddress = '0x...'
 * const host = 'wss://n.hazbase.com'
 * const clientKey = 'dummy'
 *
 * const projects = await hazlib.searchProjects({walletAddress, host, clientKey})
 * ...
 *
 * @param {object} params parameters
 * @param {string} params.walletAddress wallet address
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {string} [params.clientKey] client key
 * @returns {Project[]} list of projects
 */
lib.searchProjects = async ({
  walletAddress,
  host = HOST,
  clientKey,
  isCount = true,
}) => {
  const caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 26,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (!walletAddress) throw new Error(messages.WALLET_ADDRESS_UNDEFINED)

    if (!host) throw new Error(messages.HOST_UNDEFINED)

    const result = await searchProjectsOnHazamaNetwork({ walletAddress, host })

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 26,
      walletAddress: caller,
      status: RequestTransactionStatus.SUCCEEDED,
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 26,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * Get project on HAZAMA network
 *
 * @private
 * @param {object} params parameters
 * @param {string} params.walletAddress wallet address
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {string} params.projectId project id
 * @returns {Promise<Project>} project
 */
async function getProjectOnHazamaNetwork({
  walletAddress,
  host = HOST,
  projectId,
}) {
  return new Promise(async (resolve, reject) => {
    try {
      if (!walletAddress) throw new Error(messages.WALLET_ADDRESS_UNDEFINED)

      if (!host) throw new Error(messages.HOST_UNDEFINED)

      if (!projectId) throw new Error(messages.PROJECT_ID_UNDEFINED)

      lib.connect(host, async () => {
        try {
          const project = await lib.get(
            `/hazama/${walletAddress}/projects/${projectId}`,
            null,
            false,
          )
          resolve(project)
        } catch (e) {
          reject(e)
        }
      })
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Get project by project id (HAZAMA network only)
 *
 * @function
 * @name getProject
 * @memberof hazama-lib/Others
 *
 * @example <caption>Get project on HAZAMA network</caption>
 *
 * import hazlib from 'hazama-lib'
 * ...
 * const walletAddress = '0x...'
 * const host = 'wss://n.hazbase.com'
 * const projectId = 'id'
 * const clientKey = 'dummy'
 *
 * const project = await hazlib.getProject({walletAddress, host, projectId, clientKey})
 * ...
 *
 * @param {object} params parameters
 * @param {string} params.walletAddress wallet address
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {string} params.projectId project id
 * @param {string} [params.clientKey] client key
 * @returns {Project} project
 */
lib.getProject = async ({
  walletAddress,
  host = HOST,
  projectId,
  clientKey,
  isCount = true,
}) => {
  const caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 17,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (!walletAddress) throw new Error(messages.WALLET_ADDRESS_UNDEFINED)

    if (!host) throw new Error(messages.HOST_UNDEFINED)

    if (!projectId) throw new Error(messages.PROJECT_ID_UNDEFINED)

    const result = await getProjectOnHazamaNetwork({
      walletAddress,
      host,
      projectId,
    })

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 17,
      walletAddress: caller,
      status: RequestTransactionStatus.SUCCEEDED,
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 17,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * Search contracts on EVM network
 *
 * @private
 *
 * @param {Object} params parameters
 * @param {string} params.network network
 * @param {string} params.walletAddress wallet address
 * @param {string} params.projectId project id
 * @param {any} [params.queries={}] queries
 * @returns {Promise<Contract[]>} list of contract
 */
async function searchContractsOnEVMNetwork({
  network,
  walletAddress,
  projectId,
  queries = {},
}) {
  if (!network) throw new Error(messages.NETWORK_UNDEFINED)

  if (!walletAddress) throw new Error(messages.WALLET_ADDRESS_UNDEFINED)

  if (!projectId) throw new Error(messages.PROJECT_ID_UNDEFINED)

  const sdk = new ThirdwebSDK(network, {
    ...(isNil(thirdwebSecretKey) &&
    (!isNil(thirdwebClientId) || !isNil(thirdwebClientIdFromAPI))
      ? { clientId: thirdwebClientId || thirdwebClientIdFromAPI }
      : {}),
    ...(!isNil(thirdwebSecretKey) ? { secretKey: thirdwebSecretKey } : {}),
  })

  const contractList = await sdk.getContractList(walletAddress)

  const promises = contractList.map((cntr) => {
    return new Promise(async (resolve, reject) => {
      try {
        const metadata = await cntr?.metadata()

        if (!metadata?.description) {
          return resolve(null)
        }

        const matches = metadata.description.match(/hazbase:([a-zA-Z0-9]{32})/)

        const contractType = await cntr.contractType()
        if (
          !(
            matches &&
            matches[1] === projectId &&
            Object.keys(StandardContractType).includes(contractType)
          )
        ) {
          return resolve(null)
        }

        if (queries?.search) {
          if (!new RegExp(`${queries.search}`, 'gi').test(metadata.name)) {
            return resolve(null)
          }
        }

        let transferable = true

        const contract = await sdk.getContract(cntr.address, contractType)

        const roles = await contract.roles.get('transfer')
        transferable = roles?.includes(ethers.constants.AddressZero)

        return resolve({
          Address: walletAddress,
          Author: walletAddress,
          ContractHash: cntr.address,
          ContractVersion: 'v1',
          Id: cntr.address,
          Name: metadata.name,
          Type: StandardContractType[contractType],
          ContractTypes: contractType,
          Transferable: transferable,
        })
      } catch (e) {
        reject(e)
      }
    })
  })
  const contracts = await Promise.all(promises)

  return rejectNull(contracts)
}

/**
 * Search contracts on HAZAMA network
 *
 * @private
 *
 * @param {Object} params parameters
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {string} params.walletAddress wallet address
 * @param {string} params.projectId project id
 * @param {any} [params.queries={}] queries
 * @returns {Promise<Contract[]>} list of contract
 */
async function searchContractsOnHazamaNetwork({
  host = HOST,
  walletAddress,
  projectId,
  queries = {},
}) {
  return new Promise(async (resolve, reject) => {
    try {
      if (!host) throw new Error(messages.HOST_UNDEFINED)

      if (!walletAddress) throw new Error(messages.WALLET_ADDRESS_UNDEFINED)

      if (!projectId) throw new Error(messages.PROJECT_ID_UNDEFINED)

      lib.connect(host, async () => {
        const ret = []
        const iter = await lib.iterate(
          {
            gt: `/hazama/${walletAddress}/projects/${projectId}/contracts/`,
            lt: `/hazama/${walletAddress}/projects/${projectId}/contracts/~`,
            match: `/hazama/${walletAddress}/projects/${projectId}/contracts/([^/]+).meta$`,
          },
          10,
        )
        let chunk = await iter()
        while (chunk !== null) {
          for (let i = 0; i < chunk.length; i++) {
            const { key, value } = chunk[i]
            if (queries.search) {
              if (!new RegExp(`${queries.search}`, 'gi').test(value.Name))
                continue
            }
            ret.push([key, value])
          }

          chunk = await iter()
        }

        const wallet = await lib.getWallet({
          walletAddress,
          host,
          isCount: false,
        })

        const data = ret.map((cntr) => {
          const contractPath = cntr[0].replace(/\.meta/, '')
          const txid = wallet[contractPath]
          if (txid) {
            const timestamp = wallet.txs[txid]
            cntr[1].Timestamp = timestamp
          }

          cntr[1].Id = contractPath.split('/').pop()
          cntr[1].ContractHash = `0x${cntr[1].ContractHash || ''}`
          cntr[1].Type = StandardContractType[cntr[1].Type]
          return cntr[1]
        })

        data.sort((a, b) => b.Timestamp - a.Timestamp)

        resolve(data)
      })
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Search contracts
 *
 * @function
 * @name searchContracts
 * @memberof hazama-lib/Others
 *
 * @example <caption>Search contracts on EVM network</caption>
 *
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const walletAddress = '0x...'
 * const projectId = 'project id'
 * const query = 'search=contractName'
 * const clientKey = 'dummy'
 *
 * const contracts = await hazlib.searchContracts({network, walletAddress, projectId, query, clientKey})
 * ...
 *
 *
 * @example <caption>Search contracts on HAZAMA network</caption>
 *
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'hazama'
 * const host = 'wss://n.hazbase.com'
 * const walletAddress = '0x...'
 * const projectId = 'project id'
 * const query = 'search=contractName'
 * const clientKey = 'dummy'
 *
 * const contracts = await hazlib.searchContracts({network, host, walletAddress, projectId, query, clientKey})
 * ...
 *
 * @param {Object} params parameters
 * @param {string} params.network network
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {string} params.walletAddress wallet address
 * @param {string} params.projectId project id
 * @param {string} [params.query=''] query
 * @param {string} [params.clientKey] client key
 * @returns {Contract[]} list of contract
 */
lib.searchContracts = async ({
  network,
  host = HOST,
  walletAddress,
  projectId,
  query = '',
  clientKey,
  isCount = true,
}) => {
  const caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 25,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (!network) throw new Error(messages.NETWORK_UNDEFINED)

    if (network === 'hazama' && !host) throw new Error(messages.HOST_UNDEFINED)

    if (!walletAddress) throw new Error(messages.WALLET_ADDRESS_UNDEFINED)

    if (!projectId) throw new Error(messages.PROJECT_ID_UNDEFINED)

    const queries = parse(query)

    const result =
      network !== 'hazama'
        ? await searchContractsOnEVMNetwork({
            network,
            walletAddress,
            projectId,
            queries,
          })
        : await searchContractsOnHazamaNetwork({
            host,
            walletAddress,
            projectId,
            queries,
          })

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 25,
      walletAddress: caller,
      status: RequestTransactionStatus.SUCCEEDED,
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 25,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * Check contract is transferable on EVM Network
 *
 * @private
 * @param {object} params parameters for getting contract type
 * @param {string} params.network the chain name or rpc url
 * @param {string} params.contractAddress the contract address
 * @return {Promise<boolean>} promise
 */
async function isTransferableContractOnEVMNetwork({
  network,
  contractAddress,
}) {
  return new Promise(async (resolve, reject) => {
    try {
      if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

      const sdk = new ThirdwebSDK(network, {
        ...(isNil(thirdwebSecretKey) &&
        (!isNil(thirdwebClientId) || !isNil(thirdwebClientIdFromAPI))
          ? { clientId: thirdwebClientId || thirdwebClientIdFromAPI }
          : {}),
        ...(!isNil(thirdwebSecretKey) ? { secretKey: thirdwebSecretKey } : {}),
      })
      const contractType = await sdk.resolveContractType(contractAddress)

      let isTransferable = false
      let transferAddresses = []
      const contract = await sdk.getContract(contractAddress, contractType)
      switch (contractType) {
        case ContractType.DropERC1155:
        case ContractType.DropERC721:
          transferAddresses = (await contract?.roles?.get(Role.TRANSFER)) || []
          if (transferAddresses.includes(ethers.constants.AddressZero))
            isTransferable = true
          break
        default:
          throw new Error(messages.CONTRACT_TYPE_IS_INVALID)
      }

      resolve(isTransferable)
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Check contract is transferable on Hazama Network
 *
 * @private
 * @param {object} params parameters for getting contract type
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {string} params.contractAddress the contract address
 * @return {Promise<boolean>} promise
 */
async function isTransferableContractOnHazamaNetwork({
  host = HOST,
  contractAddress,
}) {
  return new Promise(async (resolve, reject) => {
    try {
      if (!host) throw new Error(messages.HOST_UNDEFINED)

      if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

      lib.connect(host, async () => {
        try {
          const path = await lib.get(
            `/_/contract.${contractAddress}.namespace`,
            null,
            false,
          )
          if (path === null) reject(new Error(messages.TOKEN_NOTFOUND))
          const meta = await lib.get(`${path}.meta`, null, false)
          const isTransferable = meta.Type === ContractType.NFT
          resolve(isTransferable)
        } catch (e) {
          reject(e)
        }
      })
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Is transferable contract ?
 * @function
 * @name isTransferableContract
 * @memberof hazama-lib/Others
 *
 * @example <caption>[Check contract is transferable on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const contractAddress = '0x...'
 * const clientKey = 'dummy'
 *
 * const isTransferable = await hazlib.isTransferableContract({ network, contractAddress, clientKey })
 * ...
 *
 * @example <caption>[Check contract is transferable on HAZAMA network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'hazama'
 * const host = 'wss://n.hazbase.com'
 * const contractAddress = '29...'
 * const clientKey = 'dummy'
 *
 * const isTransferable = await hazlib.isTransferableContract({ network, host, contractAddress, clientKey })
 * ...
 *
 * @param {object} params parameters for getting contract type
 * @param {string} params.network the chain name or rpc url
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {string} params.contractAddress contract address
 * @param {string} [params.clientKey] client key
 * @return {boolean}
 */
lib.isTransferableContract = async ({
  network,
  host = HOST,
  contractAddress,
  clientKey,
  isCount = true,
}) => {
  const caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 23,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (!network) throw new Error(messages.NETWORK_UNDEFINED)

    if (network === 'hazama' && !host) throw new Error(messages.HOST_UNDEFINED)

    if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

    const result =
      network !== 'hazama'
        ? await isTransferableContractOnEVMNetwork({ network, contractAddress })
        : await isTransferableContractOnHazamaNetwork({
            host,
            contractAddress,
          })

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 23,
      walletAddress: caller,
      status: RequestTransactionStatus.SUCCEEDED,
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 23,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * Get contract on EVM network
 *
 * @private
 *
 * @param {Object} params parameters
 * @param {string} params.network network
 * @param {string} params.walletAddress wallet address
 * @param {string} params.contractAddress contract address
 * @returns {Promise<Contract>} contract
 */
async function getContractOnEVMNetwork({
  network,
  walletAddress,
  contractAddress,
}) {
  if (!network) throw new Error(messages.NETWORK_UNDEFINED)

  if (!walletAddress) throw new Error(messages.WALLET_ADDRESS_UNDEFINED)

  if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

  const sdk = new ThirdwebSDK(network, {
    ...(isNil(thirdwebSecretKey) &&
    (!isNil(thirdwebClientId) || !isNil(thirdwebClientIdFromAPI))
      ? { clientId: thirdwebClientId || thirdwebClientIdFromAPI }
      : {}),
    ...(!isNil(thirdwebSecretKey) ? { secretKey: thirdwebSecretKey } : {}),
  })

  const contractType = await resolveContractTypeWithCustom({
    sdk,
    contractAddress,
  })
  const contract = await sdk.getContract(contractAddress, contractType)
  const metadata = await contract?.metadata?.get()

  return {
    Address: contractAddress,
    Author: walletAddress,
    ContractVersion: 'v1',
    Name: metadata?.name,
    Type: StandardContractType[contractType],
    Version: '1',
  }
}

/**
 * Get contract on HAZAMA network
 *
 * @private
 *
 * @param {Object} params parameters
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {string} params.walletAddress wallet address
 * @param {string} params.projectId project id
 * @param {string} params.contractAddress contract address
 * @returns {Promise<Contract>} contract
 */
async function getContractOnHazamaNetwork({
  host = HOST,
  walletAddress,
  projectId,
  contractAddress,
}) {
  return new Promise(async (resolve, reject) => {
    try {
      if (!host) throw new Error(messages.HOST_UNDEFINED)

      if (!walletAddress) throw new Error(messages.WALLET_ADDRESS_UNDEFINED)

      if (!projectId) throw new Error(messages.PROJECT_ID_UNDEFINED)

      if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

      lib.connect(host, async () => {
        try {
          const path = generatePath({
            walletAddress,
            projectId,
            contractAddress,
          })
          const data = await lib.get(`${path}.meta`, null, false)
          if (data === null) reject(messages.CONTRACT_ADDRESS_INVALID)

          data.Type = StandardContractType[data.Type]

          resolve(data)
        } catch (e) {
          reject(e)
        }
      })
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Get contract by contract address
 *
 * @function
 * @name getContract
 * @memberof hazama-lib/Others
 *
 * @example <caption>Get contract on EVM network</caption>
 *
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const walletAddress = '0x...'
 * const contractAddress = '0x...'
 * const clientKey = 'dummy'
 *
 * const contract = await hazlib.getContract({network, walletAddress, contractAddress, clientKey})
 * ...
 *
 *
 * @example <caption>Get contract on HAZAMA network</caption>
 *
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'hazama'
 * const host = 'wss://n.hazbase.com'
 * const walletAddress = '0x...'
 * const projectId = 'project id'
 * const contractAddress = '0x...'
 * const clientKey = 'dummy'
 *
 * const contract = await hazlib.getContract({network, host, walletAddress, projectId, contractAddress, clientKey})
 * ...
 *
 * @param {Object} params parameters
 * @param {string} params.network network
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {string} params.walletAddress wallet address
 * @param {string} [params.projectId] project id
 * @param {string} params.contractAddress contract address
 * @param {string} [params.clientKey] client key
 * @returns {Contract} contract
 */
lib.getContract = async ({
  network,
  host,
  walletAddress,
  projectId,
  contractAddress,
  clientKey,
  isCount = true,
}) => {
  const caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 14,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (!network) throw new Error(messages.NETWORK_UNDEFINED)

    if (network === 'hazama' && !host) throw new Error(messages.HOST_UNDEFINED)

    if (!walletAddress) throw new Error(messages.WALLET_ADDRESS_UNDEFINED)

    if (network === 'hazama' && !projectId)
      throw new Error(messages.PROJECT_ID_UNDEFINED)

    if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

    const result =
      network !== 'hazama'
        ? await getContractOnEVMNetwork({
            network,
            walletAddress,
            contractAddress,
          })
        : await getContractOnHazamaNetwork({
            host,
            walletAddress,
            projectId,
            contractAddress,
          })

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 14,
      walletAddress: caller,
      status: RequestTransactionStatus.SUCCEEDED,
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 14,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * Search tokens on EVM network
 *
 * @private
 *
 * @param {Object} params parameters
 * @param {string} params.network network
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions.<br/>If private key is not set, Signer is required.
 * @param {string} [params.privateKey] wallet's private key.<br/>If Signer is not set, private key is required.
 * @param {string} params.contractAddress contract address
 * @param {string} [params.walletAddress] wallet address
 * @returns {Promise<EVMToken[]>} list of tokens
 */
async function searchTokensOnEVMNetwork({
  network,
  signer,
  privateKey,
  contractAddress,
  walletAddress,
  isMultipleClaimConditionsSupported = false,
}) {
  if (!network) throw new Error(messages.NETWORK_UNDEFINED)

  if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

  const options = {
    readonlySettings: {
      rpcUrl: getRpcUrl(network),
    },
    ...(isArrayFull(gatewayUrls) ? { gatewayUrls } : {}),
    ...(isNil(thirdwebSecretKey) &&
    (!isNil(thirdwebClientId) || !isNil(thirdwebClientIdFromAPI))
      ? { clientId: thirdwebClientId || thirdwebClientIdFromAPI }
      : {}),
    ...(!isNil(thirdwebSecretKey) ? { secretKey: thirdwebSecretKey } : {}),
  }
  let sdk = new ThirdwebSDK(network, options)
  const contractType = await resolveContractTypeWithCustom({
    sdk,
    contractAddress,
  })

  if (contractType !== ContractType.VoteERC20 && !signer && !privateKey)
    throw new Error(messages.SIGNER_PRIVATEKEY_UNDEFINED)

  if (signer) {
    sdk.updateSignerOrProvider(signer)
  }

  if (privateKey) {
    sdk = new ThirdwebSDK(
      new ethers.Wallet(
        privateKey,
        ethers.getDefaultProvider(getRpcUrl(network)),
      ),
      options,
    )
  }

  const contract = await sdk.getContract(contractAddress, contractType)

  let tokens = []
  let transferable = true

  if (contractType === ContractType.TokenERC20) {
    const [meta, totalSupply, owned] = await Promise.all([
      contract.get(),
      contract.totalSupply(),
      contract.balance(),
    ])
    return [
      {
        ...meta,
        Transferable: transferable,
        totalSupply: `${Number(totalSupply.displayValue)} ${meta.symbol}`,
        owned: Number(owned.displayValue),
      },
    ]
  }

  let list
  if (contractType === ContractType.DropERC721) {
    const [claimed, unclaimed, roles] = await Promise.all([
      contract.getAllClaimed(),
      contract.getAllUnclaimed(),
      contract.roles.get(Role.TRANSFER),
    ])
    const claimedSupply = claimed?.length
    const unclaimedSupply = unclaimed?.length
    list = claimed.concat(unclaimed).map((d) => ({
      metadata: d.metadata || d,
      owner: d.owner || '',
      claimedSupply,
      unclaimedSupply,
      totalSupply: claimedSupply + unclaimedSupply,
    }))
    transferable = roles.includes(ethers.constants.AddressZero)
  } else if (contractType === ContractType.Multiwrap) {
    list = await lib.getWrappedBundles({
      network,
      contractAddress,
      isCount: false,
    })
  } else if (
    contractType === ContractType.TokenERC721 ||
    contractType === ContractType.TokenERC1155
  ) {
    list = await contract.getAll()
  } else if (contractType === ContractType.DropERC1155) {
    list = await contract.getAll()
    const roles = await contract.roles.get(Role.TRANSFER)
    transferable = roles.includes(ethers.constants.AddressZero)
  } else if (contractType === ContractType.VoteERC20) {
    async function getAll() {
      return Promise.all(
        (await contract.contractWrapper.readContract.getAllProposals()).map(
          async (data, index) => {
            const { title, description } =
              data.description?.length > 0
                ? JSON.parse(data.description.replace(/\"/gi, '"')) // eslint-disable-line no-useless-escape
                : {}

            const proposalId = data.proposalId.toString()
            const choices = data.choices

            const [tokenAddress, hasVoted, votes, state] = await Promise.all([
              contract.call('token'),
              walletAddress
                ? await contract.call('hasVoted', [proposalId, walletAddress])
                : true,
              getProposalVotes({
                proposalId,
                choices,
                contract,
              }),
              contract.contractWrapper.readContract.state(proposalId),
            ])
            return {
              proposalId,
              proposer: data.proposer,
              title,
              description,
              startTime: Number(data.startTime.toString()),
              endTime: Number(data.endTime.toString()),
              state,
              votes: votes,
              executions: data[3].map((c, i) => ({
                toAddress: data.targets[i],
                nativeTokenValue: c,
                transactionData: data.calldatas[i],
              })),
              hasVoted,
              tokenAddress,
              index,
            }
          },
        ),
      )
    }
    list = await getAll()
  }

  tokens = await Promise.all(
    list.map((cntr) => {
      return new Promise(async (resolve, reject) => {
        let token = {
          type: StandardContractType[contractType],
          contractType,
          contractId: contractAddress,
        }
        if (contractType === ContractType.VoteERC20) {
          token.id = cntr.proposalId
          token.owner = cntr.proposer
          token.state = cntr.state
          token.startTime = cntr.startTime
          token.endTime = cntr.endTime
          token.votes = cntr.votes
          token.executions = cntr.executions
          token.title = cntr.title
          token.description = cntr.description
          token.hasVoted = cntr.hasVoted
          token.tokenAddress = cntr.tokenAddress
          token.index = cntr.index
        } else {
          token.id = Number(cntr.metadata.id)
          token.name = cntr.metadata.name
          token.owner = cntr.owner
          token.description = cntr?.metadata?.description
            ?.replace(/\\n/g, ' ')
            .replace(/\n/g, ' ')
          token.image = cntr.metadata.image || cntr.metadata.animation_url
          token.owner_index = 0
          token.Transferable = transferable
        }
        if (contractType === ContractType.TokenERC1155) {
          try {
            const balance = await contract.balance(Number(cntr.metadata.id))
            token.owned = Number(balance)
            token.decimals = Number(balance)
            token.supply = parseInt(cntr.supply, 10)

            const keyTokens = await contract.call('currentAccessKeyTokens')
            const burnTokens = {
              burnContractAddress: keyTokens[0],
              burnTokens: keyTokens[1].map((t) => ({
                tokenId: Number(t[0].toString()),
                quantity: Number(t[1].toString()),
              })),
            }
            token.burnTokens = burnTokens
          } catch (e) {
            token.burnTokens = {
              burnContractAddress: null,
              burnTokens: [{ tokenId: null, quantity: null }],
            }
          }
        }
        if (StandardContractType[contractType] === HazBaseContractType.DROP) {
          const balance = await contract.balance(Number(cntr.metadata.id))
          token.owned = Number(balance)
          token.decimals = Number(balance)
          token.supply = parseInt(cntr.supply, 10)
          if (contractType === ContractType.DropERC721) {
            token.claimedSupply = cntr.claimedSupply
            token.unclaimedSupply = cntr.unclaimedSupply
            token.totalSupply = cntr.totalSupply
          }
          try {
            const conditions =
              contractType === ContractType.DropERC1155
                ? await contract.claimConditions.getAll(token.id, {
                    withAllowList: true,
                  })
                : await contract.claimConditions.getAll({
                    withAllowList: true,
                  })

            if (conditions && conditions.length > 0) {
              token.conditions = []
              conditions.forEach((condition) => {
                const {
                  startTime,
                  availableSupply,
                  maxClaimableSupply,
                  snapshot,
                } = condition
                token.conditions.push({
                  startTime: moment(startTime).format('YYYY/MM/DD HH:mm'),
                  startTimeOrigin: startTime,
                  availableSupply:
                    availableSupply.length > 18 ? 'Unlimited' : availableSupply,
                  maxQuantity:
                    maxClaimableSupply === 'unlimited'
                      ? 'Unlimited'
                      : formatUnits(maxClaimableSupply, 0),
                  snapshot: snapshot || [],
                  claimData: { condition },
                })
              })

              let condition
              if (isMultipleClaimConditionsSupported) {
                try {
                  condition =
                    contractType === ContractType.DropERC1155
                      ? await contract.claimConditions.getActive(token.id, {
                          withAllowList: true,
                        })
                      : await contract.claimConditions.getActive({
                          withAllowList: true,
                        })
                } catch (e) {}
              }

              if (!isObjectFull(condition)) {
                condition = conditions?.[0]
              }

              const {
                startTime,
                availableSupply,
                maxClaimableSupply,
                snapshot,
              } = condition || {}
              token = {
                ...token,
                startTime,
                startTimeOrigin: startTime,
                availableSupply,
                maxQuantity: maxClaimableSupply,
                snapshot,
                claimData: { condition },
              }
              if (condition) {
                token.startTime = moment(startTime).format('YYYY/MM/DD HH:mm')
                token.availableSupply =
                  availableSupply?.length > 18 ? 'Unlimited' : availableSupply
                token.maxQuantity =
                  maxClaimableSupply === 'unlimited'
                    ? 'Unlimited'
                    : formatUnits(maxClaimableSupply, 0)
                token.snapshot = snapshot || []
              }
            }
          } catch (e) {
            throw new Error('')
          }
        }
        resolve(token)
      })
    }),
  )

  return tokens
}

/**
 * Search tokens on HAZAMA network
 *
 * @private
 *
 * @param {Object} params parameters
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {string} params.walletAddress wallet address
 * @param {string} params.contractAddress contract address
 * @returns {Promise<HazamaToken[]>} list of contracts
 */
async function searchTokensOnHazamaNetwork({
  host = HOST,
  walletAddress,
  contractAddress,
}) {
  return new Promise(async (resolve, reject) => {
    try {
      if (!host) throw new Error(messages.HOST_UNDEFINED)

      if (!walletAddress) throw new Error(messages.WALLET_ADDRESS_UNDEFINED)

      if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

      lib.connect(host, async () => {
        const path = await lib.get(
          `/_/contract.${contractAddress}.namespace`,
          null,
          false,
        )
        if (path === null) reject(new Error(messages.TOKEN_NOTFOUND))
        let meta = await lib.get(`${path}.meta`, null, false)

        let tokens = []
        if (meta.Type === ContractType.FUNGIBLE) {
          let [totalSupply, owned] = (
            await Promise.all([
              lib.get(`${path}/totalSupply`, null, false),
              lib.get(`${path}/${walletAddress}`, null, false),
            ])
          ).map((number) => Number(number || 0))

          meta = objectToLowerCase(meta)

          if (totalSupply > 0) {
            const bigNum = Number(totalSupply).toLocaleString('fullwide', {
              useGrouping: false,
            })
            totalSupply = Number(
              formatUnits((bigNum || 0).toString(), meta.info.decimals),
            ).toFixed()
          }
          if (owned > 0) {
            const bigNum = Number(owned).toLocaleString('fullwide', {
              useGrouping: false,
            })
            owned = Number(
              formatUnits((bigNum || 0).toString(), meta.info.decimals),
            ).toFixed()
          }

          tokens = [
            {
              ...meta,
              totalSupply: `${totalSupply} ${meta.symbol}`,
              owned,
              type: StandardContractType[meta.type],
              contractType: meta.type,
            },
          ]
        } else {
          const iter = await lib.iterate(
            {
              gt: `${path}/`,
              lt: `${path}/~`,
              match: `${path}/([^/|current_index]+)$`,
            },
            10,
          )

          const ret = []
          let chunk = await iter()
          while (chunk !== null) {
            for (let i = 0; i < chunk.length; i++) {
              const { key, value } = chunk[i]
              ret.push([key, value])
            }
            chunk = await iter()
          }

          tokens = await Promise.all(
            ret.map(async (cntr) => {
              cntr[1].id = cntr[0].split('/').pop()
              cntr[1].contractId = meta.ContractHash
              cntr[1].description = cntr[1].description
                ?.replace(/\\n/g, ' ')
                .replace(/\n/g, ' ')
              cntr[1].type = StandardContractType[meta.Type]
              cntr[1].contractType = meta.Type
              cntr[1].image = cntr[1].image.replace(
                'gateway.ipfscdn.io',
                'hzbs.infura-ipfs.io',
              )
              try {
                const condition = await lib.get(
                  `${cntr[0]}/conditions`,
                  null,
                  false,
                )

                if (condition) {
                  cntr[1].startTime = moment(
                    condition.startTime.toString(),
                  ).format('YYYY/MM/DD HH:mm')
                  cntr[1].startTimeOrigin = condition.startTime
                  cntr[1].availableSupply =
                    condition.availableSupply.length > 18
                      ? 'Unlimited'
                      : condition.availableSupply
                  const bigNum = Number(
                    condition.maxQuantity === 'unlimited'
                      ? 0
                      : condition.maxQuantity,
                  ).toLocaleString('fullwide', { useGrouping: false })
                  cntr[1].maxQuantity =
                    condition.maxQuantity === 'unlimited'
                      ? 'Unlimited'
                      : formatUnits(bigNum, 0)

                  const existsSnap = (
                    await iterate({
                      lib,
                      path: `${cntr[0]}/permissions/claim/`,
                      match: '0x[0-9a-zA-Z]+',
                    })
                  )?.map((snap) => ({
                    address: snap[0].split('/').pop(),
                    maxClaimable: Number(snap[1].maxClaimable),
                  }))
                  cntr[1].snapshot = existsSnap || []
                }
              } catch (e) {
                console.error(e)
              }
              return cntr[1]
            }),
          )
        }

        resolve(tokens)
      })
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Search tokens
 *
 * @function
 * @name searchTokens
 * @memberof hazama-lib/Others
 *
 * @example <caption>Search tokens on EVM network</caption>
 *
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const privateKey = 'dummy'
 * const contractAddress = '0x...'
 * const query = 'search=tokenName'
 * const clientKey = 'dummy'
 *
 * const tokens = await hazlib.searchTokens({network, privateKey, contractAddress, query, clientKey})
 * ...
 *
 *
 * @example <caption>Search tokens on HAZAMA network</caption>
 *
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'hazama'
 * const host = 'wss://n.hazbase.com'
 * const walletAddress = '0x...'
 * const contractAddress = '0x...'
 * const query = 'search=tokenName'
 * const clientKey = 'dummy'
 *
 * const tokens = await hazlib.searchTokens({network, host, walletAddress, contractAddress, query, clientKey})
 * ...
 *
 * @param {Object} params parameters
 * @param {string} params.network network
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions.<br/>If private key is not set, Signer is required.
 * @param {string} [params.privateKey] wallet's private key.<br/>If Signer is not set, private key is required.
 * @param {string} [params.walletAddress] wallet address
 * @param {string} params.contractAddress contract address
 * @param {string} [params.query=''] query
 * @param {string} [params.clientKey] client key
 * @returns {(EVMToken[]|HazamaToken[])} list of tokens
 */
lib.searchTokens = async ({
  network,
  host = HOST,
  signer,
  privateKey,
  walletAddress,
  contractAddress,
  query = '',
  clientKey,
  isCount = true,
  isMultipleClaimConditionsSupported = false,
}) => {
  let caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 27,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (!network) throw new Error(messages.NETWORK_UNDEFINED)

    if (network === 'hazama' && !host) throw new Error(messages.HOST_UNDEFINED)

    if (network === 'hazama' && !walletAddress)
      throw new Error(messages.WALLET_ADDRESS_UNDEFINED)

    if (network !== 'hazama') {
      const { type: contractType } = await lib.getContractType({
        network,
        contractAddress,
        isCount: false,
      })

      if (contractType !== ContractType.VoteERC20 && !signer && !privateKey)
        throw new Error(messages.SIGNER_PRIVATEKEY_UNDEFINED)
    }

    if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

    caller = await getCaller({ network, host, signer, privateKey })

    const queries = parse(query)
    const data =
      network !== 'hazama'
        ? await searchTokensOnEVMNetwork({
            network,
            signer,
            privateKey,
            contractAddress,
            walletAddress,
            isMultipleClaimConditionsSupported,
          })
        : await searchTokensOnHazamaNetwork({
            host,
            walletAddress,
            contractAddress,
          })

    const filteredData =
      'search' in queries
        ? data?.filter((token) =>
            new RegExp(`${queries.search}`, 'gi').test(token.name),
          ) || []
        : [...(data || [])]

    filteredData?.sort((a, b) => a.id - b.id)

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 27,
      walletAddress: caller,
      status: RequestTransactionStatus.SUCCEEDED,
      isCount,
    })

    return filteredData
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 27,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * Get claim metadata on EVM network
 *
 * @private
 * @param {object} params parameters
 * @param {string} [params.contractNetwork] network
 * @param {string} params.contractAddress contract address
 * @param {BigNumberish} params.tokenId the id of the token
 * @return {Promise<ClaimMetadata>} claim metadata
 */
async function getClaimMetadataOnEVMNetwork({
  contractNetwork,
  contractAddress,
  tokenId,
  walletAddress,
  isMultipleClaimConditionsSupported = false,
}) {
  return new Promise(async (resolve, reject) => {
    try {
      if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

      if (!tokenId && tokenId !== 0)
        throw new Error(messages.TOKEN_ID_UNDEFINED)

      if (!walletAddress) throw new Error(messages.WALLET_ADDRESS_UNDEFINED)

      const res = await resolveContractTypeWithMultiNetwork({
        contractNetwork,
        contractAddress,
      })

      if (res) {
        const { contractType, network, sdk } = res

        if (!!contractNetwork && contractNetwork !== network) {
          throw new Error(messages.TOKEN_NOTFOUND)
        }

        let claims, activeClaimCondition

        const contract = await sdk.getContract(contractAddress, contractType)
        switch (contractType) {
          case ContractType.DropERC1155:
            claims = await Promise.all([
              contract.get(tokenId),
              contract.claimConditions.getAll(tokenId, { withAllowList: true }),
              contract.claimConditions.canClaim(tokenId, 1, walletAddress),
              contract.claimConditions.getClaimIneligibilityReasons(
                tokenId,
                1,
                walletAddress,
              ),
            ])

            // get active claim condition
            try {
              activeClaimCondition = await contract.claimConditions.getActive(
                tokenId,
                {
                  withAllowList: true,
                },
              )
            } catch (e) {}
            claims.push(activeClaimCondition)
            break
          case ContractType.DropERC721:
            tokenId = 0
            claims = await Promise.all([
              contract.get(tokenId),
              contract.claimConditions.getAll({ withAllowList: true }),
              contract.claimConditions.canClaim(1, walletAddress),
              contract.claimConditions.getClaimIneligibilityReasons(
                1,
                walletAddress,
              ),
              undefined,
              contract.getAllClaimed(),
              contract.getAllUnclaimed(),
            ])

            // get active claim condition
            try {
              activeClaimCondition = await contract.claimConditions.getActive({
                withAllowList: true,
              })
            } catch (e) {}
            claims[4] = activeClaimCondition
            break
          case ContractType.TokenERC1155: {
            let metadata
            try {
              metadata = await contract.get(tokenId)
              const currentAccessKeyTokens = await contract.call(
                'currentAccessKeyTokens',
              )

              const burnTokens = {
                burnContractAddress: currentAccessKeyTokens[0],
                burnTokens: currentAccessKeyTokens[1].map((t) => ({
                  tokenId: Number(t[0].toString()),
                  quantity: Number(t[1].toString()),
                })),
              }

              const newContract = await sdk.getContract(
                burnTokens.burnContractAddress,
                ContractType.DropERC1155,
              )
              let hasBalance = true

              const burnTokensMetadata = await Promise.all(
                burnTokens.burnTokens.map(async (token) => {
                  const balance = await newContract.call('balanceOf', [
                    walletAddress,
                    token.tokenId,
                  ])
                  if (token.quantity > Number(balance)) hasBalance = false

                  const metadata = await newContract.get(token.tokenId)
                  return {
                    ...metadata,
                    hasBalance: !(token.quantity > Number(balance)),
                  }
                }),
              )

              claims = [
                metadata,
                [{ burnTokens, burnTokensMetadata }],
                hasBalance,
                [hasBalance ? '' : 'You have no enough access tokens'],
                { burnTokens, burnTokensMetadata },
              ]
            } catch (e) {
              if (e.message?.includes('No tokens has been set yet.')) {
                claims = [
                  metadata,
                  [],
                  false,
                  ['No tokens has been set yet.'],
                  undefined,
                ]
              }
            }
            break
          }
          default:
            throw new Error(messages.CONTRACT_TYPE_IS_NOT_SUPPORTED)
        }

        if (claims[0]?.metadata?.name.includes('Failed to load NFT metadata'))
          throw new Error(messages.TOKEN_NOTFOUND)

        claims[0].metadata = {
          ...claims[0].metadata,
          supply: parseInt(claims[0].supply, 10),
          id: Number(claims[0].metadata.id),
        }

        let owned = await contract.getOwned(walletAddress)
        // eslint-disable-next-line eqeqeq
        owned = owned.filter((nft) => nft?.metadata?.id == tokenId)
        if (owned?.length > 0 && contractType !== ContractType.TokenERC1155) {
          claims[3].unshift('You have already claimed the token.')
        }

        resolve({
          conditions: claims[1],
          metadata: claims[0].metadata,
          condition:
            isMultipleClaimConditionsSupported && isObjectFull(claims[4])
              ? claims[4]
              : claims[1]?.[0],
          canClaim: claims[2],
          claimReason: claims[3].shift(),
          network,
          claimedSupply: claims?.[5]?.length || undefined,
          unclaimedSupply: claims?.[6]?.length || undefined,
          totalSupply:
            isArray(claims?.[5]) && isArray(claims?.[6])
              ? claims?.[5]?.length + claims?.[6]?.length
              : undefined,
        })
      }
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Get claim metadata on EVM network
 *
 * @private
 * @param {object} params parameters
 * @param {string} [params.network] network
 * @param {string} params.contractAddress contract address
 * @param {BigNumberish} params.tokenId the id of the token
 * @return {Promise<ClaimMetadata>} claim metadata
 */
async function getClaimMetadataOnEVMNetworkForHazBase({
  network,
  contractAddress,
  tokenId,
  walletAddress,
  isMultipleClaimConditionsSupported = false,
}) {
  return new Promise(async (resolve, reject) => {
    try {
      if (!network) throw new Error(messages.CONTRACT_NETWORK_UNDEFINED)

      if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

      if (!tokenId && tokenId !== 0)
        throw new Error(messages.TOKEN_ID_UNDEFINED)

      if (!walletAddress) throw new Error(messages.WALLET_ADDRESS_UNDEFINED)

      const sdk = new ThirdwebSDK(network, {
        ...(isNil(thirdwebSecretKey) &&
        (!isNil(thirdwebClientId) || !isNil(thirdwebClientIdFromAPI))
          ? { clientId: thirdwebClientId || thirdwebClientIdFromAPI }
          : {}),
        ...(!isNil(thirdwebSecretKey) ? { secretKey: thirdwebSecretKey } : {}),
      })

      const contractType = await resolveContractTypeWithCustom({
        sdk,
        contractAddress,
      })

      let claims, activeClaimCondition

      const contract = await sdk.getContract(contractAddress, contractType)

      switch (contractType) {
        case ContractType.DropERC1155:
          claims = await Promise.all([
            contract.get(tokenId),
            contract.claimConditions.getAll(tokenId, {
              withAllowList: true,
            }),
            contract.claimConditions.canClaim(tokenId, 1, walletAddress),
            contract.claimConditions.getClaimIneligibilityReasons(
              tokenId,
              1,
              walletAddress,
            ),
          ])

          // get active claim condition
          try {
            activeClaimCondition = await contract.claimConditions.getActive(
              tokenId,
              {
                withAllowList: true,
              },
            )
          } catch (e) {}
          claims.push(activeClaimCondition)
          break
        case ContractType.DropERC721:
          tokenId = 0
          claims = await Promise.all([
            contract.get(tokenId),
            contract.claimConditions.getAll({
              withAllowList: true,
            }),
            contract.claimConditions.canClaim(1, walletAddress),
            contract.claimConditions.getClaimIneligibilityReasons(
              1,
              walletAddress,
            ),
            undefined,
            contract.getAllClaimed(),
            contract.getAllUnclaimed(),
          ])

          // get active claim condition
          try {
            activeClaimCondition = await contract.claimConditions.getActive({
              withAllowList: true,
            })
          } catch (e) {}
          claims[4] = activeClaimCondition
          break
        case ContractType.TokenERC1155: {
          let metadata
          try {
            metadata = await contract.get(tokenId)
            const currentAccessKeyTokens = await contract.call(
              'currentAccessKeyTokens',
            )

            const burnTokens = {
              burnContractAddress: currentAccessKeyTokens[0],
              burnTokens: currentAccessKeyTokens[1].map((t) => ({
                tokenId: Number(t[0].toString()),
                quantity: Number(t[1].toString()),
              })),
            }

            const newContract = await sdk.getContract(
              burnTokens.burnContractAddress,
              ContractType.DropERC1155,
            )
            let hasBalance = true

            let unclaimedSupply = 999999999
            const burnTokensMetadata = await Promise.all(
              burnTokens.burnTokens.map(async (token) => {
                const balance = await newContract.call('balanceOf', [
                  walletAddress,
                  token.tokenId,
                ])
                if (token.quantity > Number(balance)) hasBalance = false

                const totalSupplyByToken = Math.floor(balance / token.quantity)
                unclaimedSupply = Math.min(unclaimedSupply, totalSupplyByToken)

                const metadata = await newContract.get(token.tokenId)
                return {
                  ...metadata,
                  hasBalance: !(token.quantity > Number(balance)),
                  balance: Number(balance),
                }
              }),
            )

            claims = [
              metadata,
              [{ burnTokens, burnTokensMetadata }],
              hasBalance,
              [hasBalance ? '' : 'You have no enough access tokens'],
              { burnTokens, burnTokensMetadata },
              unclaimedSupply,
            ]
          } catch (e) {
            if (e.message?.includes('No tokens has been set yet.')) {
              claims = [
                metadata,
                [],
                false,
                ['No tokens has been set yet.'],
                undefined,
              ]
            }
          }
          break
        }
        default:
          throw new Error(messages.CONTRACT_TYPE_IS_NOT_SUPPORTED)
      }

      if (claims[0]?.metadata?.name.includes('Failed to load NFT metadata'))
        throw new Error(messages.TOKEN_NOTFOUND)

      claims[0].metadata = {
        ...claims[0]?.metadata,
        supply: parseInt(claims[0]?.supply, 10),
        id: tokenId,
      }

      performance.mark('getOwned-start')

      const settings = {
        apiKey: getAlchemyApiKey(network),
        network: getAlchemyNetwork(network),
      }
      const alchemy = new Alchemy(settings)

      let owned = []
      let numberOfOwnedTokens

      try {
        const nfts = []
        // Get the async iterable for the owner's NFTs.
        const nftsIterable = alchemy.nft.getNftsForOwnerIterator(walletAddress)

        // Iterate over the NFTs and add them to the nfts array.
        for await (const nft of nftsIterable) {
          nfts.push(nft)
        }

        owned = nfts.filter(
          (nft) =>
            nft.contract?.address?.toLowerCase() ===
              contractAddress.toLowerCase() && nft.tokenId == tokenId, // eslint-disable-line eqeqeq
        )
        numberOfOwnedTokens = owned?.[0]?.balance
      } catch (error) {
        owned = await contract.getOwned(walletAddress)
        owned = owned.filter(
          // eslint-disable-next-line eqeqeq
          (nft) => nft.metadata?.id == tokenId,
        )
        numberOfOwnedTokens = owned?.[0]?.quantityOwned
      }
      const currentClaimCondition =
        isMultipleClaimConditionsSupported && isObjectFull(claims[4])
          ? claims[4]
          : claims[1]?.[0]
      const maxClaimable =
        isMultipleClaimConditionsSupported &&
        isObjectFull(currentClaimCondition)
          ? currentClaimCondition.maxClaimablePerWallet
          : 1
      if (
        maxClaimable !== 'unlimited' &&
        numberOfOwnedTokens >= maxClaimable &&
        contractType !== ContractType.TokenERC1155
      ) {
        claims[3].unshift('You have already claimed the token.')
      }

      const contractVersion =
        contractType !== ContractType.TokenERC1155
        ? 1
        : await contract.call('contractVersion')
            
      resolve({
        conditions: claims[1],
        metadata: claims[0].metadata,
        condition: currentClaimCondition,
        canClaim: claims[2],
        claimReason: claims[3].shift(),
        network,
        numberOfOwnedTokens: numberOfOwnedTokens || 0,
        claimedSupply: claims?.[5]?.length || undefined,
        unclaimedSupply:
          contractType !== ContractType.TokenERC1155
          ? claims?.[6]?.length || undefined
          : (
            contractVersion === 1
            ? 1
            : claims?.[5] || 0
          ),
        totalSupply:
          isArray(claims?.[5]) && isArray(claims?.[6])
            ? claims?.[5]?.length + claims?.[6]?.length
            : undefined,
      })
    } catch (e) {
      reject(e)
    }
  })
}
/**
 * Get claim metadata on HAZAMA network
 *
 * @private
 * @param {object} params parameters
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {string} params.contractAddress contract address
 * @param {BigNumberish} params.tokenId the id of the token
 * @param {string} params.walletAddress wallet address
 * @return {Promise<ClaimMetadata>} claim metadata
 */
async function getClaimMetadataOnHazamaNetwork({
  host = HOST,
  contractAddress,
  tokenId,
  walletAddress,
}) {
  return new Promise(async (resolve, reject) => {
    try {
      if (!host) throw new Error(messages.HOST_UNDEFINED)

      if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

      if (!tokenId && tokenId !== 0)
        throw new Error(messages.TOKEN_ID_UNDEFINED)

      if (!walletAddress) throw new Error(messages.WALLET_ADDRESS_UNDEFINED)

      lib.connect(host, async () => {
        try {
          const path = await lib.get(
            `/_/contract.${contractAddress}.namespace`,
            null,
            false,
          )
          if (path === null) reject(new Error(messages.TOKEN_NOTFOUND))

          const data = await lib.get(`${path}/${tokenId}`, null, false)
          if (data === null) reject(new Error(messages.TOKEN_NOTFOUND))

          const condition = await lib.get(
            `${path}/${tokenId}/conditions`,
            null,
            false,
          )
          const claimCount = await lib.get(
            `${path}/${tokenId}/addresses/${walletAddress}`,
            null,
            false,
          )

          data.id = tokenId
          data.image = data.image.replace(
            'gateway.ipfscdn.io',
            'hzbs.infura-ipfs.io',
          )

          let claimReason
          let canClaim = false
          if (condition) {
            const existsSnap = await iterate({
              lib,
              path: `${path}/${tokenId}/permissions/claim/`,
              match: '0x[0-9a-zA-Z]+',
            })

            const filterdSnap = existsSnap?.filter(
              (snap) => walletAddress === snap[0].split('/').pop(),
            )

            if (existsSnap?.length) {
              if (!filterdSnap?.length) {
                claimReason = 'This address is not on the allowlist.'
              } else {
                const allowInfo = await lib.get(
                  `${path}/${tokenId}/permissions/claim/${walletAddress}`,
                  null,
                  false,
                )
                if (claimCount > 0 && allowInfo.maxClaimable >= claimCount) {
                  claimReason = 'You have already claimed the token.'
                }
              }
            }
            if (!claimReason) {
              if (moment(condition.startTime).unix() > moment().unix()) {
                claimReason = 'Not started yet'
              } else if (
                claimCount > 0 &&
                // eslint-disable-next-line eqeqeq
                condition.waitInSeconds == '[object Object]'
              ) {
                claimReason = 'You have already claimed the token.'
              } else if (data.supply >= condition.maxQuantity) {
                claimReason = 'Exceed max quantity.'
              } else {
                canClaim = true
              }
            }
          } else if (!condition) {
            claimReason = 'There is no claim condition set.'
          }
          resolve({
            metadata: data,
            condition,
            canClaim,
            claimReason,
            path,
            network: 'hazama',
          })
        } catch (e) {
          reject(e)
        }
      })
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Get claim metadata
 * @function
 * @name getClaimMetadata
 * @memberof hazama-lib/Others
 *
 * @example <caption>[Get claim metadata on EVM network]</caption>
 * import hazlib from hazama-lib
 * ...
 * const network = 'polygon'
 * const contractAddress = '0x...'
 * const tokenId = 0
 * const clientKey = 'dummy'
 *
 * const claimMetadata = hazlib.getClaimMetadata({ network, contractAddress, tokenId, clientKey })
 * ...
 *
 * @example <caption>[Get claim metadata on HAZAMA network]</caption>
 * import hazlib from hazama-lib
 * ...
 * const network = 'hazama'
 * const host = 'ws://...'
 * const contractAddress = '0x...'
 * const tokenId = 0
 * const walletAddress = '0x...'
 * const clientKey = 'dummy'
 *
 * const claimMetadata = await hazlib.getClaimMetadata({ network, host, contractAddress, tokenId, walletAddress, clientKey })
 * ...
 *
 * @param {object} params parameters for getting contents of a wrapped token bundle
 * @param {string} params.network the chain name or rpc url
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {string} params.contractAddress the address of the deployed contract
 * @param {BigNumberish} params.tokenId the id of the wrapped token bundle
 * @param {string} params.walletAddress wallet address
 * @param {string} [params.clientKey] client key
 * @returns {ClaimMetadata} Output for the contents of a bundle wrapped token
 */
lib.getClaimMetadata = async ({
  network,
  host = HOST,
  contractAddress,
  tokenId,
  walletAddress,
  clientKey,
  isCount = true,
  isMultipleClaimConditionsSupported = false,
}) => {
  const caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 13,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

    if (!tokenId && tokenId !== 0) throw new Error(messages.TOKEN_ID_UNDEFINED)

    if (!walletAddress) throw new Error(messages.WALLET_ADDRESS_UNDEFINED)

    const result =
      network !== 'hazama'
        ? await getClaimMetadataOnEVMNetwork({
            contractNetwork: network,
            contractAddress,
            tokenId,
            walletAddress,
            isMultipleClaimConditionsSupported,
          })
        : await getClaimMetadataOnHazamaNetwork({
            host,
            contractAddress,
            tokenId,
            walletAddress,
          })

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 13,
      walletAddress: caller,
      status: RequestTransactionStatus.SUCCEEDED,
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 13,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * Get claim metadata for HazBase.com usage
 *
 * @private
 * @param {object} params parameters for getting contents of a wrapped token bundle
 * @param {string} params.network the chain name or rpc url
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {string} params.contractAddress the address of the deployed contract
 * @param {BigNumberish} params.tokenId the id of the wrapped token bundle
 * @param {string} params.walletAddress wallet address
 * @param {string} [params.clientKey] client key
 * @returns {ClaimMetadata} Output for the contents of a bundle wrapped token
 */
lib.getClaimMetadataForHazBase = async ({
  network,
  host = HOST,
  contractAddress,
  tokenId,
  walletAddress,
  clientKey,
  isCount = true,
  isMultipleClaimConditionsSupported = false,
}) => {
  const caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 13,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (!network) throw new Error(messages.CONTRACT_NETWORK_UNDEFINED)

    if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

    if (!tokenId && tokenId !== 0) throw new Error(messages.TOKEN_ID_UNDEFINED)

    if (!walletAddress) throw new Error(messages.WALLET_ADDRESS_UNDEFINED)

    const result =
      network !== 'hazama'
        ? await getClaimMetadataOnEVMNetworkForHazBase({
            network,
            contractAddress,
            tokenId,
            walletAddress,
            isMultipleClaimConditionsSupported,
          })
        : await getClaimMetadataOnHazamaNetwork({
            host,
            contractAddress,
            tokenId,
            walletAddress,
          })

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 13,
      walletAddress: caller,
      status: RequestTransactionStatus.SUCCEEDED,
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 13,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * Get token on EVM network
 *
 * @private
 *
 * @param {Object} params parameters
 * @param {string} params.network network
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions.<br/>If private key is not set, Signer is required.
 * @param {string} [params.privateKey] wallet's private key.<br/>If Signer is not set, private key is required.
 * @param {string} params.contractAddress contract address
 * @param {string} [params.walletAddress] wallet address
 * @param {string} params.tokenId token id
 * @returns {Promise<EVMToken>} token
 */
async function getTokenOnEVMNetwork({
  network,
  signer,
  privateKey,
  contractAddress,
  walletAddress,
  tokenId,
  isMultipleClaimConditionsSupported = false,
}) {
  const { type: contractType } = await lib.getContractType({
    network,
    contractAddress,
    isCount: false,
  })
  const isVoteContract = contractType === ContractType.VoteERC20

  if (!network) throw new Error(messages.NETWORK_UNDEFINED)

  if (!isVoteContract && !signer && !privateKey)
    throw new Error(messages.SIGNER_PRIVATEKEY_UNDEFINED)

  if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

  if (!tokenId && tokenId !== 0) throw new Error(messages.TOKEN_ID_UNDEFINED)

  const tokens = await searchTokensOnEVMNetwork({
    network,
    signer,
    privateKey,
    contractAddress,
    walletAddress,
    isMultipleClaimConditionsSupported,
  })

  return !isVoteContract
    ? tokens[tokenId]
    : tokens.find((token) => token.id === tokenId)
}

/**
 * Search tokens on HAZAMA network
 *
 * @private
 *
 * @param {Object} params parameters
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {string} params.walletAddress wallet address
 * @param {string} params.contractAddress contract address
 * @param {string} params.tokenId token id
 * @returns {Promise<HazamaToken>} token
 */
async function getTokenOnHazamaNetwork({
  host = HOST,
  walletAddress,
  contractAddress,
  tokenId,
}) {
  return new Promise(async (resolve, reject) => {
    try {
      if (!host) throw new Error(messages.HOST_UNDEFINED)

      if (!walletAddress) throw new Error(messages.WALLET_ADDRESS_UNDEFINED)

      if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

      if (!tokenId && tokenId !== 0)
        throw new Error(messages.TOKEN_ID_UNDEFINED)

      lib.connect(host, async () => {
        const path = await lib.get(
          `/_/contract.${contractAddress}.namespace`,
          null,
          false,
        )
        if (path === null) reject(new Error(messages.TOKEN_NOTFOUND))
        let meta = await lib.get(`${path}.meta`, null, false)

        let token
        if (meta.Type === ContractType.FUNGIBLE) {
          let [totalSupply, owned] = await Promise.all([
            lib.get(`${path}/totalSupply`, null, false),
            lib.get(`${path}/${walletAddress}`, null, false),
          ])

          meta = objectToLowerCase(meta)

          if (totalSupply > 0) {
            const bigNum = Number(totalSupply).toLocaleString('fullwide', {
              useGrouping: false,
            })
            totalSupply = Number(
              formatUnits((bigNum || 0).toString(), meta.info.decimals),
            ).toFixed()
          }
          if (owned > 0) {
            const bigNum = Number(owned).toLocaleString('fullwide', {
              useGrouping: false,
            })
            owned = Number(
              formatUnits((bigNum || 0).toString(), meta.info.decimals),
            ).toFixed()
          }

          token = {
            ...meta,
            totalSupply: `${totalSupply} ${meta.symbol}`,
            owned,
            type: StandardContractType[meta.type],
            contractType: meta.type,
          }
        } else {
          token = await lib.get(`${path}/${tokenId}`, null, false)

          if (!token) reject(new Error(messages.TOKEN_NOTFOUND))

          token.id = tokenId
          token.contractId = meta.ContractHash
          token.contractVersion = meta.ContractVersion
          token.description = token.description
            ?.replace(/\\n/g, ' ')
            .replace(/\n/g, ' ')
          token.type = StandardContractType[meta.Type]
          token.contractType = meta.Type
          token.image = token.image.replace(
            'gateway.ipfscdn.io',
            'hzbs.infura-ipfs.io',
          )
          try {
            const condition = await lib.get(
              `${path}/${tokenId}/conditions`,
              null,
              false,
            )

            if (condition) {
              token.startTime = moment(condition.startTime.toString()).format(
                'YYYY/MM/DD HH:mm',
              )
              token.startTimeOrigin = condition.startTime
              token.availableSupply =
                condition.availableSupply.length > 18
                  ? 'Unlimited'
                  : condition.availableSupply
              token.maxQuantity =
                condition.maxQuantity === 'unlimited'
                  ? 'Unlimited'
                  : formatUnits(condition.maxQuantity, 0)

              const existsSnap = (
                await iterate({
                  lib,
                  path: `${path}/${tokenId}/permissions/claim/`,
                  match: '0x[0-9a-zA-Z]+',
                })
              )?.map((snap) => ({
                address: snap[0].split('/').pop(),
                maxClaimable: Number(snap[1].maxClaimable),
              }))
              token.snapshot = existsSnap || []
            }
          } catch (e) {
            console.error(e)
          }
        }

        if (token?.info?.decimals) {
          token.decimals = token.info.decimals
        }

        resolve(token)
      })
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Get token by id
 *
 * @function
 * @name getToken
 * @memberof hazama-lib/Others
 *
 * @example <caption>Get token by id on EVM network</caption>
 *
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const privateKey = 'dummy'
 * const contractAddress = '0x...'
 * const tokenId = 0
 * const clientKey = 'dummy'
 *
 * const token = await hazlib.getToken({network, privateKey, contractAddress, tokenId, clientKey})
 * ...
 *
 *
 * @example <caption>Get token by id on HAZAMA network</caption>
 *
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'hazama'
 * const host = 'wss://n.hazbase.com'
 * const walletAddress = '0x...'
 * const contractAddress = '0x...'
 * const tokenId = 0
 * const clientKey = 'dummy'
 *
 * const token = await hazlib.getToken({network, host, walletAddress, contractAddress, tokenId, clientKey})
 * ...
 *
 * @param {Object} params parameters
 * @param {string} params.network network
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions.<br/>If private key is not set, Signer is required.
 * @param {string} [params.privateKey] wallet's private key.<br/>If Signer is not set, private key is required.
 * @param {string} [params.walletAddress] wallet address
 * @param {string} params.contractAddress contract address
 * @param {string} params.tokenId token id
 * @param {string} [params.clientKey] client key
 * @returns {(EVMToken|HazamaToken)} token
 */
lib.getToken = async ({
  network,
  host = HOST,
  signer,
  privateKey,
  walletAddress,
  contractAddress,
  tokenId,
  clientKey,
  isCount = true,
  isMultipleClaimConditionsSupported = false,
}) => {
  let caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 18,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (!network) throw new Error(messages.NETWORK_UNDEFINED)

    if (network === 'hazama' && !host) throw new Error(messages.HOST_UNDEFINED)

    if (network === 'hazama' && !walletAddress)
      throw new Error(messages.WALLET_ADDRESS_UNDEFINED)

    if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

    if (!tokenId && tokenId !== 0) throw new Error(messages.TOKEN_ID_UNDEFINED)

    caller = await getCaller({ network, host, signer, privateKey })

    const result =
      network !== 'hazama'
        ? await getTokenOnEVMNetwork({
            network,
            signer,
            privateKey,
            contractAddress,
            walletAddress,
            tokenId,
            isMultipleClaimConditionsSupported,
          })
        : await getTokenOnHazamaNetwork({
            host,
            walletAddress,
            contractAddress,
            tokenId,
          })

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 18,
      walletAddress: caller,
      status: RequestTransactionStatus.SUCCEEDED,
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 18,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * Get token in wallet (HAZAMA network only)
 *
 * @function
 * @name getWalletTokens
 * @memberof hazama-lib/Others
 *
 * @example
 *
 * import hazlib from 'hazama-lib'
 * ...
 * const host = 'wss://n.hazbase.com'
 * const walletAddress = '0x...'
 * const query = 'search=tokenName'
 * const clientKey = 'dummy'
 * const historyIndex = 0
 * const limit = 10
 *
 * const tokens = await hazlib.getWalletTokens({host, walletAddress, query, clientKey, historyIndex, limit})
 * ...
 *
 * @param {Object} params parameters
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {string} params.walletAddress wallet address
 * @param {string} [params.query=''] query
 * @param {string} [params.clientKey] client key
 * @param {number} [params.historyIndex] history index
 * @param {number} [params.limit] limit
 * @returns {Token[]} list of tokens
 */
lib.getWalletTokens = async ({
  host = HOST,
  walletAddress,
  query = '',
  clientKey,
  isCount = true,
  historyIndex = 0,
  limit = 10,
}) => {
  const caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 22,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    const promise = new Promise(async (resolve, reject) => {
      try {
        if (!host) throw new Error(messages.HOST_UNDEFINED)

        if (!walletAddress) throw new Error(messages.WALLET_ADDRESS_UNDEFINED)

        const queries = parse(query)
        const endpointData = url.parse(host)
        const httpService = new HttpService(`https://${endpointData.hostname}/`)

        try {
          const { data: dataForGetHistory } = await httpService.post({
            url: '',
            data: {
              jsonrpc: '2.0',
              method: 'core.get',
              params: [`/_/addr.${walletAddress}`],
              id: 1,
            },
          })
          const { result: history } = dataForGetHistory

          // filter history
          let filteredHistory = []
          Object.keys(history).forEach((hs) => {
            if (
              new RegExp(
                `^/hazama/[a-zA-Z0-9]+/projects/[0-9a-z]+/contracts/[0-9a-z]+/${walletAddress}$`,
              ).test(hs) ||
              /^\/hazama\/\w+\/projects\/[0-9a-z]+\/contracts\/[0-9a-z]+\/\d+$/.test(
                hs,
              )
            ) {
              filteredHistory.push({
                timestamp: history?.txs?.[history[hs]],
                path: hs,
              })
            }
          })

          // sort history in descending order
          filteredHistory.sort((a, b) => b.timestamp - a.timestamp)
          filteredHistory = filteredHistory.slice(historyIndex)

          const wallet = []
          let lastHistoryIndex
          let count = 1

          for await (const hs of filteredHistory) {
            lastHistoryIndex = count
            count++
            const { path: x } = hs
            const { data: dataForGetInfo } = await httpService.post({
              url: '',
              data: {
                jsonrpc: '2.0',
                method: 'core.get',
                params: [x],
                id: 1,
              },
            })
            const { result: info } = dataForGetInfo

            if (typeof info === 'number') {
              const pathArr = x.split('/')
              const last = pathArr.pop()

              if (last === walletAddress) {
                const metaPath = pathArr.join('/')
                const { data: dataForGetMeta } = await httpService.post({
                  url: '',
                  data: {
                    jsonrpc: '2.0',
                    method: 'core.get',
                    params: [metaPath + '.meta'],
                    id: 1,
                  },
                })
                const { result } = dataForGetMeta
                const meta = objectToLowerCase(result)

                if (meta) {
                  if (
                    queries?.search &&
                    !new RegExp(`${queries.search}`, 'gi').test(meta.name)
                  )
                    continue

                  const bigNum = Number(info).toLocaleString('fullwide', {
                    useGrouping: false,
                  })
                  meta.owned = Number(
                    formatUnits((bigNum || 0).toString(), meta.info.decimals),
                  ).toFixed()
                  meta.type = StandardContractType[meta.type]
                  meta.timestamp = history.txs[history[x]]
                  meta.decimals = meta.info.decimals
                  meta.path = metaPath

                  wallet.push(meta)
                  if (wallet.length === limit) {
                    break
                  }
                }
              }
            } else if (typeof info === 'object' && info?.owner_index) {
              if (info.owner === walletAddress) {
                const paths = x.split('/')
                info.issuer = paths[2]
                info.projectId = paths[4]
                info.contractId = paths[6]
                info.path = paths.slice(0, -1).join('/')

                if (
                  queries?.search &&
                  !new RegExp(`${queries.search}`, 'gi').test(info.name)
                )
                  continue

                const contractPath = generatePath({
                  walletAddress: info.issuer,
                  projectId: info.projectId,
                  contractAddress: info.contractId,
                })
                const { data: dataForGetMeta } = await httpService.post({
                  url: '',
                  data: {
                    jsonrpc: '2.0',
                    method: 'core.get',
                    params: [`${contractPath}.meta`],
                    id: 1,
                  },
                })
                const { result: meta } = dataForGetMeta

                if (meta.Type === ContractType.NONTRANSFERABLE) {
                  const tokenId = paths[paths.length - 2]

                  if (isNaN(tokenId)) continue
                  else {
                    const idx = wallet.findIndex(
                      (data) => data.path === info.path,
                    )
                    if (idx >= 0) {
                      wallet[idx].owned++
                      continue
                    }
                  }
                  info.path = paths.slice(0, -2).join('/')
                }

                info.id = paths[7]
                info.network = 'hazama'
                info.type = StandardContractType[meta.Type]
                info.timestamp = history.txs[history[x]]
                info.owned = 1

                wallet.push(info)
                if (wallet.length === limit) {
                  break
                }
              }
            }
          }

          createRequestTransaction({
            clientKey: targetClientKey,
            functionId: 22,
            walletAddress: caller,
            status: RequestTransactionStatus.SUCCEEDED,
            isCount,
          })

          resolve({
            data: wallet,
            lastHistoryIndex: historyIndex + lastHistoryIndex,
            total: historyIndex + filteredHistory.length,
          })
        } catch (e) {
          reject(e)
        }
      } catch (e) {
        reject(e)
      }
    })

    const result = await promise
    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 22,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * Get owned tokens on EVM network
 *
 * @private
 * @param {object} params parameters
 * @param {string} [params.contractNetwork] network
 * @param {string} params.contractAddress contract address
 * @param {string} params.walletAddress wallet address
 * @return {Promise<Token[]>} owned tokens
 */
async function getOwnedOnEVMNetwork({
  contractNetwork,
  contractAddress,
  walletAddress,
}) {
  return new Promise(async (resolve, reject) => {
    try {
      if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

      if (!walletAddress) throw new Error(messages.WALLET_ADDRESS_UNDEFINED)

      const res = await resolveContractTypeWithMultiNetwork({
        contractNetwork,
        contractAddress,
      })

      if (res) {
        const { contractType, network, sdk } = res

        if (!!contractNetwork && contractNetwork !== network) {
          throw new Error(messages.TOKEN_NOTFOUND)
        }

        if (
          ![
            ContractType.TokenERC1155,
            ContractType.DropERC1155,
            ContractType.DropERC721,
          ].includes(contractType)
        ) {
          throw new Error(messages.CONTRACT_TYPE_IS_NOT_SUPPORTED)
        }

        const contract = await sdk.getContract(contractAddress, contractType)
        const nfts = await contract.getOwned(walletAddress)
        const tokens = nfts.map((nft) => {
          return {
            ...nft.metadata,
            image: nft.metadata.image || nft.metadata.animation_url,
            id: Number(nft.metadata.id),
            contractId: contractAddress,
            quantityOwned: parseInt(nft.quantityOwned, 10),
            supply: parseInt(nft.supply, 10),
            network,
          }
        })

        resolve(tokens)
      }
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Get owned tokens on HAZAMA network
 *
 * @private
 * @param {object} params parameters
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {string} params.contractAddress contract address
 * @param {string} params.walletAddress wallet address
 * @return {Promise<Token[]>} owned tokens
 */
async function getOwnedOnHazamaNetwork({
  host = HOST,
  contractAddress,
  walletAddress,
}) {
  return new Promise(async (resolve, reject) => {
    try {
      if (!host) throw new Error(messages.HOST_UNDEFINED)

      if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

      if (!walletAddress) throw new Error(messages.WALLET_ADDRESS_UNDEFINED)

      lib.connect(HOST, async () => {
        try {
          const path = await lib.get(
            `/_/contract.${contractAddress}.namespace`,
            null,
            false,
          )
          if (path === null) reject(new Error(messages.TOKEN_NOTFOUND))
          const history = await lib.get(`/_/addr.${walletAddress}`, null, false)

          const list = []
          const check = {}
          for (const x in history) {
            if (new RegExp(`^${path}/[0-9]+/[0-9]+$`).test(x)) {
              const meta = await lib.get(x, null, false)
              if (meta.owner !== walletAddress) continue
              const index = x.split('/').splice(-2, 1).join('/')
              if (check[index] > -1) {
                list[check[index]].owned++
                continue
              }
              check[index] = list.length
              list.push({ path: x, owned: 1 })
            }
          }

          const nfts = await Promise.all(
            list.map((x) => {
              return new Promise(async (resolve, reject) => {
                const metadata = await lib.get(x.path, null, false)
                metadata.image = metadata.image.replace(
                  'gateway.ipfscdn.io',
                  'hzbs.infura-ipfs.io',
                )

                const arr = x.path.split('/')
                const id = arr[arr.length - 2]

                const nftPath = arr.join('/')
                const meta = await lib.get(nftPath, null, false)

                resolve({
                  ...metadata,
                  id,
                  contractId: contractAddress,
                  quantityOwned: x.owned,
                  supply: meta.supply,
                })
              })
            }),
          )
          nfts?.sort((a, b) => a.id - b.id)

          resolve(nfts)
        } catch (e) {
          reject(e)
        }
      })
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Get owned tokens
 * @function
 * @name getOwned
 * @memberof hazama-lib/Others
 *
 * @example <caption>[Get owned tokens on EVM network]</caption>
 * import hazlib from hazama-lib
 * ...
 * const network = 'polygon'
 * const contractAddress = '0x...'
 * const walletAddress = '0x...'
 * const clientKey = 'dummy'
 *
 * const tokens = hazlib.getOwned({ network, contractAddress, walletAddress, clientKey })
 * ...
 *
 * @example <caption>[Get owned tokens on HAZAMA network]</caption>
 * import hazlib from hazama-lib
 * ...
 * const network = 'hazama'
 * const host = 'ws://...'
 * const contractAddress = '0x...'
 * const walletAddress = '0x...'
 * const clientKey = 'dummy'
 *
 * const tokens = await hazlib.getOwned({ network, host, contractAddress, walletAddress, clientKey })
 * ...
 *
 * @param {object} params parameters for getting owned tokens
 * @param {string} params.network the chain name or rpc url
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {string} params.contractAddress the address of the deployed contract
 * @param {string} params.walletAddress wallet address
 * @param {string} [params.clientKey] client key
 * @returns {ClaimMetadata} Output for the contents of a bundle wrapped token
 */
lib.getOwned = async ({
  network,
  host = HOST,
  contractAddress,
  walletAddress,
  clientKey,
  isCount = true,
}) => {
  const caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 16,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

    if (!walletAddress) throw new Error(messages.WALLET_ADDRESS_UNDEFINED)

    const result =
      network !== 'hazama'
        ? await getOwnedOnEVMNetwork({
            contractNetwork: network,
            contractAddress,
            walletAddress,
          })
        : await getOwnedOnHazamaNetwork({
            host,
            contractAddress,
            walletAddress,
          })

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 16,
      walletAddress: caller,
      status: RequestTransactionStatus.SUCCEEDED,
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 16,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * Set client key
 * @function
 * @name setClientKey
 * @memberof hazama-lib/Others
 *
 * @example
 * import hazlib from hazama-lib
 * ...
 * const network = 'polygon'
 * const contractAddress = '0x...'
 * const walletAddress = '0x...'
 * const clientKey = 'dummy'
 *
 * hazlib.setClientKey({ clientKey })
 * const tokens = hazlib.getOwned({ network, contractAddress, walletAddress })
 * ...
 *
 * @param {object} params parameters for setting client key
 * @param {string} params.clientKey client key
 * @returns {void}
 */
lib.setClientKey = ({ clientKey }) => {
  if (!clientKey) throw new Error(messages.CLIENT_KEY_UNDEFINED)
  currentClientKey = clientKey
}

/**
 * get roles on EVM Network
 *
 * @private
 * @param {object} params parameters for getting roles
 * @param {string} params.network the chain name or rpc url
 * @param {string} params.contractAddress the token contract address
 * @return {Promise<RolesObject>} promise
 */
async function getRolesOnEVMNetwork({ network, contractAddress }) {
  return new Promise(async (resolve, reject) => {
    try {
      let contract = null
      const options = {
        readonlySettings: {
          rpcUrl: getRpcUrl(network),
        },
        ...(isArrayFull(gatewayUrls) ? { gatewayUrls } : {}),
        ...(isNil(thirdwebSecretKey) &&
        (!isNil(thirdwebClientId) || !isNil(thirdwebClientIdFromAPI))
          ? { clientId: thirdwebClientId || thirdwebClientIdFromAPI }
          : {}),
        ...(!isNil(thirdwebSecretKey) ? { secretKey: thirdwebSecretKey } : {}),
      }

      if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

      if (!ethers.utils.isAddress(contractAddress))
        throw new Error(messages.CONTRACT_ADDRESS_INVALID)

      const sdk = new ThirdwebSDK(network, options)

      const contractType = await resolveContractTypeWithCustom({
        sdk,
        contractAddress,
      })

      if (
        ![
          ContractType.DropERC721,
          ContractType.TokenERC721,
          ContractType.DropERC1155,
          ContractType.TokenERC20,
          ContractType.Multiwrap,
          ContractType.TokenERC1155,
        ].includes(contractType)
      )
        throw new Error(messages.CONTRACT_TYPE_IS_INVALID)

      switch (contractType) {
        case ContractType.DropERC721:
        case ContractType.TokenERC721:
        case ContractType.DropERC1155:
        case ContractType.TokenERC20:
        case ContractType.Multiwrap:
        case ContractType.TokenERC1155:
          contract = await sdk.getContract(contractAddress, contractType)
          break
        default:
          throw new Error(messages.TOKEN_TYPE_IS_NOT_SUPPORTED)
      }

      const result = await contract.roles.getAll()

      resolve(result)
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Get permissions
 *
 * @function
 * @name getPermissions
 * @memberof hazama-lib/ERC20/ERC721/ERC1155
 *
 * @example <caption>Get permissions on EVM network</caption>
 *
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const contractAddress = '0x...'
 * const clientKey = 'dummy'
 *
 * const result = hazlib.getPermissions({ network, contractAddress, clientKey })
 * ...
 *
 * @param {object} params parameters for getting permissions
 * @param {string} params.network the chain name or rpc url
 * @param {string} params.contractAddress the token contract address
 * @param {string} [params.clientKey] client key
 * @return {RolesObject} permission list
 */
lib.getPermissions = async ({
  network,
  contractAddress,
  clientKey,
  isCount = true,
}) => {
  const caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 49,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (!network) throw new Error(messages.NETWORK_UNDEFINED)

    if (network === 'hazama') throw new Error(messages.NETWORK_IS_NOT_SUPPORTED)

    const result = await getRolesOnEVMNetwork({ network, contractAddress })

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 49,
      walletAddress: caller,
      status: RequestTransactionStatus.SUCCEEDED,
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 49,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * set all roles on EVM Network
 *
 * @private
 * @param {object} params parameters for granting roles
 * @param {string} params.network the chain name or rpc url
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions
 * @param {string} [params.privateKey] wallet's private key
 * @param {string} params.contractAddress the token contract address
 * @param {RolesObject} params.roles roles of contract
 * @param {string} [params.relayerUrl] relayer url (gasless transaction mode)
 * @return {Promise<TransactionResult[]>} promise
 */
async function setAllRolesOnEVMNetwork({
  network,
  signer,
  privateKey,
  contractAddress,
  roles,
  relayerUrl,
}) {
  return new Promise(async (resolve, reject) => {
    try {
      let contract = null
      let options = {
        readonlySettings: {
          rpcUrl: getRpcUrl(network),
        },
        ...(isArrayFull(gatewayUrls) ? { gatewayUrls } : {}),
        ...(isNil(thirdwebSecretKey) &&
        (!isNil(thirdwebClientId) || !isNil(thirdwebClientIdFromAPI))
          ? { clientId: thirdwebClientId || thirdwebClientIdFromAPI }
          : {}),
        ...(!isNil(thirdwebSecretKey) ? { secretKey: thirdwebSecretKey } : {}),
      }

      if (!roles) throw new Error(messages.ROLES_UNDEFINED)

      if (!signer && !privateKey)
        throw new Error(messages.SIGNER_PRIVATEKEY_UNDEFINED)

      if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

      if (!ethers.utils.isAddress(contractAddress))
        throw new Error(messages.CONTRACT_ADDRESS_INVALID)

      if (relayerUrl) {
        options = {
          ...options,
          gasless: {
            openzeppelin: { relayerUrl },
          },
        }
      }

      let sdk = new ThirdwebSDK(network, options)
      if (signer) {
        sdk.updateSignerOrProvider(signer)
      }

      if (privateKey) {
        sdk = new ThirdwebSDK(
          new ethers.Wallet(
            privateKey,
            ethers.getDefaultProvider(getRpcUrl(network)),
          ),
          options,
        )
      }

      const contractType = await resolveContractTypeWithCustom({
        sdk,
        contractAddress,
      })

      if (
        ![
          ContractType.DropERC721,
          ContractType.TokenERC721,
          ContractType.DropERC1155,
          ContractType.TokenERC20,
          ContractType.Multiwrap,
          ContractType.TokenERC1155,
        ].includes(contractType)
      )
        throw new Error(messages.CONTRACT_TYPE_IS_INVALID)

      switch (contractType) {
        case ContractType.DropERC721:
        case ContractType.TokenERC721:
        case ContractType.DropERC1155:
        case ContractType.TokenERC20:
        case ContractType.Multiwrap:
        case ContractType.TokenERC1155:
          contract = await sdk.getContract(contractAddress, contractType)
          break
        default:
          throw new Error(messages.TOKEN_TYPE_IS_NOT_SUPPORTED)
      }

      const result = await contract.roles.setAll(roles)

      resolve(result)
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Overwrite permissions<br/><br/>
 * Important Note:
 * 1. If wallet address exists in role, it will be skipped
 * 2. If zero address(0x0000000000000000000000000000000000000000) exists in role, all wallet address of role will be skipped
 * @function
 * @name overwritePermissions
 * @memberof hazama-lib/ERC20/ERC721/ERC1155
 *
 * @example <caption>[overwrite permissions with privateKey on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const privateKey = 'dummy'
 * const contractAddress = '0x...'
 * const roles = {
 *   admin: ['0x...'],
 *   minter: ['0x...'],
 *   transfer: ['0x...'],
 * }
 * const clientKey = 'dummy'
 *
 * const result = hazlib.overwritePermissions({ network, privateKey, contractAddress, roles, clientKey })
 * ...
 *
 * @example <caption>[overwrite permissions with signer and relayer on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const signer = <signer object>
 * const relayerUrl = 'https://api.defender.openzeppelin.com/autotasks/0f7cf308-.../runs/webhook/fd447953-...' // Openzeppelin webhook URI
 * const contractAddress = '0x...'
 * const roles = {
 *   admin: ['0x...'],
 *   minter: ['0x...'],
 *   transfer: ['0x...'],
 * }
 * const clientKey = 'dummy'
 *
 * const result = hazlib.overwritePermissions({ network, signer, contractAddress, roles, relayerUrl, clientKey })
 * ...
 *
 * @param {object} params parameters for adding permissions
 * @param {string} params.network the chain name or rpc url
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions
 * @param {string} [params.privateKey] wallet's private key
 * @param {string} params.contractAddress the token contract address
 * @param {RolesObject} params.roles roles of contract
 * @param {string} [params.relayerUrl] relayer url (gasless transaction mode)
 * @param {string} [params.clientKey] client key
 * @return {TransactionResult[]} transaction id (Hazama) or transaction result
 */
lib.overwritePermissions = async ({
  network,
  signer,
  privateKey,
  contractAddress,
  roles,
  relayerUrl,
  clientKey,
  isCount = true,
}) => {
  let caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 50,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (!network) throw new Error(messages.NETWORK_UNDEFINED)

    if (network === 'hazama') throw new Error(messages.NETWORK_IS_NOT_SUPPORTED)

    caller = await getCaller({ network, signer, privateKey })

    const result = await setAllRolesOnEVMNetwork({
      network,
      signer,
      privateKey,
      contractAddress,
      roles,
      relayerUrl,
    })

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 50,
      walletAddress: caller,
      status: RequestTransactionStatus.SUCCEEDED,
      transactionHash: formatTransactionHash({ input: result }),
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 50,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * Resolve contract type on EVM Network
 *
 * @private
 * @param {object} params parameters for resolving contract type
 * @param {string} params.network the chain name or rpc url
 * @param {string} params.contractAddress the contract address
 * @return {Promise<ContractTypeResult>} promise
 */
async function resolveContractTypeOnEVMNetwork({ network, contractAddress }) {
  return new Promise(async (resolve, reject) => {
    try {
      if (!network) throw new Error(messages.NETWORK_UNDEFINED)

      if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

      const sdk = new ThirdwebSDK(network, {
        ...(isNil(thirdwebSecretKey) &&
        (!isNil(thirdwebClientId) || !isNil(thirdwebClientIdFromAPI))
          ? { clientId: thirdwebClientId || thirdwebClientIdFromAPI }
          : {}),
        ...(!isNil(thirdwebSecretKey) ? { secretKey: thirdwebSecretKey } : {}),
      })

      let contractType = await sdk.resolveContractType(contractAddress)
      if (
        typeof contractType !== 'string' ||
        contractType === ContractType.Custom
      ) {
        try {
          const contractMetadata = await sdk
            .getPublisher()
            .fetchCompilerMetadataFromAddress(contractAddress)
          contractType = contractMetadata?.name || null
          if (
            ![
              CustomContract.ContractType.BurnToRedeem1155,
              CustomContract.ContractType.Vote,
            ].includes(contractType)
          )
            contractType = ContractType.Custom
        } catch (e) {}
      }
      resolve({ type: contractType || '' })
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Resolve contract type on Hazama Network
 *
 * @private
 * @param {object} params parameters for resolving contract type
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {string} params.walletAddress wallet address
 * @param {string} params.projectId project id
 * @param {string} params.contractAddress contract address
 * @return {Promise<ContractTypeResult>} promise
 */
async function resolveContractTypeOnHazamaNetwork({
  host = HOST,
  walletAddress,
  projectId,
  contractAddress,
}) {
  return new Promise(async (resolve, reject) => {
    try {
      if (!host) throw new Error(messages.HOST_UNDEFINED)

      if (!walletAddress) throw new Error(messages.WALLET_ADDRESS_UNDEFINED)

      if (!projectId) throw new Error(messages.PROJECT_ID_UNDEFINED)

      if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

      lib.connect(host, async () => {
        try {
          const path = generatePath({
            walletAddress,
            projectId,
            contractAddress,
          })
          const data = await lib.get(`${path}.meta`, null, false)
          resolve({ type: data?.Type || '' })
        } catch (e) {
          reject(e)
        }
      })
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Resolve contract type
 * @function
 * @name resolveContractType
 * @memberof hazama-lib/Others
 *
 * @example <caption>[Resolve contract type on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const contractAddress = '0x...'
 * const clientKey = 'dummy'
 *
 * const contractType = await hazlib.resolveContractType({ network, contractAddress, clientKey })
 * ...
 *
 * @example <caption>[Resolve contract type on HAZAMA network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'hazama'
 * const host = 'wss://n.hazbase.com'
 * const walletAddress = '0x...'
 * const projectId = 'project id'
 * const contractAddress = '0x...'
 * const clientKey = 'dummy'
 *
 * const contractType = await hazlib.resolveContractType({ network, host, walletAddress, projectId, contractAddress, clientKey })
 * ...
 *
 * @param {object} params parameters for getting contract type
 * @param {string} params.network the chain name or rpc url
 * @param {string} [params.host='wss://n.hazbase.com'] host
 * @param {string} params.contractAddress contract address
 * @param {string} [params.walletAddress] wallet address
 * @param {string} [params.projectId] project id
 * @param {string} [params.clientKey] client key
 * @return {ContractTypeResult}
 */
lib.resolveContractType = async ({
  network,
  host = HOST,
  contractAddress,
  projectId,
  walletAddress,
  clientKey,
  isCount = true,
}) => {
  const caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 15,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (!network) throw new Error(messages.NETWORK_UNDEFINED)

    if (network === 'hazama' && !host) throw new Error(messages.HOST_UNDEFINED)

    if (network === 'hazama' && !walletAddress)
      throw new Error(messages.WALLET_ADDRESS_UNDEFINED)

    if (network === 'hazama' && !projectId)
      throw new Error(messages.PROJECT_ID_UNDEFINED)

    if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

    const result =
      network !== 'hazama'
        ? await resolveContractTypeOnEVMNetwork({ network, contractAddress })
        : await resolveContractTypeOnHazamaNetwork({
            host,
            walletAddress,
            projectId,
            contractAddress,
          })

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 15,
      walletAddress: caller,
      status: RequestTransactionStatus.SUCCEEDED,
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 15,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * Delegate token (ERC20)
 * @function
 * @name delegate
 * @memberof hazama-lib/Others
 *
 * @example <caption>Delegate token with privateKey</caption>
 * import hazlib from hazama-lib
 * ...
 * const network = 'polygon'
 * const privateKey = 'dummy'
 * const contractAddress = '0x...'
 * const delegateeAddress = '0x...'
 * const clientKey = 'dummy'
 *
 * const result = hazlib.delegate({ network, privateKey, contractAddress, delegateeAddress, clientKey })
 * ...
 *
 * @example <caption>Delegate token with signer and relayer</caption>
 * import hazlib from hazama-lib
 * ...
 * const network = 'polygon'
 * const signer = <signer object>
 * const relayerUrl = 'https://api.defender.openzeppelin.com/autotasks/0f7cf308-.../runs/webhook/fd447953-...' // Openzeppelin webhook URI
 * const contractAddress = '0x...'
 * const delegateeAddress = '0x...'
 * const clientKey = 'dummy'
 *
 * const result = hazlib.delegate({ network, signer, contractAddress, delegateeAddress, relayerUrl, clientKey })
 * ...
 *
 * @param {object} params parameters for delegating token
 * @param {string} params.network the chain name or rpc url
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions.<br/>If private key is not set, Signer is required.
 * @param {string} [params.privateKey] wallet's private key.<br/>If Signer is not set, private key is required.
 * @param {string} params.contractAddress the operator/spender's address
 * @param {string} params.delegateeAddress delegatee wallet address
 * @param {string} [params.relayerUrl] relayer url (gasless transaction mode)
 * @param {string} [params.clientKey] client key
 * @returns {boolean} result
 */
lib.delegate = async ({
  network,
  signer,
  privateKey,
  contractAddress,
  delegateeAddress,
  relayerUrl,
  clientKey,
  isCount = true,
}) => {
  let caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 51,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    let result = false
    let options = {
      readonlySettings: {
        rpcUrl: getRpcUrl(network),
      },
      ...(isArrayFull(gatewayUrls) ? { gatewayUrls } : {}),
      ...(isNil(thirdwebSecretKey) &&
      (!isNil(thirdwebClientId) || !isNil(thirdwebClientIdFromAPI))
        ? { clientId: thirdwebClientId || thirdwebClientIdFromAPI }
        : {}),
      ...(!isNil(thirdwebSecretKey) ? { secretKey: thirdwebSecretKey } : {}),
    }
    let transactionHash = null

    if (!network) throw new Error(messages.NETWORK_UNDEFINED)

    if (!signer && !privateKey)
      throw new Error(messages.SIGNER_PRIVATEKEY_UNDEFINED)

    if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

    if (!delegateeAddress) throw new Error(messages.DELEGATEE_ADDRESS_UNDEFINED)

    if (relayerUrl) {
      options = {
        ...options,
        gasless: {
          openzeppelin: { relayerUrl },
        },
      }
    }

    let sdk = new ThirdwebSDK(network, options)
    if (signer) {
      sdk.updateSignerOrProvider(signer)
    }

    if (privateKey) {
      sdk = new ThirdwebSDK(
        new ethers.Wallet(
          privateKey,
          ethers.getDefaultProvider(getRpcUrl(network)),
        ),
        options,
      )
    }

    caller = await getCaller({ network, signer, privateKey })
    const contractType = await sdk.resolveContractType(contractAddress)
    const contract = await sdk.getContract(contractAddress, contractType)
    switch (contractType) {
      case ContractType.TokenERC20:
        const currentDelegatee = await contract.getDelegationOf(caller)
        if (currentDelegatee !== delegateeAddress) {
          const transactionResult = await contract.delegateTo(delegateeAddress)
          transactionHash = transactionResult?.receipt?.transactionHash || null
        }
        result = true
        break
      default:
        throw new Error(messages.CONTRACT_TYPE_IS_NOT_SUPPORTED)
    }

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 51,
      walletAddress: caller,
      transactionHash,
      status: RequestTransactionStatus.SUCCEEDED,
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 51,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * Cancel proposal on EVM Network
 *
 * @private
 * @param {object} params parameters for cancelling proposal
 * @param {string} params.network the chain name or rpc url
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions.<br/>If private key is not set, Signer is required.
 * @param {string} [params.privateKey] wallet's private key.<br/>If Signer is not set, private key is required.
 * @param {string} params.contractAddress the token contract address
 * @param {string} params.proposalId proposal id
 * @param {string} [params.relayerUrl] relayer url (gasless transaction mode)
 * @return {Promise<TransactionResult>} promise
 */
async function cancelProposalOnEVMNetwork({
  network,
  signer,
  privateKey,
  contractAddress,
  proposalId,
  relayerUrl,
}) {
  // eslint-disable-next-line no-async-promise-executor
  return new Promise(async (resolve, reject) => {
    try {
      let result = null
      let options = {
        readonlySettings: {
          rpcUrl: getRpcUrl(network),
        },
        ...(isArrayFull(gatewayUrls) ? { gatewayUrls } : {}),
        ...(isNil(thirdwebSecretKey) &&
        (!isNil(thirdwebClientId) || !isNil(thirdwebClientIdFromAPI))
          ? { clientId: thirdwebClientId || thirdwebClientIdFromAPI }
          : {}),
        ...(!isNil(thirdwebSecretKey) ? { secretKey: thirdwebSecretKey } : {}),
      }

      if (!network) throw new Error(messages.NETWORK_UNDEFINED)

      if (!signer && !privateKey)
        throw new Error(messages.SIGNER_PRIVATEKEY_UNDEFINED)

      if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

      if (!proposalId) throw new Error(messages.PROPOSAL_ID_UNDEFINED)

      if (relayerUrl) {
        options = {
          ...options,
          gasless: {
            openzeppelin: { relayerUrl },
          },
        }
      }

      let sdk = new ThirdwebSDK(network, options)
      if (signer) {
        sdk.updateSignerOrProvider(signer)
      }

      if (privateKey) {
        sdk = new ThirdwebSDK(
          new ethers.Wallet(
            privateKey,
            ethers.getDefaultProvider(getRpcUrl(network)),
          ),
          options,
        )
      }

      const contractType = await resolveContractTypeWithCustom({
        sdk,
        contractAddress,
      })
      const contract = await sdk.getContract(contractAddress, contractType)
      result = await contract.call('cancel', [proposalId])
      resolve(result)
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Cancel proposal
 * @function
 * @name cancelProposal
 * @memberof hazama-lib/Others
 *
 * @example <caption>Cancel proposal with privateKey</caption>
 * import hazlib from hazama-lib
 * ...
 * const network = 'polygon'
 * const privateKey = 'dummy'
 * const contractAddress = '0x...'
 * const proposalId = '12...'
 * const clientKey = 'dummy'
 *
 * const result = hazlib.cancelProposal({ network, privateKey, contractAddress, proposalId, clientKey })
 * ...
 *
 * @example <caption>Cancel proposal with signer and relayer</caption>
 * import hazlib from hazama-lib
 * ...
 * const network = 'polygon'
 * const signer = <signer object>
 * const relayerUrl = 'https://api.defender.openzeppelin.com/autotasks/0f7cf308-.../runs/webhook/fd447953-...' // Openzeppelin webhook URI
 * const contractAddress = '0x...'
 * const proposalId = '12...'
 * const clientKey = 'dummy'
 *
 * const result = hazlib.cancelProposal({ network, signer, contractAddress, proposalId, relayerUrl, clientKey })
 * ...
 *
 * @param {object} params parameters for cancelling proposal
 * @param {string} params.network the chain name or rpc url
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions.<br/>If private key is not set, Signer is required.
 * @param {string} [params.privateKey] wallet's private key.<br/>If Signer is not set, private key is required.
 * @param {string} params.contractAddress the operator/spender's address
 * @param {string} params.proposalId proposal id
 * @param {string} [params.relayerUrl] relayer url (gasless transaction mode)
 * @param {string} [params.clientKey] client key
 * @returns {TransactionResult} result
 */
lib.cancelProposal = async ({
  network,
  signer,
  privateKey,
  contractAddress,
  proposalId,
  relayerUrl,
  clientKey,
  isCount = true,
}) => {
  let caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 52,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (!network) throw new Error(messages.NETWORK_UNDEFINED)

    caller = await getCaller({ network, signer, privateKey })
    const isHazamaNetwork = network === 'hazama'

    const result = !isHazamaNetwork
      ? await cancelProposalOnEVMNetwork({
          network,
          signer,
          privateKey,
          contractAddress,
          proposalId,
          relayerUrl,
        })
      : null

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 52,
      walletAddress: caller,
      status: RequestTransactionStatus.SUCCEEDED,
      transactionHash: formatTransactionHash({ input: result }),
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 52,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * Vote on EVM Network
 *
 * @private
 * @param {object} params parameters for voting proposal
 * @param {string} params.network the chain name or rpc url
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions.<br/>If private key is not set, Signer is required.
 * @param {string} [params.privateKey] wallet's private key.<br/>If Signer is not set, private key is required.
 * @param {string} params.contractAddress the token contract address
 * @param {string} params.proposalId proposal id
 * @param {(0|1|2)} params.voteType the vote type you want to cast, can be VoteType.Against = 0, VoteType.For = 1, or VoteType.Abstain = 2
 * @param {string} [params.relayerUrl] relayer url (gasless transaction mode)
 * @return {Promise<TransactionResult>} promise
 */
async function voteOnEVMNetwork({
  network,
  signer,
  privateKey,
  contractAddress,
  proposalId,
  voteType,
  relayerUrl,
}) {
  // eslint-disable-next-line no-async-promise-executor
  return new Promise(async (resolve, reject) => {
    try {
      let result = null
      let options = {
        readonlySettings: {
          rpcUrl: getRpcUrl(network),
        },
        ...(isArrayFull(gatewayUrls) ? { gatewayUrls } : {}),
        ...(isNil(thirdwebSecretKey) &&
        (!isNil(thirdwebClientId) || !isNil(thirdwebClientIdFromAPI))
          ? { clientId: thirdwebClientId || thirdwebClientIdFromAPI }
          : {}),
        ...(!isNil(thirdwebSecretKey) ? { secretKey: thirdwebSecretKey } : {}),
      }

      if (!network) throw new Error(messages.NETWORK_UNDEFINED)

      if (!signer && !privateKey)
        throw new Error(messages.SIGNER_PRIVATEKEY_UNDEFINED)

      if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

      if (!proposalId) throw new Error(messages.PROPOSAL_ID_UNDEFINED)

      if (!voteType && voteType !== 0)
        throw new Error(messages.VOTE_TYPE_UNDEFINED)

      if (relayerUrl) {
        options = {
          ...options,
          gasless: {
            openzeppelin: { relayerUrl },
          },
        }
      }

      let sdk = new ThirdwebSDK(network, options)
      if (signer) {
        sdk.updateSignerOrProvider(signer)
      }

      if (privateKey) {
        sdk = new ThirdwebSDK(
          new ethers.Wallet(
            privateKey,
            ethers.getDefaultProvider(getRpcUrl(network)),
          ),
          options,
        )
      }

      const contractType = await resolveContractTypeWithCustom({
        sdk,
        contractAddress,
      })
      const contract = await sdk.getContract(contractAddress, contractType)
      result = await contract.call('castVote', [proposalId, voteType])
      resolve(result)
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Vote on an active proposal
 * @function
 * @name vote
 * @memberof hazama-lib/Others
 *
 * @example <caption>Vote with privateKey</caption>
 * import hazlib from hazama-lib
 * ...
 * const network = 'polygon'
 * const privateKey = 'dummy'
 * const contractAddress = '0x...'
 * const proposalId = '12...'
 * const voteType = 1
 * const clientKey = 'dummy'
 *
 * const result = hazlib.vote({ network, privateKey, contractAddress, proposalId, voteType, clientKey })
 * ...
 *
 * @example <caption>Vote with signer and relayer</caption>
 * import hazlib from hazama-lib
 * ...
 * const network = 'polygon'
 * const signer = <signer object>
 * const relayerUrl = 'https://api.defender.openzeppelin.com/autotasks/0f7cf308-.../runs/webhook/fd447953-...' // Openzeppelin webhook URI
 * const contractAddress = '0x...'
 * const proposalId = '12...'
 * const voteType = 1
 * const clientKey = 'dummy'
 *
 * const result = hazlib.vote({ network, signer, contractAddress, proposalId, voteType, relayerUrl, clientKey })
 * ...
 *
 * @param {object} params parameters for delegating token
 * @param {string} params.network the chain name or rpc url
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions.<br/>If private key is not set, Signer is required.
 * @param {string} [params.privateKey] wallet's private key.<br/>If Signer is not set, private key is required.
 * @param {string} params.contractAddress the operator/spender's address
 * @param {string} params.proposalId proposal id
 * @param {(0|1|2)} params.voteType the vote type you want to cast, can be VoteType.Against = 0, VoteType.For = 1, or VoteType.Abstain = 2
 * @param {string} [params.relayerUrl] relayer url (gasless transaction mode)
 * @param {string} [params.clientKey] client key
 * @returns {TransactionResult} result
 */
lib.vote = async ({
  network,
  signer,
  privateKey,
  contractAddress,
  proposalId,
  voteType,
  relayerUrl,
  clientKey,
  isCount = true,
}) => {
  let caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 53,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (!network) throw new Error(messages.NETWORK_UNDEFINED)

    caller = await getCaller({ network, signer, privateKey })
    const isHazamaNetwork = network === 'hazama'

    const result = !isHazamaNetwork
      ? await voteOnEVMNetwork({
          network,
          signer,
          privateKey,
          contractAddress,
          proposalId,
          voteType,
          relayerUrl,
        })
      : null

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 53,
      walletAddress: caller,
      status: RequestTransactionStatus.SUCCEEDED,
      transactionHash: formatTransactionHash({ input: result }),
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 53,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * Execute proposal on EVM Network
 *
 * @private
 * @param {object} params parameters for executing proposal
 * @param {string} params.network the chain name or rpc url
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions.<br/>If private key is not set, Signer is required.
 * @param {string} [params.privateKey] wallet's private key.<br/>If Signer is not set, private key is required.
 * @param {string} params.contractAddress the token contract address
 * @param {string} params.description description
 * @param {ProposalExecution[]} params.executions executions
 * @param {string} [params.relayerUrl] relayer url (gasless transaction mode)
 * @return {Promise<TransactionResult>} promise
 */
async function executeProposalOnEVMNetwork({
  network,
  signer,
  privateKey,
  contractAddress,
  description,
  executions,
  relayerUrl,
}) {
  // eslint-disable-next-line no-async-promise-executor
  return new Promise(async (resolve, reject) => {
    try {
      let result = null
      let options = {
        readonlySettings: {
          rpcUrl: getRpcUrl(network),
        },
        ...(isArrayFull(gatewayUrls) ? { gatewayUrls } : {}),
        ...(isNil(thirdwebSecretKey) &&
        (!isNil(thirdwebClientId) || !isNil(thirdwebClientIdFromAPI))
          ? { clientId: thirdwebClientId || thirdwebClientIdFromAPI }
          : {}),
        ...(!isNil(thirdwebSecretKey) ? { secretKey: thirdwebSecretKey } : {}),
      }

      if (!network) throw new Error(messages.NETWORK_UNDEFINED)

      if (!signer && !privateKey)
        throw new Error(messages.SIGNER_PRIVATEKEY_UNDEFINED)

      if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

      if (!description) throw new Error(messages.DESCRIPTION_UNDEFINED)

      if (!isArrayFull(executions))
        throw new Error(messages.EXECUTIONS_UNDEFINED)

      if (relayerUrl) {
        options = {
          ...options,
          gasless: {
            openzeppelin: { relayerUrl },
          },
        }
      }

      let sdk = new ThirdwebSDK(network, options)
      if (signer) {
        sdk.updateSignerOrProvider(signer)
      }

      if (privateKey) {
        sdk = new ThirdwebSDK(
          new ethers.Wallet(
            privateKey,
            ethers.getDefaultProvider(getRpcUrl(network)),
          ),
          options,
        )
      }

      const contractType = await resolveContractTypeWithCustom({
        sdk,
        contractAddress,
      })
      const contract = await sdk.getContract(contractAddress, contractType)

      const tos = executions.map((p) => p.toAddress)
      const values = executions.map((p) => p.nativeTokenValue)
      const datas = executions.map((p) => p.transactionData)
      const descriptionHash = ethers.utils.id(description)

      result = await contract.call('execute', [
        tos,
        values,
        datas,
        descriptionHash,
      ])

      resolve(result)
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Execute proposal
 * @function
 * @name executeProposal
 * @memberof hazama-lib/Others
 *
 * @example <caption>Execute proposal with privateKey</caption>
 * import hazlib from hazama-lib
 * ...
 * const network = 'polygon'
 * const privateKey = 'dummy'
 * const contractAddress = '0x...'
 * const description = 'dummy'
 * const executions = []
 * const clientKey = 'dummy'
 *
 * const result = hazlib.executeProposal({ network, privateKey, contractAddress, description, executions, clientKey })
 * ...
 *
 * @example <caption>Execute proposal with signer and relayer</caption>
 * import hazlib from hazama-lib
 * ...
 * const network = 'polygon'
 * const signer = <signer object>
 * const relayerUrl = 'https://api.defender.openzeppelin.com/autotasks/0f7cf308-.../runs/webhook/fd447953-...' // Openzeppelin webhook URI
 * const contractAddress = '0x...'
 * const description = 'dummy'
 * const executions = []
 * const clientKey = 'dummy'
 *
 * const result = hazlib.executeProposal({ network, signer, contractAddress, description, executions, relayerUrl, clientKey })
 * ...
 *
 * @param {object} params parameters for executing proposal
 * @param {string} params.network the chain name or rpc url
 * @param {Signer} [params.signer] Signer to be used to sign messages and transactions.<br/>If private key is not set, Signer is required.
 * @param {string} [params.privateKey] wallet's private key.<br/>If Signer is not set, private key is required.
 * @param {string} params.contractAddress the operator/spender's address
 * @param {string} params.description description
 * @param {ProposalExecution[]} params.executions executions
 * @param {string} [params.relayerUrl] relayer url (gasless transaction mode)
 * @param {string} [params.clientKey] client key
 * @returns {TransactionResult} result
 */
lib.executeProposal = async ({
  network,
  signer,
  privateKey,
  contractAddress,
  description,
  executions,
  relayerUrl,
  clientKey,
  isCount = true,
}) => {
  let caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 54,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (!network) throw new Error(messages.NETWORK_UNDEFINED)

    caller = await getCaller({ network, signer, privateKey })
    const isHazamaNetwork = network === 'hazama'

    const result = !isHazamaNetwork
      ? await executeProposalOnEVMNetwork({
          network,
          signer,
          privateKey,
          contractAddress,
          description,
          executions,
          relayerUrl,
        })
      : null

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 54,
      walletAddress: caller,
      status: RequestTransactionStatus.SUCCEEDED,
      transactionHash: formatTransactionHash({ input: result }),
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 54,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * Vote on EVM Network
 *
 * @private
 * @param {object} params parameters
 * @param {string} params.proposalId proposalId
 * @param {string[]} params.choices choices
 * @param {SmartContract} params.contract contract
 * @return {Promise<any>} promise
 */
async function getProposalVotes({ proposalId, choices, contract }) {
  let result = []
  const votes = await contract.call('proposalVotes', [proposalId])
  if (isArrayFull(choices)) {
    for (let i = 0; i < choices.length; i++) {
      result.push({
        type: i,
        label: choices[i],
        count: parseFloat(ethers.utils.formatUnits(votes?.[i] || 0)),
      })
    }
  } else {
    result = [
      {
        type: VoteType.Against,
        label: 'No',
        count: parseFloat(ethers.utils.formatUnits(votes[0])),
      },
      {
        type: VoteType.For,
        label: 'Yes',
        count: parseFloat(ethers.utils.formatUnits(votes[1])),
      },
      {
        type: VoteType.Abstain,
        label: 'Abstain',
        count: parseFloat(ethers.utils.formatUnits(votes[2])),
      },
    ]
  }
  return result
}

/**
 * Has delegated on EVM Network
 *
 * @private
 * @param {object} params parameters for checking status of field "delegated"
 * @param {string} params.network the chain name or rpc url
 * @param {string} params.contractAddress contract address
 * @param {string} params.walletAddress wallet address
 * @return {Promise<boolean>} promise
 */
async function hasDelegatedOnEVMNetwork({
  network,
  contractAddress,
  walletAddress,
}) {
  return new Promise(async (resolve, reject) => {
    try {
      let result = false
      if (!network) throw new Error(messages.NETWORK_UNDEFINED)

      if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

      if (!walletAddress) throw new Error(messages.WALLET_ADDRESS_UNDEFINED)

      const sdk = new ThirdwebSDK(network, {
        ...(isNil(thirdwebSecretKey) &&
        (!isNil(thirdwebClientId) || !isNil(thirdwebClientIdFromAPI))
          ? { clientId: thirdwebClientId || thirdwebClientIdFromAPI }
          : {}),
        ...(!isNil(thirdwebSecretKey) ? { secretKey: thirdwebSecretKey } : {}),
      })
      const contractType = await resolveContractTypeWithCustom({
        sdk,
        contractAddress,
      })

      if (contractType !== ContractType.TokenERC20)
        throw new Error(messages.CONTRACT_TYPE_IS_INVALID)

      const contract = await sdk.getContract(contractAddress, contractType)
      const delegatee = await contract.call('delegates', [walletAddress])
      result = delegatee === walletAddress

      resolve(result)
    } catch (e) {
      reject(e)
    }
  })
}

/**
 * Has delegated (ERC20)
 * @function
 * @name hasDelegated
 * @memberof hazama-lib/Others
 *
 * @example <caption>[Has delegated on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const contractAddress = '0x...'
 * const walletAddress = '0x...'
 * const clientKey = 'dummy'
 *
 * const hasDelegated = await hazlib.hasDelegated({ network, contractAddress, walletAddress, clientKey })
 * ...
 *
 * @param {object} params parameters for checking status of field "delegated"
 * @param {string} params.network the chain name or rpc url
 * @param {string} params.contractAddress contract address
 * @param {string} params.walletAddress wallet address
 * @param {string} [params.clientKey] client key
 * @return {boolean}
 */
lib.hasDelegated = async ({
  network,
  contractAddress,
  walletAddress,
  clientKey,
  isCount = true,
}) => {
  const caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 56,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (!network) throw new Error(messages.NETWORK_UNDEFINED)

    const isHazamaNetwork = network === 'hazama'
    const result = !isHazamaNetwork
      ? await hasDelegatedOnEVMNetwork({
          network,
          contractAddress,
          walletAddress,
        })
      : null

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 56,
      walletAddress: caller,
      status: RequestTransactionStatus.SUCCEEDED,
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 56,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * Get vote info on EVM Network
 *
 * @private
 * @param {object} params parameters for getting vote info
 * @param {string} params.network the chain name or rpc url
 * @param {string} params.contractAddress contract address
 * @param {string} params.walletAddress wallet address
 * @param {string} params.proposalId proposal id
 * @return {Promise<any>} promise
 */
async function getVoteInfoOnEVMNetwork({
  network,
  contractAddress,
  walletAddress,
  proposalId,
}) {
  return new Promise(async (resolve, reject) => {
    try {
      if (!network) throw new Error(messages.NETWORK_UNDEFINED)

      if (!contractAddress) throw new Error(messages.CONTRACT_ADDRESS_UNDEFINED)

      if (!walletAddress) throw new Error(messages.WALLET_ADDRESS_UNDEFINED)

      const sdk = new ThirdwebSDK(network, {
        ...(isNil(thirdwebSecretKey) &&
        (!isNil(thirdwebClientId) || !isNil(thirdwebClientIdFromAPI))
          ? { clientId: thirdwebClientId || thirdwebClientIdFromAPI }
          : {}),
        ...(!isNil(thirdwebSecretKey) ? { secretKey: thirdwebSecretKey } : {}),
      })
      const contractType = await resolveContractTypeWithCustom({
        sdk,
        contractAddress,
      })

      if (contractType !== ContractType.VoteERC20)
        throw new Error(messages.CONTRACT_TYPE_IS_INVALID)

      const contract = await sdk.getContract(contractAddress, contractType)
      const tokenAddress = await contract.call('token')
      const tokenContract = await sdk.getContract(
        tokenAddress,
        ContractType.TokenERC20,
      )
      const [
        voteThreshold,
        proposalThreshold,
        getWeightForProposal,
        balance,
        tokenMetadata,
      ] = await Promise.all([
        contract.call('voteThreshold'),
        contract.call('proposalThreshold'),
        proposalId
          ? contract.call('getWeightForProposal', [proposalId, walletAddress])
          : 0,
        tokenContract.balanceOf(walletAddress),
        tokenContract.get(),
      ])
      const { name: tokenName, symbol: tokenSymbol } = tokenMetadata

      resolve({
        voteThreshold: parseFloat(formatUnitsWithBugNum(voteThreshold, 18)),
        proposalThreshold: parseFloat(
          formatUnitsWithBugNum(proposalThreshold, 18),
        ),
        getWeightForProposal: isArrayFull(getWeightForProposal)
          ? getWeightForProposal.map((item) =>
              parseFloat(formatUnitsWithBugNum(item, 18)),
            )
          : getWeightForProposal,
        balance: parseFloat(formatUnitsWithBugNum(balance.value, 18)),
        tokenAddress,
        tokenName,
        tokenSymbol,
      })
    } catch (e) {
      reject(e)
    }
  })
}

function formatUnitsWithBugNum(num, decimals) {
  const bigNum = Number(num).toLocaleString('fullwide', {
    useGrouping: false,
  })
  return Number(formatUnits((bigNum || 0).toString(), decimals)).toFixed()
}

/**
 * Get vote info
 * @function
 * @name getVoteInfo
 * @memberof hazama-lib/Others
 *
 * @example <caption>[Get vote info on EVM network]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const network = 'polygon'
 * const contractAddress = '0x...'
 * const walletAddress = '0x...'
 * const proposalId = '12...'
 * const clientKey = 'dummy'
 *
 * const result = await hazlib.getVoteInfo({ network, contractAddress, walletAddress, proposalId, clientKey })
 * ...
 *
 * @param {object} params parameters for getting vote info
 * @param {string} params.network the chain name or rpc url
 * @param {string} params.contractAddress contract address
 * @param {string} params.walletAddress wallet address
 * @param {string} params.proposalId proposal id
 * @param {string} [params.clientKey] client key
 * @return {any}
 */
lib.getVoteInfo = async ({
  network,
  contractAddress,
  walletAddress,
  proposalId,
  clientKey,
  isCount = true,
}) => {
  const caller = null
  const targetClientKey = clientKey || currentClientKey || null
  await checkRequestTransactionStatus({
    functionId: 57,
    clientKey: targetClientKey,
    isCount,
  })

  try {
    if (!network) throw new Error(messages.NETWORK_UNDEFINED)

    const isHazamaNetwork = network === 'hazama'
    const result = !isHazamaNetwork
      ? await getVoteInfoOnEVMNetwork({
          network,
          contractAddress,
          walletAddress,
          proposalId,
        })
      : null

    createRequestTransaction({
      clientKey: targetClientKey,
      functionId: 57,
      walletAddress: caller,
      status: RequestTransactionStatus.SUCCEEDED,
      isCount,
    })

    return result
  } catch (e) {
    await createFailedRequestTransaction({
      clientKey: targetClientKey,
      functionId: 57,
      walletAddress: caller,
      reason: e?.message || null,
      isCount,
    })
    lib.emit('error', e)
  }
}

/**
 * Set gateway urls
 * @function
 * @name setGatewayUrls
 * @memberof hazama-lib/Others
 *
 * @example <caption>[Set gateway urls]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const urls = ['https://ipfs.io/ipfs']
 *
 * const result = hazlib.setGatewayUrls({ urls })
 * ...
 *
 * @param {object} params parameters for setting gateway urls
 * @param {string[]} params.urls gateway urls
 * @return {void}
 */
lib.setGatewayUrls = async ({ urls }) => {
  gatewayUrls = urls
}

/**
 * Set thirdweb credentials
 * @function
 * @name setThirdwebCredentials
 * @memberof hazama-lib/Others
 *
 * @example <caption>[Set thirdweb credentials]</caption>
 * import hazlib from 'hazama-lib'
 * ...
 * const clientId = 'dummy'
 * const secretKey = 'dummy'
 *
 * const result = hazlib.setThirdwebCredentials({ clientId, secretKey })
 * ...
 *
 * @param {object} params parameters for setting credentials
 * @param {string} [params.clientId] client id
 * @param {string} [params.secretKey] secret key
 * @return {void}
 */
lib.setThirdwebCredentials = async ({ clientId, secretKey }) => {
  thirdwebClientId = clientId
  thirdwebSecretKey = secretKey
}

module.exports = lib
